#!/usr/bin/perl
#
# Bitflu - (C) 2006-2007 Adrian Ulrich
#
# Released under the terms of The "Artistic License 2.0".
# http://www.perlfoundation.org/legal/licenses/artistic-2_0.txt
#
#use warnings;
use strict;


### Performance: ut_pex kann ziemlich flooden und die CPU hochtreiben!
### Performance: hunter verbessern...

### === Ideas / Features ===
###
###  * Kill TIMEOUT_HUNT* => killSock could simply start a FastHunt ?
###  * No Endgame
###  * XXX_UGLY: Dynamic loop could need some cleanups : But it seems to work okay so far..
###  * CONSTANT Perlmod?
###  * Commit could stream via http? (-> preview)
###  * Bencoding may use a lot of memory (unfixable?)
###  * Fast doesn't send/receive Suggest messages (Never seen one)


use IO::Socket;
use IO::Select;
use Digest::SHA1;
use Data::Dumper;
use Getopt::Long;
use List::Util 'shuffle';
use Carp qw(confess croak cluck carp);
use URI::Escape;
use Scalar::Util; #For chroot
use File::Find;
use POSIX;

no utf8;


use constant BITFLU_BUILD   => '7224';  # YMDD (Y+M => HEX)
use constant BITFLU_VERSION => '0.32';

$SIG{PIPE} = $SIG{CHLD} = 'IGNORE';

$SIG{INT} = $SIG{HUP} = \&ConfDown;


## LogLevel constants. Everything above 'LOGLEVEL' will
## be printed to STDOUT.
use constant L_DEBUG       =>   1;  # LowLevel Debug
use constant L_DEBUG_STORE =>   2;  # Storemanager
use constant L_DEBUG_RPICE =>   3;  # ReadPiece
use constant L_DEBUG_SUGST =>   4;  # SuggestNewPiece
use constant L_DEBUG_CRX   =>   5;  # CreateRequest
use constant L_DEBUG_PBLSH =>   6;  # Publish
use constant L_DEBUG_CONFIG=> 899;  # dumps config
use constant L_DEBUG_SLACK =>   7;  # Choking/Interested stuff
use constant L_DEBUG_EXT   =>   8;  # Extended Protocol
use constant L_INFO        => 100;  # Info messages
use constant L_WARN        => 200;  # Warning messages
use constant L_PANIC       => 900;  # Panics client

## BitTorrentState constants. Do not change unless you know what
## you are doing.
use constant BTS_WRITE_HANDSHAKE          => 100; # -> Switches to BTS_READ_HANDSHAKE_RESPONSE
use constant BTS_READ_HANDSHAKE_RESPONSE  => 101; # -> Switches to BTS_WRITE_BITFIELD
use constant BTS_WRITE_BITFIELD           => 200; # -> Switches to BTS_IDLE
use constant BTS_IDLE                     => 300; # -> MAIN()
use constant BTS_CONNECTING               => 400; # -> Switches to BTS_WRITE_HANDSHAKE
use constant BTS_READ_CLIENT_HANDSHAKE    => 500;
use constant BTS_WRITE_CLIENT_HANDSHAKE_RESPONSE_AND_BITFIELD => 600;
use constant BTS_WRITE_EXTENSIONS         => 700;
use constant BTS_KILLME_PLEASE            => 998; # -> Request to close the connection on next bts_writer run
use constant BTS_UNDEF                    => 999; # -> Undef state

use constant EPROTO_UT_PEX   => 1;
use constant EPROTO_BF_SPEED => 2;

## State 'flow':
#### If *WE* connect to a client
# -> BTS_CONNECTING -> BTS_WRITE_HANDSHAKE -> BTS_READ_HANDSHAKE_RESPONSE -> BTS_WRITE_BITFIELD -> BTS_WRITE_EXTENSIONS -> BTS_IDLE
#
### If a client connects to US
# -> BTS_READ_CLIENT_HANDSHAKE -> BTS_WRITE_CLIENT_HANDSHAKE_RESPONSE_AND_BITFIELD -> BTS_WRITE_BITFIELD -> BTS_WRITE_EXTENSIONS -> BTS_IDLE


## Dynamic select loop
## we are running with DYNASLEEP_MAX but are going
## to switch to DYNASLEEP_MIN if we need to (Fast download)
## Change this to reduce CPU-Usage and/or improve download speed
use constant DYNASLEEP_MIN =>0.00001;
use constant DYNASLEEP_MAX =>0.05;

## Piece Size; do not change unless you know what you are doing!
use constant PIECESIZE    => (2**14);      #Default size for pieces
use constant MAXPIECESIZE => (2**17);      #Drop the connection if a client requests more..
use constant MAXONWIRE    => 512*1024;     #Drop connection if buffer grows above limit

## Various Timeout-Values.
use constant TIMEOUT_SHAKE             => 16;     # Small timeout for things like 'connecting' and handshaking / bitfield exchange                 {TIMEOUT}
use constant TIMEOUT_INTERESTED        => 63;     # ReScan interested chain
use constant TIMEOUT_HUNTER            => 60*15;  # Request a piece each X seconds even if we may have nothing good                               {TIMEOUT_HUNTER]
use constant TIMEOUT_LASTEVENT         => 60*20;  # Close connection with peer if it didn't request anything for X seconds or if we didn't send anything to it.
use constant TIMEOUT_WAITFORPIECE      => 150;    # Wait X seconds for a piece-chunk to arrive                                                     {TIMEOUT_WAITFORPIECE}
use constant TIMEOUT_FASTHUNTER        => 1;      # Set TIMEOUT_HUNTER after getting unchoked (Anti DOS protection -> unchoke floods)
use constant TIMEOUT_KEEPALIVE         => 110;    # Send a keepalive each X seconds                                                               {SEND_KEEPALIVE}
use constant TIMEOUT_KEEPALIVEDRIFT    => 115;    # Kill connection if last keepalive is older than TIMEOUT_KEEPALIVE + TIMEOUT_KEEPALIVEDRIFT    {SEND_KEEPALIVE}
use constant TIMEOUT_UNCHOKEROUND      => 30;     # Run an unchoke round each 30 seconds
use constant TIMEOUT_SENDPEX           => 82;     # Sendout PEX messages each 82 seconds


use constant MSG_CHOKE          => 0;
use constant MSG_UNCHOKE        => 1;
use constant MSG_INTERESTED     => 2;
use constant MSG_NOT_INTERESTED => 3;
use constant MSG_HAVE           => 4;
use constant MSG_BITFIELD       => 5;
use constant MSG_REQUEST        => 6;
use constant MSG_PIECE          => 7;
use constant MSG_CANCEL         => 8;
use constant MSG_PORT           => 9;  # DHT Extension
use constant MSG_WANT_METAINFO  => 10; # Unused
use constant MSG_METAINFO       => 11; # Unused
use constant MSG_SUSPECT_PIECE  => 12; # Unused
use constant MSG_SUGGEST_PIECE  => 13; # FastPeers
use constant MSG_HAVE_ALL       => 14; # FastPeers   | IMPLEMENTED  (Recv)       (Silly extension: no need to send this)
use constant MSG_HAVE_NONE      => 15; # FastPeers   | IMPLEMENTED  (Recv)       (Silly extension: no need to send this)
use constant MSG_REJECT_REQUEST => 16; # FastPeers   | IMPLEMENTED (Recv|Send)
use constant MSG_ALLOWED_FAST   => 17; # FastPeers   | IMPLEMENTED (Recv|Send)
use constant MSG_HOLE_PUNCH     => 18; # NAT-Unused
use constant MSG_UTORRENT_MSG   => 19; # Unused (??)
use constant MSG_UT_EXT         => 20; # Extended Handshake protocol



# RoundTrip requests:
use constant MAX_PREREQUEST_PIECES => 3;

# For people with way too many torrents:
use constant MAX_PEERS_HARDLIMIT => 450;

# Refuse to auto-commit if ration is below this:
use constant MIN_ACRATION        => 1.0;

# Default config file
use constant DEFAULT_CONFIG_FILE => '.bitflu.config';


my %BWSTATS            = (); #{UP}|{DOWN}
my %OUTQ               = (); #Outqueue
my %BFH                = (); #BufferHash (= Unpared data)
my %TCSTAT             = (); #TorrentClientStat
my %ACSTAT             = (); #AdminClientStat (same as TCSTAT but for telnet and http : user)
my %CONNECTING         = (); #BTS_CONNECTING sockets
my $TORRENT_UNIVERSE   = (); #Stores information about all torrents ; used {config} values are: seeding;uploading;downloading

my %getopts = ();
GetOptions(\%getopts, "telnet=n", "http=n", "torrent=n", "config=s", "help|h", "version|v");
$getopts{config} ||= DEFAULT_CONFIG_FILE;

my $config = ();

if($getopts{help}) {
	die << "EOF";
Usage: $0 [--telnet 4001] [--http 4081] [--torrent 6688] [--config .bitflu.config] [--help] [--version]

  -h, --help         Display this help and exit
  -v, --version      Display version information and exit
      --telnet       TCP-Port to listen for incoming telnet requests
      --http         TCP-Port to listen for incoming HTTP requests
      --torrent      TCP/UDP Port to use for BitTorrent / DHT communication
      --config       Use optional configuration file.

(C) 2006-2007 Adrian Ulrich , report problems/bugs/patches to <adrian\@blinkenlights.ch>
Released under the terms of The "Artistic License 2.0".
Web: http://bitflu.workaround.ch

EOF
}
elsif($getopts{version}) {
	die "This is BitFlu ".BITFLU_VERSION." (Build: ".BITFLU_BUILD.")\n";
}
elsif($getopts{config}) {
	($config->{client}, $config->{SYS}->{configfh}) = __load_config($getopts{config});
}

#Options with getopt support:
# 1. prio: commandline switch , 2.nd: rcfile 3th: defaults
$config->{client}->{port}   = $getopts{torrent} || $config->{client}->{port} || 6688;     # BitTorrent incoming port (UDP and TCP)

$config->{client}->{telnet} = $getopts{telnet} if defined($getopts{telnet});
$config->{client}->{telnet} = 4001 unless defined($config->{client}->{telnet});

$config->{client}->{telnetbind} ||= 0;                                          # Bind telnet to interface ; defaults to 0 (= ALL)


$config->{client}->{http}       = $getopts{http} if defined($getopts{http});    # Args
$config->{client}->{http}       = 0 unless defined($config->{client}->{http});  # No config? default

$config->{client}->{httpbind} ||= 0;                                            # Bind http-gui to interface ; defaults to 0 (= ALL)


$config->{client}->{maxpeers}    ||= 48; #Number of connections PER TORRENT
$config->{client}->{grabpeers}   ||= 32; #Connect to maximal X peers per request
$config->{client}->{maxup}       ||= 45; # Max upload bandwidth
$config->{client}->{unchokeup}   ||= 5;  # Reserve X KiB for each unchoked peer
$config->{client}->{rootdir}     ||= "./workdir";
$config->{client}->{autoloaddir} ||= "./autoload";
$config->{client}->{publishdir}  ||= "./publish";
$config->{client}->{commitdir}   ||= "./commited";
$config->{client}->{canceldir}   ||= "./canceled";
$config->{client}->{downloaddir} ||= "./downloading";
$config->{client}->{historydir}  ||= "./history";

$config->{client}->{kick_refreshpeers} ||= 111; #Refresh new peers each X seconds
$config->{client}->{kick_autoload}     ||= 60*6 unless defined($config->{client}->{kick_autoload});  #Scan Autoload folder each X-Seconds (may be null)
$config->{client}->{writes_per_sec}    ||= 3; # How often to call __bt_writer

$config->{client}->{trackerblacklist} ||= '';
$config->{client}->{loglevel}         ||= (L_WARN - 1);

$config->{client}->{userauth}      ||= '';
$config->{client}->{suppresshave}    = 0 unless defined($config->{client}->{suppresshave});
$config->{client}->{notifycompleted} = 0 unless defined($config->{client}->{notifycompleted}); #WALL about completed downloads

# Lock Non-Runtime vars
foreach (qw(rootdir downloaddir http httpbind port telnet telnetbind key peerid chroot runas_gid runas_uid userauth)) {
	$config->{client_locked}->{$_} = 1;
}

$config->{SYS}->{boottime} = time();
$config->{SYS}->{unclean_storage} = undef;
$config->{SYS}->{boot_id} = Bitflu::Utils::GetLinuxBootId(); # random string on !Linux
$config->{SYS}->{runtime_unchoked} = 0;

$config->{SYS}->{peer_key} = sprintf("%X",int(rand(0xFFFFFFFF))).sprintf("%X",int(rand(0xFFFFFFFF)));
$config->{SYS}->{peer_key} .= "X" x 20;
$config->{SYS}->{peer_key} = unpack("H40",$config->{SYS}->{peer_key});

$config->{SYS}->{peer_id}  = "-BF".BITFLU_BUILD."-".sprintf("%X",int(rand(0xFFFFFFFF))).sprintf("%X",int(rand(0xFFFFFFFF)));
$config->{SYS}->{peer_id} .= "X" x 20;
$config->{SYS}->{peer_id} = pack("A20",$config->{SYS}->{peer_id});

$config->{SYS}->{kick_refreshpeers} = 0;
$config->{SYS}->{kick_autoload}     = 0;  # ugly: used by other classes

print Data::Dumper::Dumper($config->{client});

# DHT:
# should fork, secure_process() and launch DHT


##############################################################################################################################


l(L_WARN, "Bitflu ".BITFLU_VERSION." (build: ".BITFLU_BUILD.") starting ...");

	my $torrent_handle = new IO::Select();	
	my $torrent_socket  = IO::Socket::INET->new(LocalPort=>$config->{client}->{port},
	                                           Proto=>'tcp', Listen=>1, ReuseAddr=>1) or die "$0 Unable to bind to $config->{client}->{port} : $!\n";
	$torrent_handle->add($torrent_socket);


	
	my $telnet_handle = new IO::Select();
	my $telnet_socket   = undef;
	
	if($config->{client}->{telnet}) {
		$telnet_socket = IO::Socket::INET->new(LocalPort=>$config->{client}->{telnet}, LocalAddr=>$config->{client}->{telnetbind},
		                                     Proto=>'tcp', Listen=>1, ReuseAddr=>1) 
		                                     or die "$0 Unable to bind to LocalAddr=>$config->{client}->{telnetbind}:$config->{client}->{telnet} : $!\n";
		$telnet_handle->add($telnet_socket);
	}
	
	
	my $http_handle = new IO::Select();
	my $http_socket = undef;
	
	if($config->{client}->{http}) {
		$http_socket = IO::Socket::INET->new(LocalPort=>$config->{client}->{http}, LocalAddr=>$config->{client}->{httpbind},
		                                           Proto=>'tcp', Listen=>1, ReuseAddr=>1) or die "$0 Unable to bind to $config->{client}->{http} : $!\n";
		$http_handle->add($http_socket);
	}
	
	
	my $ipc_handle = new IO::Select();
	my ($ipc_read, $ipc_write) = POSIX::pipe();
	   $ipc_handle->add($ipc_read);
	my $ipc_buff = undef;
	
	
	
	
secure_process();

l(L_WARN, "Init completed, now running with pid $$ ; uid = ($>|$<) / gid = ($)|$()");
l(L_WARN, "Initializing storage-layer and contacting trackers. This may take a few seconds ...");
# Init Bitflus Store
Bitflu::Store::_INIT();
# -> Load EXISTING torrent files
__load_old_torrents();
__init_lock();

l(L_WARN, "> Bitflu is ready!");
l(L_WARN, "> Run 'telnet $config->{client}->{telnetbind} $config->{client}->{telnet}' to use the telnet interface") if $config->{client}->{telnet};
l(L_WARN, "> Use http://$config->{client}->{httpbind}:$config->{client}->{http} to use the http interface")         if $config->{client}->{http};
l(L_WARN, "");
# -> Enter mainloop
main();


sub main {

	my ($NOWTIME, $KICK_UNCHOKER, $KICK_WRITE, $NUM_WRITES, $KICK_SEND_PEX) = undef;
	
	# Mainloop
	for(;;) {
			
			$NOWTIME = time();
			
			
			##########################################
			# Init a non-blocking connection
			foreach my $s (keys(%CONNECTING)) {
				my $sock = $CONNECTING{$s};
				my $sin  = $TCSTAT{$sock}->{SIN};
				connect($sock, $sin);
				if($!{'EISCONN'}) {
					$torrent_handle->add($sock);
					$TCSTAT{$sock}->{BTS}     = BTS_WRITE_HANDSHAKE;
					$TCSTAT{$sock}->{TIMEOUT} = $NOWTIME+TIMEOUT_SHAKE;
					delete($CONNECTING{$sock});
				}
				else {
					if($TCSTAT{$sock}->{TIMEOUT} < $NOWTIME) {
						killSock(undef,$sock);
					}
				}
				select(undef,undef,undef,DYNASLEEP_MIN);
			}
			
			# READ torrent
			my $XXX_UGLY_READ = __bt_reader($NOWTIME);
			
			if($NUM_WRITES != $config->{client}->{writes_per_sec} or $KICK_WRITE != $NOWTIME) {
				__bt_writer($NOWTIME);
				
				if($KICK_WRITE != $NOWTIME) {
					$KICK_WRITE = $NOWTIME;
					$NUM_WRITES = 0;
				}
				else {
					$NUM_WRITES++;
				}
			}
			
			
			if($config->{SYS}->{kick_autoload} < $NOWTIME && $config->{client}->{kick_autoload} != 0) {
				$config->{SYS}->{kick_autoload} = $NOWTIME+$config->{client}->{kick_autoload};
				__auto_loader(Bitflu::Store::__get_autoloadbase());
			}
			
			
			# Grab new nodes from .nodes and ask trackers
			if($config->{SYS}->{kick_refreshpeers} <= $NOWTIME) {
				$config->{SYS}->{kick_refreshpeers} = $NOWTIME+$config->{client}->{kick_refreshpeers};
				foreach my $xt (keys(%$TORRENT_UNIVERSE)) {
					Bitflu::Tracker::StartBoot($xt);
					GrabPeers($xt);
				}
			}
			
			if($KICK_SEND_PEX <= $NOWTIME) {
				$KICK_SEND_PEX = $NOWTIME + TIMEOUT_SENDPEX;
				foreach my $xt (keys(%$TORRENT_UNIVERSE)) {
					SendPEX($xt);
				}
			}
			
			if($KICK_UNCHOKER <= $NOWTIME) {
				$KICK_UNCHOKER = $NOWTIME + TIMEOUT_UNCHOKEROUND;
				StartUnchokeRound($NOWTIME);
			}
			
			
			## TELNET START
			__telnet_reader($NOWTIME);
			__telnet_writer($NOWTIME);
			## TELNET END
			
			__http_reader($NOWTIME);
			__http_writer($NOWTIME);
			my $XXX_UGLY_SLEEP = DYNASLEEP_MAX;
			if($XXX_UGLY_READ >= POSIX::BUFSIZ) {
				$XXX_UGLY_SLEEP = DYNASLEEP_MIN;
			}
			
			__ipc_reader($NOWTIME);
			
			CalculateBandwidhtStats($NOWTIME);
			select(undef,undef,undef,$XXX_UGLY_SLEEP);
	} #END mainloop
	
}



########################################################################
# Write lockfile and force verify if needed
sub __init_lock {
	my $storage_lockid = Bitflu::Store::LockStorage($config->{SYS}->{boot_id});
	
	if($storage_lockid ne $config->{SYS}->{boot_id}) {
		l(L_WARN, "** Bitflu didn't exit cleanly , forcing a HashCheck of all pieces");
		my $storage_errors = 0;
		$config->{SYS}->{unclean_storage} = 1;
		foreach(keys(%$TORRENT_UNIVERSE)) {
			l(L_WARN, "** Checking $_ ...");
			Bitflu::Store::VerifyStorageDir($_) or $storage_errors++;
		}
		l(L_PANIC, "** Corruptions found for $storage_errors torrents, please restart $0 to re-check everything.") if $storage_errors;
	}
	$config->{SYS}->{unclean_storage} = 0; # Storage is clean
}

########################################################################
# Chroots and drops privileges
sub secure_process {
	
	Carp::longmess("CARP_INIT");
	
	if($config->{client}->{chroot}) {
		l(L_WARN, "Chrooting into $config->{client}->{chroot}");
		chdir($config->{client}->{chroot})  or Carp::confess("chdir($config->{client}->{chroot}) failed : $!");
		chroot($config->{client}->{chroot}) or Carp::confess("chroot($config->{client}->{chroot}) failed : $!");
	}
	else {
		l(L_WARN, "No chroot option set :-( ; starting up without jail");
	}
	
	if($config->{client}->{runas_gid}) {
		l(L_WARN, "+ Changing GID to $config->{client}->{runas_gid}");
		$! = undef;
		$) = "$config->{client}->{runas_gid} $config->{client}->{runas_gid}";
			Carp::confess("Unable to set EGID: $!") if $!;
		
		$! = undef;
		$( = "$config->{client}->{runas_gid}";
			Carp::confess("Unable to set GID: $!") if $!;
	}
	
	if($config->{client}->{runas_uid}) {
		l(L_WARN, "+ Changing UID to $config->{client}->{runas_uid}");
		POSIX::setuid($config->{client}->{runas_uid}) or Carp::confess("Unable to set UID : $!");
	}
	
	Carp::confess("$0 refuses to run with uid 0 (= root)\n") unless $>;
}


########################################################################
# Load existing .torrent files and init $TORRENT_UNIVERSE
sub __load_old_torrents {
	opendir(OLD_TORRENTS, Bitflu::Store::__get_torrentbase()) or Carp::confess("Unable to open torrentsbase : No init?");
	while(my $file = readdir(OLD_TORRENTS)) {
		my $fp = Bitflu::Store::__get_torrentbase()."/$file";
		next unless (-f $fp);
		my $info_hash = __load_torrent($fp);
		
		unless($info_hash) {
			l(L_WARN, "** BROKEN torrent file in workdir: $fp");
		}
	}
	closedir(OLD_TORRENT);
}

########################################################################
# Load a single .torrent file from disk
sub __load_torrent {
	my($file) = @_;
	my $hash = Bitflu::Bencoding::torrent2hash($file);
	if($hash->{torrent_hash} =~ /^([a-z0-9]{40})$/ && !defined($TORRENT_UNIVERSE->{$hash->{torrent_hash}})) {
		$TORRENT_UNIVERSE->{$hash->{torrent_hash}}->{torrent} = \$hash;
		$TORRENT_UNIVERSE->{$hash->{torrent_hash}}->{sourcefile} = Bitflu::Store::StoreTorrent($file, $hash->{torrent_hash});
		Bitflu::Store::InitTorrent($hash->{torrent_hash}) or return undef;
		#Overwrite zero-conf that InitTorrent setup
		$TORRENT_UNIVERSE->{$hash->{torrent_hash}}->{config}     = Bitflu::Store::GetConfig($hash->{torrent_hash});
		$config->{SYS}->{kick_refreshpeers} = ( int(time())+ 2 ); # Give telnet some time for feedback
		#Bitflu::Tracker::StartBoot($hash->{torrent_hash});
		Bitflu::Store::HistoryAdd($hash->{torrent_hash});
		GrabPeers($hash->{torrent_hash});
		return $hash->{torrent_hash};
	}
	else {
		return undef;
	}
}

########################################################################
# Parse configuration file
sub __load_config {
	my($file) = @_;
	my %h = ();
	
	unless(open(CONF,"+<".$file)) {
		if(-f $file) {
			# User-Failesave: Do not startup with an existing (but unwriteable) file
			die "$0 : $file could not be opened (missing WRITE permissions)\n";
		}
		else {
			warn "$0 : Unable to open configuration file ($file) : $! ; Continuing with default options\n";
			return (\%h, undef);
		}
	}
	while(<CONF>) {
		my $line = $_;
		chomp($line);
		
		if($line =~ /^#/ or $line =~ /^\s*$/) {
			#comment
			next;
		}
		elsif($line =~ /^([a-zA-Z_]+)\s*=\s*(.*)$/) {
			die "$0 Duplicate definition of '$1' at line $. in configuration file '$file'\n" if $h{$1};
			$h{$1} = $2;
		}
		else {
			die "$0: Mumbled line '$line' at line $. in configuration file '$file'\n";
		}
	}
	return (\%h, *CONF);
}


########################################################################
# Create Bandwith Statistics: Fixme: This needs a redesign ;-)
sub CalculateBandwidhtStats {
	my($NOWTIME) = @_;

	$BWSTATS{RAW_UPX}   += $BWSTATS{UP};
	$BWSTATS{RAW_DOWNX} += $BWSTATS{DOWN};
	$BWSTATS{AR_UPX}    += $BWSTATS{UP};
	$BWSTATS{AR_DOWNX}  += $BWSTATS{DOWN};
	
	$BWSTATS{UP} = $BWSTATS{DOWN} = 0;     # Purge old values
	
	if((my $drift = ($NOWTIME - $BWSTATS{LASTCHECK}))) {
		# -> Calculate AVG values
		$BWSTATS{LASTCHECK} = $NOWTIME;
		
		$BWSTATS{__UPX}   = $BWSTATS{_UPX};
		$BWSTATS{__DOWNX} = $BWSTATS{_DOWNX};

		$BWSTATS{_UPX}   = $BWSTATS{UPX};
		$BWSTATS{_DOWNX} = $BWSTATS{DOWNX};
		
		$BWSTATS{UPX}   = $BWSTATS{AR_UPX}   = (($BWSTATS{RAW_UPX}/$drift)+$BWSTATS{_UPX}+$BWSTATS{__UPX})/3;
		$BWSTATS{DOWNX} = $BWSTATS{AR_DOWNX} = (($BWSTATS{RAW_DOWNX}/$drift)+$BWSTATS{_DOWNX}+$BWSTATS{__DOWNX})/3;
		$BWSTATS{RAW_UPX} = $BWSTATS{RAW_DOWNX} = 0; # Purge
	}
	
}

########################################################################
# Invalidate TIMEOUT_INTERESTED for X peers
sub StartFastHunt {
	my($info_hash, $piecemap, $hits_wanted) = @_;
	foreach my $cc (keys(%TCSTAT)) {
		my $tref = $TCSTAT{$cc};
		last if     $hits_wanted < 1;
		next unless $tref->{BTS} == BTS_IDLE;
		next unless $tref->{INFO_HASH} eq $info_hash;
		next if     $tref->{ME_INTERESTED}; # No need to do an interested hunt!
		my $could_be_interesting = 0;
		foreach my $xpiece (keys(%$piecemap)) {
			if(Bitflu::Utils::GetBit(\$tref->{BITFIELD},$xpiece)) {
				$could_be_interesting = 1;
				last;
			}
		}
		next unless $could_be_interesting;
		next if int(keys(%{$tref->{PIECE_INWORK}})) == MAX_PREREQUEST_PIECES; # Full
		$hits_wanted--;
		$tref->{TIMEOUT_INTERESTED} = 0; # do it NOW
	}
}




sub StartUnchokeRound {
my($NOWTIME) = @_;

	my $bestlist = ();     # Best uploaders list
	my $optlist  = ();     # Optimistic unchoke list
	my $seedlist = ();     # Seeding torrents
	my $peercount = ();    # Peers per torrent
	my $do_unchoke = ();   # List of unchokeable peers
	my $num_torrents = 0;
	my $hard_unchoke_limit = int($config->{client}->{maxup}); # Provide 1KiB per client
	
	$config->{SYS}->{runtime_unchoked} = 0;
	
	
	# Count # of torrents and check if they are completed (SEEDING)
	foreach my $torrent_hash (keys(%$TORRENT_UNIVERSE)) {
		$num_torrents++;

		if($TORRENT_UNIVERSE->{$torrent_hash}->{is_complete}) {
			if($TORRENT_UNIVERSE->{$torrent_hash}->{config}->{autocommit} >= MIN_ACRATION &&
			   Bitflu::Utils::GetRation($torrent_hash) >= $TORRENT_UNIVERSE->{$torrent_hash}->{config}->{autocommit}) {
				# Do not autocommit again:
				$TORRENT_UNIVERSE->{$torrent_hash}->{config}->{autocommit} = 0;
				Bitflu::RPC::wall("$torrent_hash: starting autocommit");
				Bitflu::Commands::ForkCommit([$torrent_hash], 1);
			}
			next;
		}
		
		my $donestats = Bitflu::Store::GetDoneStats($torrent_hash);
		my $totalpieces = Bitflu::Store::GetNumberOfPieces($torrent_hash)+1;
		if($donestats->{pieces} == $totalpieces) {
			$TORRENT_UNIVERSE->{$torrent_hash}->{is_complete} = 1;
			if($config->{client}->{notifycompleted}) {
				Bitflu::RPC::wall("$torrent_hash: download completed");
			}
		}
		delete($TORRENT_UNIVERSE->{$torrent_hash}->{stats_uploaded});
		delete($TORRENT_UNIVERSE->{$torrent_hash}->{stats_downloaded});
	}
	
	$num_torrents ||= 1;

	my $unchokeable = int(($config->{client}->{maxup}/$config->{client}->{unchokeup})/$num_torrents)-1;
	my $optcredits  = 1;
	
	if($unchokeable < 1) {
		$optcredits = 0; # Disable optimistic unchoking.. tja
		$unchokeable = 1;
	}

	
	
	# Build the toplist
	foreach my $cc (keys(%TCSTAT)) {
		my $tref = $TCSTAT{$cc};
		next unless $tref->{BTS} == BTS_IDLE;
		
		if($TORRENT_UNIVERSE->{$tref->{INFO_HASH}}->{is_complete}) {
			push(@{$seedlist->{$tref->{INFO_HASH}}->{int($tref->{'UNCHOKED@'})}}, $cc);
		}
		else {
			push(@{$bestlist->{$tref->{INFO_HASH}}->{int($tref->{STATS_DOWNLOADED})}}, $cc);
		}
		
		push(@{$optlist->{$tref->{INFO_HASH}}->{$tref->{'CONNECT@'}}}, $cc); # Optimistic unchoking
		$peercount->{$tref->{INFO_HASH}}++;
		$TORRENT_UNIVERSE->{$tref->{INFO_HASH}}->{stats_uploaded}   += $tref->{STATS_UPLOADED};
		$TORRENT_UNIVERSE->{$tref->{INFO_HASH}}->{stats_downloaded} += $tref->{STATS_DOWNLOADED};
		delete($tref->{STATS_UPLOADED});
		delete($tref->{STATS_DOWNLOADED});
	}
	
	my $UCSTAT_NUMSORT = 0;
	my $UCSTAT_LASTUN  = 0;
	my $UCSTAT_OPTUN   = 0;
	
	foreach my $torrent_hash (keys(%$peercount)) {
		my $peers_for_torrent      = $peercount->{$torrent_hash};  # We got X connected peers
		my $credits_for_torrent    = $unchokeable;
		my $credits_for_optimistic = $optcredits;
		
		$TORRENT_UNIVERSE->{$torrent_hash}->{stats_uploaded}   = ($TORRENT_UNIVERSE->{$torrent_hash}->{stats_uploaded}/1024/TIMEOUT_UNCHOKEROUND);
		$TORRENT_UNIVERSE->{$torrent_hash}->{stats_downloaded} = ($TORRENT_UNIVERSE->{$torrent_hash}->{stats_downloaded}/1024/TIMEOUT_UNCHOKEROUND);
		
		# Get $credits_for_torrent best uploaders
		foreach my $numsort (sort({$b <=> $a} (keys(%{$bestlist->{$torrent_hash}})))) {
			foreach my $sock (@{$bestlist->{$torrent_hash}->{$numsort}}) {
				if($TCSTAT{$sock}->{PEER_INTERESTED} && $credits_for_torrent) {
					$do_unchoke->{$sock} = defined;
					$credits_for_torrent--;
					$UCSTAT_NUMSORT++;
				}
			}
		}
		
		# Seeding state: Unchoke some new clients
		foreach my $lastunchoke (sort({$a <=> $b} (keys(%{$seedlist->{$torrent_hash}})))) {
			foreach my $sock (@{$seedlist->{$torrent_hash}->{$lastunchoke}}) {
				if($TCSTAT{$sock}->{PEER_INTERESTED} && $credits_for_torrent) {
					$do_unchoke->{$sock} = defined;
					$credits_for_torrent--;
					$UCSTAT_LASTUN++;
				}
			}
		}
		
		foreach my $optunchoke (sort({$b <=> $a} (keys(%{$optlist->{$torrent_hash}})))) {
			foreach my $sock (@{$optlist->{$torrent_hash}->{$optunchoke}}) {
				if(!defined($do_unchoke->{$sock}) && $credits_for_torrent && $credits_for_optimistic && !int(rand(int($peers_for_torrent/3)))) {
					$do_unchoke->{$sock} = defined;
					$credits_for_torrent--;
					$credits_for_optimistic--;
					$UCSTAT_OPTUN++;
				}
			}
		}
	}
	
	foreach my $cc (List::Util::shuffle(keys(%TCSTAT))) {
		my $tref = $TCSTAT{$cc};
		next unless $tref->{BTS} == BTS_IDLE;
		next if $config->{SYS}->{runtime_unchoked} >= $hard_unchoke_limit;
		if(defined($do_unchoke->{$cc})) {
			$config->{SYS}->{runtime_unchoked}++;
			if($tref->{PEER_CHOKE}) {
				_xwA($cc, Bitflu::Protocol::CreateUnchoke($tref));
			}
		}
		elsif($tref->{PEER_CHOKE} == 0) {
			_xwA($cc, Bitflu::Protocol::CreateChoke($tref));
			if($tref->{EXTENSIONS}->{FAST} && $tref->{OUTSTANDING_REQUESTS}) {
				# -> Fast Peer
				for(my $i=0; $i<int(@{$tref->{OUTSTANDING_REQUESTS}});$i++) {
					my $this_piece = ${$tref->{OUTSTANDING_REQUESTS}}[$i]->{index};
					next if defined($tref->{FASTSET}->{$this_piece}); # -> Fastallowed
					_xwA($cc, Bitflu::Protocol::CreateReject(${$tref->{OUTSTANDING_REQUESTS}}[$i]));
					splice(@{$tref->{OUTSTANDING_REQUESTS}},$i,1);
					l(L_DEBUG, "Canceling OUTSTANDING_REQUEST $this_piece");
				}
			}
			else {
				delete($tref->{OUTSTANDING_REQUESTS});
			}
		}
	}
	
	


}


########################################################################
# Sendout PEX messages to peers
sub SendPEX {
	my($torrent_hash) = @_;
	
	return if $TORRENT_UNIVERSE->{$torrent_hash}->{content}->{info}->{private}; # Do not do pex for private torrents
	
	my $pexnodes_limit = 12; # Do not send more than 12 nodes per flood
	my $pexrcv_limit   = 8; # Do not send stuff to more than 8 peers
	
	
	my @pexnodes = ();
	my @pexrcv   = ();
	foreach my $cc (List::Util::shuffle(keys(%TCSTAT))) {
		my $tref = $TCSTAT{$cc};
		next unless $tref->{BTS} == BTS_IDLE;
		next unless $tref->{INFO_HASH} eq $torrent_hash;
		if($tref->{PORT} && int(@pexnodes) < $pexnodes_limit) {
			push(@pexnodes, {encoded=>Bitflu::Utils::EncodeCompactIp({ip=>$tref->{IP}, port=>$tref->{PORT}}), crypt=>int($tref->{EXTENSIONS}->{ENCRYPT})});
		}
		if($tref->{EXTENSIONS}->{_UT_PEX} && int(@pexrcv) < $pexrcv_limit) {
			push(@pexrcv, {sock=>$cc, msgid=>$tref->{EXTENSIONS}->{_UT_PEX}});
		}
	}
	
	my $xref = {dropped=>undef, added=>undef, 'added.f'=>undef};
	foreach my $nh (@pexnodes) {
		$xref->{'added'} .= $nh->{encoded};
		
		if($nh->{crypt}) {
			# -> Peer would support encryption
			$xref->{'added.f'} .= chr(1);
		}
		else {
			# -> No encryption
			$xref->{'added.f'} .= chr(0);
		}
	}
	$xref->{dropped} = undef;
	
	foreach my $hr (@pexrcv) {
		my $onwire = Bitflu::Protocol::CreateExtended($hr->{msgid},Bitflu::Bencoding::encode($xref));
		_xwA($hr->{sock}, $onwire);
	}
	
}


########################################################################
# Check how many connections this torrent has
sub CheckSingleNodeConnectX {
	my($hash,$ip) = @_;
	Carp::confess("INVALID TORRENT HASH") unless defined($TORRENT_UNIVERSE->{$hash});
	my $xnum = 0;
	   $xnum = keys(%{$TORRENT_UNIVERSE->{$hash}->{peerslist}->{$ip}})
	           if defined($TORRENT_UNIVERSE->{$hash}->{peerslist}->{$ip});
	return $xnum;
}

########################################################################
# Add a connection for this ip to torrent
sub AddSingleNodeConnectX {
	my($hash,$ip,$glob) = @_;
	Carp::confess("INVALID TORRENT HASH") unless defined($TORRENT_UNIVERSE->{$hash});
	Carp::confess("Doubleadd?") if defined($TORRENT_UNIVERSE->{$hash}->{peerslist}->{$ip}->{$glob});
	Carp::confess("No IP?")     unless $ip;
	Carp::confess("No Sock?")   unless $glob;
	return ($TORRENT_UNIVERSE->{$hash}->{peerslist}->{$ip}->{$glob}++);
}


sub IpIsBlacklisted {
	my($ip) = @_;
	return undef; #Not implemented yet
}


########################################################################
# Save conf
sub ConfDown {
	my($sig) = @_;
	l(L_WARN, "Got signal $sig : shutting down...");
	# Save config before going down
	foreach my $k (keys(%$TORRENT_UNIVERSE)) {
		l(L_WARN, " => Saving Configuration of '$k'");
		Bitflu::Store::SaveConfig($k, $TORRENT_UNIVERSE->{$k}->{config});
	}
	
	if($sig eq 'INT') {
		if($config->{SYS}->{unclean_storage} == 0) {
			Bitflu::Store::UnlockStorage() or Carp::confess("Unable to unlock storage dir!");
		}
		l(L_WARN, "Bitflu shutdown completed");
		exit(0);
	}
}


########################################################################
# Kill connection with a client and cleanup all buffers
sub killSock {
	my($select_handle,$socket) = @_;
	l(L_DEBUG, "killSock($select_handle, $socket)");
	$select_handle->remove($socket) if $select_handle;
	delete $BFH{$socket};
	delete $OUTQ{$socket};
	Bitflu::Store::FreeWorkPieces($TCSTAT{$socket});
	
	if($TCSTAT{$socket}->{INFO_HASH}) {
		Carp::confess("Invalid INFO_HASH") unless $TORRENT_UNIVERSE->{$TCSTAT{$socket}->{INFO_HASH}};
		Carp::confess("No IP?!")           unless $TCSTAT{$socket}->{IP};
		
		if ( defined($TORRENT_UNIVERSE->{$TCSTAT{$socket}->{INFO_HASH}}->{peerslist}->{$TCSTAT{$socket}->{IP}}) ) {
			# IP in Storage
			delete($TORRENT_UNIVERSE->{$TCSTAT{$socket}->{INFO_HASH}}->{peerslist}->{$TCSTAT{$socket}->{IP}}->{$socket});
			if(CheckSingleNodeConnectX($TCSTAT{$socket}->{INFO_HASH}, $TCSTAT{$socket}->{IP}) == 0) {
				delete($TORRENT_UNIVERSE->{$TCSTAT{$socket}->{INFO_HASH}}->{peerslist}->{$TCSTAT{$socket}->{IP}}); #Freeup some memory
			}
		}
	}
	
	delete $TCSTAT{$socket};
	delete $ACSTAT{$socket};
	delete $CONNECTING{$socket};
	close $socket or Carp::confess("DIEDIEDIE <$socket> : $! => $?");
}



########################################################################
# Get data read from client
sub _xGETtoread {
	my($sock) = @_;
	return $BFH{$sock};
}

########################################################################
# Clean (delete) data read from client
sub _xCLEANtoread {
	my($sock,$to) = @_;
	if($to < 0) { delete $BFH{$sock} ; return }
	$BFH{$sock} =  substr($BFH{$sock},$to);
}


########################################################################
# STORE readed data APPEND
sub _xrA {
	my($sock,$buff) = @_;
	$BFH{$sock} .= $buff;
}

########################################################################
# STORE Readed data NEW
sub _xrN {
	my($sock,$buff) = @_;
	$BFH{$sock} = $buff;
}

###

########################################################################
# Clean (delete) data read from client
sub _xCLEANtowrite {
	my($sock,$to) = @_;
	if($to < 0) { delete $OUTQ{$sock}; return }
	$OUTQ{$sock} =  substr($OUTQ{$sock},$to);
}

########################################################################
# Get data needed to send to client
sub _xGETtowrite {
	my($sock) = @_;
	return $OUTQ{$sock};
}

########################################################################
# WRITE data NEW
sub _xwN {
	my($sock, $txt) = @_;
	$OUTQ{$sock} = $txt;
}

########################################################################
# WRITE data Append
sub _xwA {
	my($sock, $txt) = @_;
	$OUTQ{$sock} .= $txt;
}

########################################################################
# Print a logmsg
sub l {
	my($level, @message) = @_;
	printf("".gmtime()." %03d # %s\n", $level, "@message") if $level >= $config->{client}->{loglevel};
	die if $level == L_PANIC;
}


########################################################################
# Connect to a BT-Client and populate TCSTAT for this connection
sub BT_initConnection {
	my($hr) = @_;
	
	Carp::confess("Called without {hash}") unless $hr->{hash};
	Carp::confess("Invalid hash")          unless $TORRENT_UNIVERSE->{$hr->{hash}};
	Carp::confess("Missing SOURCE")        unless $hr->{source};
	
	if(IpIsBlacklisted($hr->{ip})) {
		l(L_WARN, "$hr->{ip} is blacklisted, won't create an outgoing connection to this peer");
		return undef;
	}
	elsif(CheckSingleNodeConnectX($hr->{hash}, $hr->{ip})) {
		l(L_DEBUG, "Outgoing connection for $hr->{ip} ==> $hr->{hash} not established: We allready got one!");
		return undef;
	}
	
	my $proto = getprotobyname('tcp');
	my $sock  = undef;
	socket($sock, AF_INET,SOCK_STREAM,$proto) or Carp::confess("Unhandled state : $!\n");
	my $sin = sockaddr_in($hr->{port}, inet_aton($hr->{ip}));
	unblock_FH($sock) or Carp::confess("Unable to unblock FH <$sock> : $!\n");

	my $NOWTIME = int(time());

	$CONNECTING{$sock} = $sock;
	
	$TCSTAT{$sock} and Carp::confess("BUGBUG: <$sock> was closed but somehow \$TCSTAT was still alive! PANICED");
	$TCSTAT{$sock}->{BTS} = BTS_CONNECTING;
	$TCSTAT{$sock}->{SIN} = $sin;
	$TCSTAT{$sock}->{IP}  = $hr->{ip};
	$TCSTAT{$sock}->{PORT}= $hr->{port};
	$TCSTAT{$sock}->{INFO_HASH}       = $hr->{hash};
	$TCSTAT{$sock}->{ME_CHOKE}        = 1;
	$TCSTAT{$sock}->{PEER_CHOKE}      = 1;
	$TCSTAT{$sock}->{ME_INTERESTED}   = 0;
	$TCSTAT{$sock}->{PEER_INTERESTED} = 0;
	$TCSTAT{$sock}->{TIMEOUT}         = $NOWTIME + TIMEOUT_SHAKE;
	$TCSTAT{$sock}->{SEND_KEEPALIVE}  = $NOWTIME + TIMEOUT_KEEPALIVE;
	$TCSTAT{$sock}->{SOURCE}          = $hr->{source};
	$TCSTAT{$sock}->{'CONNECT@'}      = $NOWTIME;
	AddSingleNodeConnectX($hr->{hash},$hr->{ip},$sock);
}


# fixme: what is more expensive: TCSTAT keys or readdir?
sub GrabPeers {
	my($hash) = @_;
	my $nodesdir       = Bitflu::Store::_getnodesdir($hash);
	my $peerslots_left = GetMaxConnections() - int(keys(%TCSTAT));
	   $peerslots_left = $config->{client}->{grabpeers} if $peerslots_left > $config->{client}->{grabpeers};
	return undef if $peerslots_left < 1; # Fast exit
	
	my @shuffleable = ();
	
	opendir(NODES, $nodesdir) or return 1;
	while(defined (my $node = readdir(NODES))) {
		next unless $node =~ /\/?([\d\.]+):(\d+)$/;
		my $fullfile = "$nodesdir/$node";
		my $xip = $1; my $xport = $2;
		unlink($fullfile) or Carp::confess("Unable to remove $fullfile");
		push(@shuffleable, {ip=>$xip, port=>$xport, file=>$fullfile});
	}
	closedir(NODES);
	
	foreach my $xref (List::Util::shuffle(@shuffleable)) {
		next if $peerslots_left-- < 1;
		BT_initConnection({ip=>$xref->{ip}, port=>$xref->{port}, hash=>$hash, source=>"Fix DHTboot"});
	}
	
}

########################################################################
# Returns (dynamic) number of max-connections
sub GetMaxConnections {
	my $num_torrents = int(keys(%$TORRENT_UNIVERSE));                  # Number of torrents we are downloading
	my $num_maxconn  = $num_torrents * $config->{client}->{maxpeers};  # Maximal connections possible
	   $num_maxconn  = MAX_PEERS_HARDLIMIT if $num_maxconn > MAX_PEERS_HARDLIMIT;
	return $num_maxconn;
}


########################################################################
# Unblock a filehandle
sub unblock_FH {
	my($cfh) = @_;
	return undef unless $cfh;
	my $flags = fcntl($cfh, F_GETFL, 0)
							or return undef;
	fcntl($cfh, F_SETFL, $flags | O_NONBLOCK)
	or return undef;
	return 1;
}


########################################################################
# Scan given folder for files and try to load them all as .torrent
sub __auto_loader {
	my($xdir) = @_;
	opendir(AUTOLOAD, $xdir) or Carp::confess("Unable to open AutoLoad folder $xdir : $!");
	while(my $toload = readdir(AUTOLOAD)) {
		my $fullpath = $xdir."/".$toload;
		next unless -f $fullpath;
		my $result = __load_torrent($fullpath);
		if($result) {
			l(L_WARN, "[AutoLoad] '$fullpath' loaded (Hash: $result)");
			unlink($fullpath) or Carp::cluck("Unable to remove '$fullpath' : $!");
		}
		else {
			l(L_WARN, "[AutoLoad] '$fullpath' NOT loaded due to error");
		}
	}
	closedir(AUTOLOAD);
}

sub __ipc_reader {
	foreach($ipc_handle->can_read(0)) {
		my $buff = undef;
		POSIX::read($_,$buff,POSIX::BUFSIZ);
		_xrA('IPC_MAGIC', $buff);
		my $full_data = _xGETtoread('IPC_MAGIC');
		if(substr($full_data,-1,1) eq "\n") {
			_xCLEANtoread('IPC_MAGIC',-1);
			Bitflu::Commands::ExecuteIpcCommands($full_data);
		}
	}
}

sub __http_reader {
	my @all_http_can_read = $http_handle->can_read(0);
	foreach my $sock (@all_http_can_read) {
		if($sock == $http_socket) {
			#-> A new client
			my $new_sock = $sock->accept();
			$http_handle->add($new_sock);
		}
		else {
			#-> Command from connected client
			my $buffer = undef;
			my $recv = $sock->recv($buffer, POSIX::BUFSIZ,0);
			my $bufflen = length($buffer);
			if(defined($recv) and $bufflen > 0) {
				_xrA($sock,$buffer); #Append buffer
				my $full_data = _xGETtoread($sock);
				if(substr($full_data,-4,4) eq "\r\n\r\n") {
					_xCLEANtoread($sock,-1);
					_xwN($sock, Bitflu::HTTPgui::main($full_data));
				}
			}
			else {
				killSock($http_handle,$sock);
			}
		}
	}
}

########################################################################
# Write output to telnet client
sub __http_writer {
	my @all_http_can_write = $http_handle->can_write(DYNASLEEP_MIN);
	my $clients = 0;
	foreach my $sock (@all_http_can_write) {
		$clients++;
		if(my $buff = _xGETtowrite($sock)) {
			my $data_sent = syswrite($sock,$buff);
			_xCLEANtowrite($sock,$data_sent);
			if(!defined($data_sent)) {
				l(L_DEBUG, "Syswrite to <$sock> failed: $! ; Closing connection with peer");
				killSock($http_handle,$sock);
				print "[HTTP] $sock write error\n";
			}
			elsif($data_sent == length($buff)) { # Buffer empty, close connection
				killSock($http_handle,$sock);
			}
		}
	}
	return $clients;
}


########################################################################
# Read telnet commands
sub __telnet_reader {

use constant TELNET_CLIENT_AUTHENTICATED => 100;
use constant TELNET_CLIENT_WAITFORUSER   => 101;
use constant TELNET_CLIENT_WAITFORPASS   => 102;

			my @all_telnet_can_read = $telnet_handle->can_read(0);
			foreach my $sock (@all_telnet_can_read) {
				if($sock == $telnet_socket) {
					#-> A new client
					my $new_sock = $sock->accept();
					$telnet_handle->add($new_sock);
					
					_xwN($new_sock, "## Welcome to BitFlu\r\n# ");
					
					if($config->{client}->{userauth}) {
						_xwA($new_sock, "Username: ");
						$ACSTAT{$new_sock}->{AUTH_STATE} = TELNET_CLIENT_WAITFORUSER;
					}
					else {
						$ACSTAT{$new_sock}->{AUTH_STATE} = TELNET_CLIENT_AUTHENTICATED;
					}
					
					l(L_WARN,"[Telnet] New client <$sock> connected (".$new_sock->peerhost().")");
				}
				else {
					#-> Command from connected client
					my $buffer = undef;
					my $recv = $sock->recv($buffer, POSIX::BUFSIZ,0);
					my $bufflen = length($buffer);
					if(defined($recv) and $bufflen > 0) {
						_xrA($sock,$buffer); #Append buffer
						my $full_data = _xGETtoread($sock);
	
						if(substr($full_data,-2,2) eq "\r\n") {
							_xCLEANtoread($sock,-1);
							if($ACSTAT{$sock}->{AUTH_STATE} == TELNET_CLIENT_AUTHENTICATED) {
								Bitflu::Commands::ExecuteTelnetCommands($telnet_handle,$sock,$full_data);
								_xwA($sock, "# ");
							}
							elsif($ACSTAT{$sock}->{AUTH_STATE} == TELNET_CLIENT_WAITFORUSER) {
								$ACSTAT{$sock}->{GIVEN_USER} = substr($full_data,0,-2);
								_xwA($sock, "# Password : ");
								$ACSTAT{$sock}->{AUTH_STATE} = TELNET_CLIENT_WAITFORPASS;
							}
							elsif($ACSTAT{$sock}->{AUTH_STATE} == TELNET_CLIENT_WAITFORPASS) {
								my $pass = substr($full_data,0,-2);
								if(Bitflu::Utils::UserPasswordIsOkay($ACSTAT{$sock}->{GIVEN_USER},$pass)) {
									$ACSTAT{$sock}->{AUTH_STATE} = TELNET_CLIENT_AUTHENTICATED;
									_xwA($sock, "# ");
								}
								else {
									killSock($telnet_handle,$sock);
								}
							}
							else {
								Carp::confess("WRONG ACSTATE: $ACSTAT{$sock}->{AUTH_STATE}");
							}
						}
					}
					else {
						l(L_INFO, "[Telnet] Closing connection with <$sock>");
						killSock($telnet_handle,$sock);
					}
				}
			}
}


########################################################################
# Write output to telnet client
sub __telnet_writer {
	my @all_telnet_can_write =$telnet_handle->can_write(DYNASLEEP_MIN);
	my $clients = 0;
	foreach my $sock (@all_telnet_can_write) {
		$clients++;
		if(my $buff = _xGETtowrite($sock)) {
			my $data_sent = syswrite($sock,$buff);
			_xCLEANtowrite($sock,$data_sent);
			if(!defined($data_sent)) {
				l(L_DEBUG, "Syswrite to <$sock> failed: $! ; Closing connection with peer");
				killSock($telnet_handle,$sock);
			}
		}
	}
	return $clients;
}


########################################################################
# Read data from BT clients
sub __bt_reader {
	my($NOW) = @_;
			my @all_bt_can_read = $torrent_handle->can_read(0);
			my $XXX_UGLY_ALL_LEN = 0;
			foreach my $sock (@all_bt_can_read) {
				if($sock == $torrent_socket) {
				# -> New client
					my ($new_sock) = $sock->accept();
					my  $new_ip = undef;
					
					if(!defined($new_sock) or !($new_ip = $new_sock->peerhost())) {
						l(L_INFO, "$sock -> accept() failed: '$!' ; Closing connection.");
						$new_sock->close() if $new_sock; # Let IO::Socket take care of this mess
						next;
					}
					elsif(!unblock_FH($new_sock)) { #For BSD
						l(L_WARN, "Unable to unblock FH <$new_sock>");
						$new_sock->close() if $new_sock;
						next;
					}
					
					#Connection established: Build a small ENV for this client
					$torrent_handle->add($new_sock);
					$TCSTAT{$new_sock} and Carp::confess("BUGBUG: This is a duplicate!"); #Paranoia check
					$TCSTAT{$new_sock}->{BTS} = BTS_KILLME_PLEASE;
					
					#Check if we have a slot for the client
					if(int(keys(%TCSTAT)) >= GetMaxConnections() ) {
						l(L_WARN, "maxpeers reached; Closing connection with new client: $new_ip");
						# -> BTS is set to BTS_KILLME_PLEASE : this will kill the client in the next run of bt_writer
					}
					else {
					l(L_INFO, "<$new_sock> Torrent client connected, waiting for HANDSHAKE");
						# Fixme: BT_Init has almost the same code.. 
						$TCSTAT{$new_sock}->{BTS}             = BTS_READ_CLIENT_HANDSHAKE;
						$TCSTAT{$new_sock}->{IP}              = $new_ip;
						$TCSTAT{$new_sock}->{PORT}            = undef; # Unknown for incoming messages, but PEX may update this
						$TCSTAT{$new_sock}->{ME_CHOKE}        = 1;
						$TCSTAT{$new_sock}->{PEER_CHOKE}      = 1;
						$TCSTAT{$new_sock}->{ME_INTERESTED}   = 0;
						$TCSTAT{$new_sock}->{PEER_INTERESTED} = 0;
						$TCSTAT{$new_sock}->{TIMEOUT}         = $NOW+TIMEOUT_SHAKE;
						$TCSTAT{$new_sock}->{SEND_KEEPALIVE}  = $NOW+TIMEOUT_KEEPALIVE;
						$TCSTAT{$new_sock}->{SOURCE}          = "INCOMING";
						$TCSTAT{$new_sock}->{'CONNECT@'}      = $NOW;
					}
				}
				else {
				#-> Data received from client
					my $buffer        = undef;
					my $recv          = recv($sock,$buffer, POSIX::BUFSIZ,0);
					my $bufflen       = length($buffer);
					my $tref          = $TCSTAT{$sock};
					if(defined($recv) and $bufflen > 0) {
						_xrA($sock,$buffer); #Append buffer
						$tref->{STATS_DOWNLOADED}       += $bufflen;
						$tref->{STATS_DOWNLOADED_TOTAL} += $bufflen;
						$BWSTATS{DOWN}                  += $bufflen;
						
						RELOOP:
						my $full_data     = _xGETtoread($sock);
						my $full_data_len = length($full_data);
						my $state         = $tref->{BTS};
						$XXX_UGLY_ALL_LEN += $full_data_len;
						
						
						$tref->{RECEIVED_KEEPALIVE} = $NOW + TIMEOUT_KEEPALIVE; # Message received!
						
						if($full_data_len > MAXONWIRE) {
							l(L_WARN, "Too much data on wire from client $tref->{IP} ; Closing connection");
							$tref->{BTS} = BTS_KILLME_PLEASE;
							next;
						}
						elsif($state == BTS_IDLE) {
							if($full_data_len >= 4) { #Minimal length of a BT message
								my $msglen = unpack("N", substr($full_data,0,4));
								if($msglen == 0) {
									l(L_DEBUG,"   <- KEEPALIVE received");
									# No need to update RECEIVED_KEEPALIVE as we did this a few lines ago :-)
									_xCLEANtoread($sock,4); #Remove from buffer
									goto RELOOP;  #Pickup more
								}
								elsif(length($full_data) >= ($msglen+4)) {
									# -> We got a full packet with payload!
									my $msgtype = unpack("c", substr($full_data,4,1));
									_xCLEANtoread($sock, $msglen+4); #Remove parsed data from buffer
									
									###################################################################
									if($msgtype == MSG_PIECE) { # Piece
										my $piece_id          = unpack("N", substr($full_data,5,4));
										my $offset            = unpack("N", substr($full_data,9,4));
										my $content           = substr($full_data,13,$msglen-9);
										my $current_info_hash = $tref->{INFO_HASH};
										
										$TORRENT_UNIVERSE->{$current_info_hash}->{config}->{downloaded} += $msglen-9;
										
										# ..and increment the clients timeout
										$tref->{TIMEOUT_WAITFORPIECE} = $NOW + TIMEOUT_WAITFORPIECE;
										$tref->{TIMEOUT_LASTEVENT}    = $NOW + TIMEOUT_LASTEVENT;
										
										if(!defined($tref->{PIECE_INWORK}{$piece_id})) {
											# -> This client sent us something we didn't request.
											# We are going to DROP this.
											# Storing it anyway wouldn't be a good option because this client has to be considered
											# 'rouge' => The data sent may be even faulty (SHA1 corruption) : Such a case would force
											# us to download the whole piece again!
											l(L_WARN, "<$sock> sent us piece an unrequested piece: $piece_id\@$current_info_hash ; Killing connection with 'rouge' client.");
											$tref->{BTS} = BTS_KILLME_PLEASE;
											next;
										}
										
										
										# Attemp to store data
										my $stored = Bitflu::Store::StorePiece({hash=>$current_info_hash, index=>$piece_id, offset=>$offset, data=>$content});
										
										if(defined($stored->{verified})) {
											# Chunk completed + Verified
											l(L_DEBUG_STORE, " => Piece $piece_id\@$current_info_hash is VERIFIED (got it from $sock)");
											Bitflu::Utils::SetBit(\$TORRENT_UNIVERSE->{$current_info_hash}->{bitfield},$piece_id);
											
											# Pollute the network with funny HAVE messages
											foreach my $sname (keys(%TCSTAT)) {
												my $ref = $TCSTAT{$sname};
												if($ref->{BTS} == BTS_IDLE && $ref->{INFO_HASH} eq $current_info_hash &&
												   ( ($config->{client}->{suppresshave} == 0) or (Bitflu::Utils::GetBit(\$ref->{BITFIELD},$piece_id) == 0) ) ) {
													l(L_DEBUG, "  --> Sending *HAVE $piece_id* to $sname");
													_xwA($sname, Bitflu::Protocol::CreateHave($piece_id));
												}
											}
										}
										
										if($stored->{rollover}) {
											# -> Storemanager told us to do a RollOver (= Grab next piece, not the next offset)
											delete($tref->{PIECE_INWORK}{$piece_id}); # Storemanager saved piece, so delete thisone..
											my $suggested_piece = Bitflu::Store::SuggestNewPiece($tref,$piece_id); #Suggest a new piece but NOT $piece_id
											l(L_INFO, "StoreManager suggested to try a new piece -> $suggested_piece\@$tref->{INFO_HASH} (old: $piece_id)");
											if(defined($suggested_piece)) {
												#Change store
												$stored->{index} = $suggested_piece;
												($stored->{offset}, $stored->{size} ) = Bitflu::Store::GetFreePieceOffset($tref->{INFO_HASH},$stored->{index});
												Bitflu::Store::SetPieceAsInwork($tref->{INFO_HASH},$stored->{index});
												$tref->{PIECE_INWORK}{$stored->{index}} = $stored;
											}
											elsif(int(keys(%{$tref->{PIECE_INWORK}})) > 0) {
												l(L_DEBUG_SLACK, "Client has NO new pieces for me but some requests are still unfullfilled ; ");
												goto RELOOP;
											}
											else {
												l(L_DEBUG_SLACK, "Client has NO pieces left for $current_info_hash ; Sending NotInterested ");
												_xwA($sock, Bitflu::Protocol::CreateNotInterested($tref));
												goto RELOOP; #Fixme: This is ugly but we need to aviod sending a request!
											}
										}
										
										if(!Bitflu::Store::PieceRequestOkForClient($stored->{index}, $tref)) {
											# check if we are still allowed to download from the client (We could have get choked ...)
											l(L_WARN, "<$sock> Aborted request for $stored->{index} due to client state change");
											Bitflu::Store::SetPieceAsFree($tref->{INFO_HASH}, $stored->{index});
											delete($tref->{PIECE_INWORK}{$stored->{index}});
											goto RELOOP;
										}
										
										# Still here? -> Request piece from client
										_xwA($sock, Bitflu::Protocol::CreateRequest($stored,$tref));
									}
									elsif($msgtype == MSG_HAVE) {
									# -> HAVE
										my $piece_id = unpack("N", substr($full_data,5,4));
										l(L_DEBUG, "<$sock> set HAVE $piece_id");
										Bitflu::Utils::SetBit(\$tref->{BITFIELD},$piece_id);
										if(Bitflu::Utils::GetBit(\$TORRENT_UNIVERSE->{$tref->{INFO_HASH}}->{bitfield},$piece_id) == 0) {
											$tref->{TIMEOUT_INTERESTED} = $tref->{TIMEOUT_HUNTER} = $NOW+TIMEOUT_FASTHUNTER; # Kick hunt if we are interested
										}
									}
									elsif($msgtype == MSG_BITFIELD) {
									# -> Bitfield
										my $bitfield = substr($full_data,5,$msglen-1);
										if(defined($tref->{BITFIELD})) {
											l(L_WARN, "<$sock> sent us a seconds bitfield ; no thanks!");
											$tref->{BTS} = BTS_KILLME_PLEASE;
											next;
										}
										$tref->{BITFIELD} = $bitfield;
										$tref->{TIMEOUT_INTERESTED} = $tref->{TIMEOUT_HUNTER} = 0;            # Kick hunt if we are interested
										my $len_bf = length(unpack("B*",$bitfield));                          # Worst case: 1
										my $parts  = Bitflu::Store::GetNumberOfPieces($tref->{INFO_HASH});    # Worst case: 0
										if($len_bf < ($parts+1)) {
											l(L_WARN, "<$sock> Client sent a faulty bitfield ; Closing connection");
											$tref->{BTS} = BTS_KILLME_PLEASE;
											next;
										}
									}
									elsif($msgtype == MSG_UNCHOKE) {
									# -> Client unchoked me!
										if($tref->{ME_CHOKE}) {
											$tref->{ME_CHOKE} = 0;
											$tref->{TIMEOUT_INTERESTED} = $tref->{TIMEOUT_HUNTER} = $NOW + TIMEOUT_FASTHUNTER; # Kick hunt if we ain't interested
											l(L_DEBUG, "<$sock> Got unchoked :-)");
										}
									}
									elsif($msgtype == MSG_CHOKE) { ## -> Choke
										l(L_DEBUG, "<$sock> Got choked :-(");
										if($tref->{ME_CHOKE} == 0) {
											$tref->{ME_CHOKE} = 1;
											unless($tref->{EXTENSIONS}->{FAST}) {
												# Non-fast client: Free all pieces
												Bitflu::Store::FreeWorkPieces($tref);
											}
											else {
												# Void: Client shall send us REJECT messages after we got choked.
											}
										}
									}
									elsif($msgtype == MSG_INTERESTED) { ## -> interested
										$tref->{PEER_INTERESTED} = 1;
									}
									elsif($msgtype == MSG_NOT_INTERESTED) { ## -> Not interested
										$tref->{PEER_INTERESTED} = 0;
									}
									elsif($msgtype == MSG_REQUEST) { ## -> request
										my $piece_id = unpack("N", substr($full_data,5,4));
										my $offset   = unpack("N", substr($full_data,9,4));
										my $length   = unpack("N", substr($full_data,13,4));
										my $ref = {offset=>$offset, index=>$piece_id, size=>$length, hash=>$tref->{INFO_HASH}};
										
										l(L_DEBUG, "<$sock> sent a request for piece $piece_id at offset $offset with len $length");
										if($length > MAXPIECESIZE) {
											l(L_WARN, "<$sock> Requested too much data for $piece_id -> $length . Closing connection");
											$tref->{BTS} = BTS_KILLME_PLEASE;
											next;
										}
										elsif($TORRENT_UNIVERSE->{$tref->{INFO_HASH}}->{config}->{paused} == 0 &&
										        ($tref->{PEER_CHOKE} == 0 or ($tref->{EXTENSIONS}->{FAST} && defined($tref->{FASTSET}->{$piece_id})))) {
											$tref->{TIMEOUT_LASTEVENT}    = $NOW + TIMEOUT_LASTEVENT;
											push(@{$tref->{OUTSTANDING_REQUESTS}},$ref);
										}
										elsif($tref->{EXTENSIONS}->{FAST}) {
											l(L_WARN, "<$sock> Rejecting non-fast request for $piece_id");
											_xwA($sock,Bitflu::Protocol::CreateReject($ref));
										}
										else {
											l(L_INFO, "<$sock> Choked/Paused client asked for $piece_id !");
										}
									}
									elsif($msgtype == MSG_CANCEL) {
										my $piece_id = unpack("N", substr($full_data,5,4));
										my $offset   = unpack("N", substr($full_data,9,4));
										my $length   = unpack("N", substr($full_data,13,4));
										l(L_INFO, "Canceling $offset|$piece_id|$length");
										
										if($tref->{OUTSTANDING_REQUESTS}) {
											for(my $i=0;$i<int(@{$tref->{OUTSTANDING_REQUESTS}});$i++) {
												my $orkref = ${$tref->{OUTSTANDING_REQUESTS}}[$i];
												if($orkref->{offset} == $offset && $orkref->{index} == $piece_id && $orkref->{size} == $length) {
													splice(@{$tref->{OUTSTANDING_REQUESTS}},$i,1);
												}
											}
										}
									}
									elsif($msgtype == MSG_PORT) { #DHT PORT message
										my $dht_port = unpack("n", substr($full_data,5,2));
										l(L_WARN, "[DHT] New bootnode for $tref->{INFO_HASH} : ".$tref->{IP}.":$dht_port");
										Bitflu::Store::AddDhtNode($tref->{INFO_HASH}, $tref->{IP}, $dht_port);
									}
									elsif($msgtype == MSG_HAVE_ALL) {
										# FAST EXTENSION
										$tref->{BITFIELD} = pack("B*", (1 x (Bitflu::Store::GetNumberOfPieces($tref->{INFO_HASH})+1)));
										$tref->{TIMEOUT_INTERESTED} = $tref->{TIMEOUT_HUNTER} = $NOW+TIMEOUT_FASTHUNTER; # Kick hunt if we ain't interested
									}
									elsif($msgtype == MSG_HAVE_NONE) {
										# FAST EXTENSION
										$tref->{BITFIELD} = pack("B*", (0 x (Bitflu::Store::GetNumberOfPieces($tref->{INFO_HASH})+1)));
									}
									elsif($msgtype == MSG_REJECT_REQUEST) {
										my $rejected_piece = unpack("N", substr($full_data,5,4));
										if(!defined($tref->{PIECE_INWORK}{$rejected_piece})) {
											l(L_WARN, "<$sock> rejected an unrequested piece ; Killing connection with rouge fast-peer");
											$tref->{BTS} = BTS_KILLME_PLEASE;
											next;
										}
										else {
											l(L_WARN, "<$sock> rejected piece $rejected_piece ; freeing it");
											Bitflu::Store::SetPieceAsFree($tref->{INFO_HASH}, $rejected_piece);
											delete($tref->{PIECE_INWORK}{$rejected_piece});
											$tref->{TIMEOUT_HUNTER} = $NOW+TIMEOUT_HUNTER;
										}
									}
									elsif($msgtype == MSG_ALLOWED_FAST) {
										my $fast = unpack("N", substr($full_data,5,4));
										$tref->{FASTALLOWED}->{$fast} = defined;
										l(L_DEBUG, "<$sock> sent allowed_fast($fast)");
										$tref->{TIMEOUT_INTERESTED} = $tref->{TIMEOUT_HUNTER} = $NOW+TIMEOUT_FASTHUNTER; # Kick hunt if we ain't interested
									}
									elsif($msgtype == MSG_UT_EXT) {
										my $ext_type = unpack("c", substr($full_data,5,1));
										my $bencoded = substr($full_data,6,$msglen-2);
										my $utmsg = Bitflu::Bencoding::decode($bencoded);
										
										if(!$ext_type) {
											# -> Handshake
											foreach my $ext_name (keys(%{$utmsg->{m}})) {
												next unless $ext_name =~ /^(ut_pex|bf_speed)$/;
												my $ext_name_internal = "_".uc($ext_name);
												if($utmsg->{m}->{$ext_name} == 0) {
													delete($tref->{EXTENSIONS}->{$ext_name_internal});
												}
												else {
													$tref->{EXTENSIONS}->{$ext_name_internal} = $utmsg->{m}->{$ext_name};
												}
											}
											$tref->{PORT} = int($utmsg->{p}) unless defined($tref->{PORT});
											$tref->{EXTENSIONS}->{ENCRYPT} = 1 if $utmsg->{e} == 1;
										}
										elsif($ext_type == $tref->{EXTENSIONS}->{_UT_PEX}) {
											my @iplist = Bitflu::Utils::DecodeCompactIp($utmsg->{added});
											Bitflu::Store::AddTorrentNodes(\@iplist, $tref->{INFO_HASH}, "UT_PEX");
										}
										else {
											l(L_INFO, "<$sock> Unknown extension type '$ext_type' ignored");
										}
										
									}
									else {
										l(L_WARN, "  <-> Unknown msgtype $msgtype ".unpack("H*",$msgtype));
									}
									
									goto RELOOP;
								}
								else {
									# -> Not a complete chunk.. wait..
								}
								
							}
							
						}
						elsif($state == BTS_READ_HANDSHAKE_RESPONSE) {
							my $hsresp = Bitflu::Protocol::ParseHandshake($full_data);
							_xrN($sock,$hsresp->{ol}); #Save unparsed data
							
							if($hsresp->{info_hash} eq $tref->{INFO_HASH}) {
								l(L_INFO, "<$sock> Handshake finished, info hash -> $tref->{INFO_HASH}");
								$tref->{CLIENT_ID} = $hsresp->{client_id};
								$tref->{BTS}       = BTS_WRITE_BITFIELD;
								$tref->{TIMEOUT}   = $NOW+TIMEOUT_SHAKE;
								$tref->{EXTENSIONS}= $hsresp->{extensions};
								
								if($tref->{EXTENSIONS}->{FAST}) {
									#Calculate FastSet for this connection
									$tref->{FASTSET}   = Bitflu::Utils::CreateFastSet($tref->{INFO_HASH}, $tref->{IP}, (Bitflu::Store::GetNumberOfPieces($tref->{INFO_HASH})+1));
								}
								#We may have unparsed data:
								goto RELOOP;
							}
						}
						elsif($state == BTS_READ_CLIENT_HANDSHAKE) {
							my $hsresp = Bitflu::Protocol::ParseHandshake($full_data);
							_xrN($sock,$hsresp->{ol}); #Save unparsed data
							if($TORRENT_UNIVERSE->{$hsresp->{info_hash}}) { # Shared file!
								if($hsresp->{client_id} eq $config->{SYS}->{peer_id}) {
									#Whoot: This client (connected itself to us) has the same peerid as we got..
									# Most likely we are connecting to ourself:
									l(L_INFO, "<$sock> sent my own peerid ($config->{SYS}->{peer_id}) ; Looks like i've connected to myself! dropping connection!");
									$tref->{BTS} = BTS_KILLME_PLEASE;
									next;
								}
								elsif(IpIsBlacklisted($tref->{IP})) {
									l(L_WARN, "<$sock> $tref->{IP} has been blacklisted, refusing connection");
									$tref->{BTS} = BTS_KILLME_PLEASE;
									next;
								}
								elsif(CheckSingleNodeConnectX($hsresp->{info_hash}, $tref->{IP})) {
									l(L_DEBUG, "<$sock> $tref->{IP} has a connection with $hsresp->{info_hash} ; closing 2nd connection");
									$tref->{BTS} = BTS_KILLME_PLEASE;
									next;
								}
								else {
									l(L_INFO, "<$sock> Handshake parsed, sending my own bitfield to client");
									$tref->{CLIENT_ID}  = $hsresp->{client_id};
									# -> Connection not initiated by us? -> Write response and bitfield
									$tref->{BTS}        = BTS_WRITE_CLIENT_HANDSHAKE_RESPONSE_AND_BITFIELD;
									$tref->{TIMEOUT}    = $NOW+TIMEOUT_SHAKE;
									$tref->{INFO_HASH}  = $hsresp->{info_hash};
									$tref->{EXTENSIONS} = $hsresp->{extensions};
									
									if($tref->{EXTENSIONS}->{FAST}) {
										#Calculate FastSet for this connection
										$tref->{FASTSET}   = Bitflu::Utils::CreateFastSet($tref->{INFO_HASH}, $tref->{IP}, (Bitflu::Store::GetNumberOfPieces($tref->{INFO_HASH})+1))
									}
									
									AddSingleNodeConnectX($hsresp->{info_hash}, $tref->{IP}, $sock);
									goto RELOOP;
								}
							}
						}
						else {
							l(L_DEBUG,"Unhandled state: $state");
						}
					}
					else {
						l(L_INFO, "[TorrentClient] <$sock> dropped connection");
						#Note: we can't use BTS_KILLME_PLEASE here as the socket
						# IS allready dead..
						killSock($torrent_handle,$sock);
					}
				}
			}
	return $XXX_UGLY_ALL_LEN;
}

########################################################################
# Write data to BT clients
sub __bt_writer {
	my($NOW) = @_;
	my @all_bt_can_write =$torrent_handle->can_write(DYNASLEEP_MIN);
	
	
	my $unchoked_peers = $config->{SYS}->{runtime_unchoked};
	   $unchoked_peers ||= 1;
	my $upcredits = ( (($config->{client}->{maxup}*1024)-($BWSTATS{RAW_UPX}+$BWSTATS{AR_UPX})/2)/$unchoked_peers);
	
	if($upcredits < 0) {
		return undef;
	}
	
	foreach my $sock (@all_bt_can_write) {
		my $tref = $TCSTAT{$sock};
		my $state = $tref->{BTS};
		
		
		if($state == BTS_IDLE) {
			#We are mostly here :-)
			#########################################################
			# -> Check if we need to send a keepalive

			### => Keepalive checks
			if($tref->{RECEIVED_KEEPALIVE} <= ($NOW - TIMEOUT_KEEPALIVEDRIFT)) {
				l(L_DEBUG, "<$sock> Timed out: No keepalive!");
				killSock($torrent_handle,$sock);
				next;
			}
			elsif($tref->{TIMEOUT_LASTEVENT} < ($NOW - TIMEOUT_LASTEVENT)) {
				l(L_DEBUG, "<$sock> is silent ; closing connection");
				killSock($torrent_handle,$sock);
				next;
			}
			elsif($tref->{TIMEOUT_WAITFORPIECE} < $NOW && int(keys(%{$tref->{PIECE_INWORK}})) > 0) {
				l(L_INFO, "<$sock> -> Rehunting: Waited much too long for piece $tref->{PIECE_INWORK}\@$tref->{INFO_HASH}");
				StartFastHunt($tref->{INFO_HASH}, $tref->{PIECE_INWORK}, MAX_PREREQUEST_PIECES+2); # Search 5 replacement-peers for this. StartFastHunt is expensive, but it's called seldom
				foreach my $to_cancel (keys(%{$tref->{PIECE_INWORK}})) {
					_xwA($sock, Bitflu::Protocol::CreateCancel($tref->{PIECE_INWORK}{$to_cancel}));
				}
				Bitflu::Store::FreeWorkPieces($tref);
				$tref->{TIMEOUT_INTERESTED} = $NOW + TIMEOUT_INTERESTED;
				$tref->{TIMEOUT_HUNTER}     = $NOW + TIMEOUT_HUNTER;
				_xwA($sock, Bitflu::Protocol::CreateNotInterested($tref));
			}
			elsif($tref->{SENT_KEEPALIVE} <= $NOW) {
				l(L_DEBUG, "<$sock> -> Sending a Keepalive");
				$tref->{SENT_KEEPALIVE} = $NOW + TIMEOUT_KEEPALIVE;
				_xwA($sock, Bitflu::Protocol::CreateKeepalive());
			}
			

			
			if($tref->{ME_INTERESTED} == 0 && $tref->{TIMEOUT_INTERESTED} < $NOW) {
				# Check if we are interested
				if(defined(Bitflu::Store::SuggestNewUnchokedPiece($tref,undef))) {
					l(L_INFO, "<$sock> : Send 'Interested request' as client has some yummy pieces for us !");
					_xwA($sock, Bitflu::Protocol::CreateInterested($tref));
				}
				else {
					# Nope: We are not interested
					l(L_INFO, "<$sock> : We are currently NOT interested .. nothing sent");
					$tref->{TIMEOUT_INTERESTED} = $NOW + TIMEOUT_INTERESTED;
				}
			}
			elsif($tref->{ME_INTERESTED} == 1 && $tref->{TIMEOUT_HUNTER} < $NOW && int(keys(%{$tref->{PIECE_INWORK}})) < MAX_PREREQUEST_PIECES) {
				
				my $suggested_piece = Bitflu::Store::SuggestNewPiece($tref);
				
				if(defined($suggested_piece) && Bitflu::Store::PieceRequestOkForClient($suggested_piece, $tref)) {
					# -> Request this piece from client
					my ($sug_off, $sug_size) = Bitflu::Store::GetFreePieceOffset($tref->{INFO_HASH},$suggested_piece);
					_xwA($sock, Bitflu::Protocol::CreateRequest({index=>$suggested_piece, 
					      offset=> $sug_off, size=>$sug_size},$tref));
					Bitflu::Store::SetPieceAsInwork($tref->{INFO_HASH},$suggested_piece);
					$tref->{PIECE_INWORK}{$suggested_piece} = {index=>$suggested_piece, offset=>$sug_off, size=>$sug_size};
					$tref->{TIMEOUT_WAITFORPIECE} = $NOW + TIMEOUT_WAITFORPIECE;
					l(L_INFO, " Requested $suggested_piece from $sock");
				}
				elsif(defined($suggested_piece)) {
					$tref->{TIMEOUT_HUNTER} = $NOW + TIMEOUT_HUNTER;
					l(L_INFO, "Hunter <$sock> has no requestable pieces");
				}
				elsif(defined(Bitflu::Store::SuggestNewUnchokedPiece($tref)) or int(keys(%{$tref->{PIECE_INWORK}})) > 0) { # Performance: Likely to hit SNP_CACHE
					$tref->{TIMEOUT_HUNTER} = $NOW + TIMEOUT_HUNTER;
					l(L_INFO, "Hunter <$sock> has no yet-requestable pieces"); #We are still waiting for data..
				}
				else {
					l(L_INFO, "Hunter: <$sock> no pieces ; sending not-interested");
					$tref->{TIMEOUT_INTERESTED} = $NOW + TIMEOUT_INTERESTED;
					_xwA($sock, Bitflu::Protocol::CreateNotInterested($tref));
				}
			}
			
		} #########################################################
		else {
			# First checkout if this non-idle client timed out
			if($tref->{TIMEOUT} < $NOW) {
				l(L_INFO, "<$sock> timed out in state <$state>");
				killSock($torrent_handle,$sock);
				next;
			}
			
			# Still here? Ok, lets check if we got something to send
			if($state == BTS_WRITE_HANDSHAKE) {
				l(L_DEBUG, "<$sock> -> Performing Handshake");
				_xwA($sock, Bitflu::Protocol::CreateHandshake({info_hash=>$tref->{INFO_HASH}, EXT_eproto=>1, EXT_fastpeers=>1,
				                                                  peer_id=>$config->{SYS}->{peer_id}}));
				$tref->{BTS}    = BTS_READ_HANDSHAKE_RESPONSE;
				$tref->{TIMEOUT} = $NOW+TIMEOUT_SHAKE;
			}
			elsif($state == BTS_WRITE_CLIENT_HANDSHAKE_RESPONSE_AND_BITFIELD) {
				l(L_DEBUG, "<$sock> -> Sending Handshake");
				_xwA($sock, Bitflu::Protocol::CreateHandshake({info_hash=>$tref->{INFO_HASH}, EXT_eproto=>1, EXT_fastpeers=>1,
				                                                  peer_id=>$config->{SYS}->{peer_id}}));
				$tref->{BTS} = BTS_WRITE_BITFIELD;
				$tref->{TIMEOUT} = $NOW+TIMEOUT_SHAKE;
			}
			elsif($state == BTS_WRITE_BITFIELD) {
				l(L_INFO, "<$sock> -> Writing Bitfield");
				_xwA($sock, Bitflu::Protocol::CreateBitfield($TORRENT_UNIVERSE->{$tref->{INFO_HASH}}->{bitfield}));
				$tref->{BTS} = BTS_WRITE_EXTENSIONS;
			}
			elsif($state == BTS_WRITE_EXTENSIONS) {
				l(L_INFO, "<$sock> -> Writing extensions (if needed) and entering IDLE state");
				$tref->{BTS} = BTS_IDLE;
				delete($tref->{TIMEOUT}); #The Handshake timeout is not checked in BTS_IDLE mode.. freeup some memroy..
				$tref->{SENT_KEEPALIVE} = $NOW + 5; #This is WRONG but there are some broken clients so we won't wait a full tick to send the keepalive
				$tref->{RECEIVED_KEEPALIVE} = $NOW + TIMEOUT_KEEPALIVE;  #Wait 'till X for a keepalive
				$tref->{TIMEOUT_LASTEVENT}  = $NOW + TIMEOUT_LASTEVENT;  # Times out if peer doesn't do any traffic (send  /  recv)
				# => TIMEOUT_WAITFORPIECE not set because PIECE_INWORK is set to UNDEF
				# => TIMEOUT_HUNTER       not set because hunt should start ASAP
				
				if($tref->{EXTENSIONS}->{FAST}) {
					foreach(keys(%{$tref->{FASTSET}})) {
						_xwA($sock,Bitflu::Protocol::CreateAllowedFast($_));
						l(L_DEBUG, "Sending FastAllowed : $_");
					}
				}
				if($tref->{EXTENSIONS}->{EPROTO}) {
					my $eref = { e=> 0, v=>'Bitflu', p=>$config->{client}->{port}, m=>{ut_pex=>EPROTO_UT_PEX, bf_speed=>EPROTO_BF_SPEED}};
					if($TORRENT_UNIVERSE->{$tref->{INFO_HASH}}->{content}->{info}->{private}) {
						# Private / Secure torrent => Disable PEX
						delete($eref->{m}->{ut_pex});
					}
					_xwA($sock,Bitflu::Protocol::CreateExtended(0,Bitflu::Bencoding::encode($eref)));
					l(L_DEBUG_EXT, "Sending ExtendedHandshake");
				}
				if($tref->{EXTENSIONS}->{DHT}) {
					#print "FIXME: THIS IS A DHT NODE\n";
				}
			}
			elsif($state == BTS_KILLME_PLEASE) {
				l(L_DEBUG, "<$sock> killing client as requested");
				killSock($torrent_handle,$sock);
				next;
			}
		}
		
		
		
		# -> WRITE data to wire
		my $buff = substr(_xGETtowrite($sock),0,$upcredits);
		my $bufflen = length($buff);
		
		if($tref->{OUTSTANDING_REQUESTS} && $bufflen < $upcredits && (my $orkref = shift(@{$tref->{OUTSTANDING_REQUESTS}}))) {
			my $fsockbuff = length(_xGETtowrite($sock));
			Carp::confess() unless defined $orkref;
			my($ork_data, $ork_len) = Bitflu::Store::ReadPiece($orkref);
			if($ork_len != $orkref->{size}) {
				l(L_WARN, "<$sock> Requested piece $orkref->{index} at offset $orkref->{offset} with a length of $orkref->{size}. Reading failed after $ork_len bytes!");
				$tref->{BTS} = BTS_KILLME_PLEASE;
			}
			else {
				#Padd buffer
				_xwA($sock,Bitflu::Protocol::CreatePiece($orkref).$ork_data);
				$TORRENT_UNIVERSE->{$tref->{INFO_HASH}}->{config}->{uploaded} += $ork_len;
				$buff = substr(_xGETtowrite($sock),0,$upcredits);
				$bufflen = length($buff);
			}
		}



		if($bufflen) {
			my $data_sent = syswrite($sock,$buff);
			$BWSTATS{UP}+=$data_sent;
			_xCLEANtowrite($sock,$data_sent);
			$tref->{STATS_UPLOADED}       += $data_sent;
			$tref->{STATS_UPLOADED_TOTAL} += $data_sent;
			if(!defined($data_sent)) {
				l(L_WARN, "<$sock> !! SENDING data failed with error $!");
				killSock($torrent_handle,$sock); #Close connection NOW
			}
		}
	}
}


sub CancelAllPiecesOfHash {
	my($pieces, $hash) = @_;
	foreach my $sock (keys(%TCSTAT)) {
		my $tref = $TCSTAT{$sock};
		next if $tref->{INFO_HASH} ne $hash;
		foreach my $to_cancel (keys(%{$tref->{PIECE_INWORK}})) {
			if(defined($pieces->{$to_cancel})) {
				_xwA($sock, Bitflu::Protocol::CreateCancel($tref->{PIECE_INWORK}{$to_cancel}));
				delete($tref->{PIECE_INWORK}{$to_cancel});
				Bitflu::Store::SetPieceAsFree($hash,$to_cancel);
				Carp::confess("FSBUG") unless Bitflu::Store::PieceIsFree($hash,$to_cancel);
			}
		}
	}
}






package Bitflu::HTTPgui;
use URI::Escape; # Imported in main:: but we would like to get the exported vars here :-)

sub main {
	my($rq) = @_;
	
	my @HTTP_HEADER = split(/\r\n/,$rq);
	my %header_hash = ();
	my $http_buff = undef;
	my($query_method, $uri) = $HTTP_HEADER[0] =~ /^(GET|HEAD) (\S+) HTTP\//i;
	
	my($rq_hash, $rq_cmd) = $uri =~ /^\/([^\/]+)\/(\S+)$/;

	for(my $i=1;$i<int(@HTTP_HEADER);$i++) {
		chomp($HTTP_HEADER[$i]);
		if($HTTP_HEADER[$i] =~ /^([^:]+):(\s?([^\r\n]+))?/) {
			$header_hash{$1} = $3;
		}
		else {
			last;
		}
	}

	
	unless($query_method) {
		return "HTTP/1.0 500 Invalid Request\r\n\r\nInvalid Request\r\n";
	}
	
	if($config->{client}->{userauth}) {
		# Check WWW-Auth
		my ($auth) = $header_hash{Authorization} =~ /Basic (.+)$/;
		my ($wwwuser,$wwwpass) = Bitflu::Utils::DecodeBase64(substr($auth,0,80)) =~ /([^:]+):(.+)$/; #Save mem.. save base64 decoder
		unless(Bitflu::Utils::UserPasswordIsOkay($wwwuser,$wwwpass)) {
			return "HTTP/1.0 401\r\nWWW-Authenticate: Basic realm=\"/\"\r\n\r\nForibdden\n";
		}
		# else : continue
	}
	
	
	
	$http_buff = "HTTP/1.0 200 OK\r\nServer: Bitflu\r\n";
	
	
	if($rq_hash eq "Images") {
		$http_buff .= "Content-Type: image/gif\r\n\r\n";
		$http_buff .= CreateGifImage(split(/_/,$rq_cmd));
		return $http_buff;
	}
	elsif($rq_hash eq "StyleSheet") {
		$http_buff .= "Content-Type: text/css\r\n\r\n";
		$http_buff .= CreateCSS();
		return $http_buff;
	}
	else {
		$http_buff .= "Content-Type: text/html\r\n\r\n";
	}
	# Normal HTTP-View
	
	
	
	$http_buff .= "<html><header><title>Bitflu $rq_hash</title>\n";
	$http_buff .= '<link rel="stylesheet" href="/StyleSheet/fark.css" type="text/css">';
	$http_buff .= "</header><body>\n<h3>Bitflu HTTP-Interface</h3>\n";
	
	$http_buff .= "<form method=get action='/Load/'><input type=text name=file><input type=submit value='Load torrent'></form>\n";
	
	my $injected = undef;
	
	if($rq_hash eq "Load") {
		$rq_cmd = substr($rq_cmd,1); #Remove ?
		my $toload   = undef;
		my $takethis = 0;
		foreach my $xpart (split(/&/,$rq_cmd)) {
			foreach my $xm (split(/=/,$xpart)) {
				next if defined($toload);
				$toload = uri_unescape($xm) if $takethis;
				$takethis=1 if $xm eq 'file';
			}
		}
		
		my (undef,$loadmsg) = Bitflu::Commands::LoadTorrent($toload);
		$http_buff .= BuildDialogWindow(_htmlIze($loadmsg));
	}
	elsif(!$TORRENT_UNIVERSE->{$rq_hash}) {
		$http_buff .= "<!-- No valid hash //-->\n";
	}
	elsif($rq_cmd eq "ClientStats") {
		$injected = BuildDialogWindow(BuildClientStats($rq_hash));
	}
	elsif($rq_cmd eq "Cancel") {
		$injected = BuildDialogWindow("Do you really want to cancel $rq_hash ?", [{ 'Yes' => "/$rq_hash/RealCancel"}, {'Abort' => "/$rq_hash/Selected"}]);
	}
	elsif($rq_cmd eq "RealCancel") {
		Bitflu::Commands::CancelTorrent($rq_hash);
		$http_buff .= "<script language=\"JavaScript\">this.location = '/';</script>";
	}
	elsif($rq_cmd eq "Commit") {
		$injected = BuildDialogWindow("Start commiting $rq_hash ?", [ { 'Start Commit' => "/$rq_hash/RealCommit"} , {'Abort' => "/$rq_hash/Selected"}]);
	}
	elsif($rq_cmd eq "RealCommit") {
		my(undef,$fork_out) = Bitflu::Commands::ForkCommit([$rq_hash], 1);
		$injected = BuildDialogWindow("$fork_out");
	}

	
	$http_buff .= BuildViewDownloads($rq_hash,$injected);
	
	$http_buff .= "</body></html>\n";
	return $http_buff;
}

sub BuildDialogWindow {
	my($txt, $aref) = @_;
	
	my $buff = "\n<div class='msgbox'><div>$txt</div>\n";
	
	my $buttons = undef;
	foreach my $ref (@$aref) {
		foreach my $k (keys(%$ref)) {
			$buttons .= "<a class='linkbox' href='$ref->{$k}'>$k</a>";
		}
	}
	
	$buff .= "<br>$buttons<br><font size-2> " if $buttons;
	
	$buff .= "</div>\n";
	return $buff;
}


##################################################################################################################################
# Show Commandlinks
sub BuildCommandList {
	my($hash) = @_;
	return BuildDialogWindow("<i>==&gt;$hash</i>", [ { "Commit" => "/$hash/Commit" },
	                                                 { "Cancel" => "/$hash/Cancel" },
	                                                 { "Show statistics" => "/$hash/ClientStats" },
	                                                 { "Close" => "/" } ]);
}

##################################################################################################################################
# Shows verbose information about a torrent
sub BuildClientStats {
	my($hash) = @_;
	
	my $http_buff = "<table border=1><tr><td>PeerID</td><td>IP</td><td>MeChoked</td><td>PeerChoked</td><td>Downloading</td><td>Uploaded</td></tr>\n";
	
	foreach my $cs (keys(%TCSTAT)) {
		next if lc($TCSTAT{$cs}->{INFO_HASH}) ne lc($hash);
		my $px = undef;
		foreach(keys(%{$TCSTAT{$cs}->{PIECE_INWORK}})) { $px .= "$_;"; }
		$http_buff .= "<tr>";
		$http_buff .= "<td>"._htmlIze(_prfy($TCSTAT{$cs}->{CLIENT_ID}))."</td>";
		$http_buff .= "<td>".$TCSTAT{$cs}->{IP}."</td>";
		$http_buff .= "<td>".$TCSTAT{$cs}->{ME_CHOKE}."</td>";
		$http_buff .= "<td>".$TCSTAT{$cs}->{PEER_CHOKE}."</td>";
		$http_buff .= "<td>$px</td>";
		$http_buff .= "<td>".$TCSTAT{$cs}->{STATS_UPLOADED_TOTAL}." X</td>";
		$http_buff .= "</tr>\n";
	}
	$http_buff .= "</table>\n";
	return $http_buff;
}


##################################################################################################################################
# Displays an HTML-Version of 'vd'
sub BuildViewDownloads {
	my($injectAt, $injected) = @_;
	my $http_buff = undef;
	$http_buff .= "<div class='xdesc'>Upload: ".sprintf("%.3f",$BWSTATS{UPX}/1024)." KiB/s &nbsp;&nbsp; Download: ".sprintf("%.3f",$BWSTATS{DOWNX}/1024)." KiB/s</div>\n<table style='border:thin solid black; border-spacing:0px'>\n";
	$http_buff .= "<tr><td>Name</td><td>Hash</td><td>Peers (Active/Total)</td><td>Pieces (Done/Total)</td><td>Size (Done/Total)</td><td>Done %</td></tr>\n";
			my %list = ();
			
			foreach my $s (keys(%TCSTAT)) {
				my $ih = $TCSTAT{$s}->{INFO_HASH};
				my $mc = $TCSTAT{$s}->{ME_CHOKE};
				$list{$ih}->{peers}++;
				$list{$ih}->{active}++ if $mc == 0;
				foreach(keys(%{$TCSTAT{$s}->{PIECE_INWORK}})) {
					$list{$ih}->{sucking}++;
				}
			}
	my $dlcount = 0;
	foreach my $k (sort keys(%{$TORRENT_UNIVERSE})) {
		my $donestats = Bitflu::Store::GetDoneStats($k);
		my $blunt = "table1";
		   $blunt = "table2" if (++$dlcount%2);
		   $blunt = "selected" if lc($k) eq lc($injectAt);
		my $torrentstate = 'idle';
		   $torrentstate = 'downloading' if $list{$k}->{sucking} > 0;
		   $torrentstate = 'stalled'     if $list{$k}->{peers}  < 1;
		   $torrentstate = 'completed'   if $donestats->{pieces} == int(Bitflu::Store::GetNumberOfPieces($k)+1);
		
		my $percent_done = sprintf("%6.2f",(($donestats->{pieces}/(Bitflu::Store::GetNumberOfPieces($k)+1))*100));
		
		$http_buff .= "<tr class=$blunt>\n";
		$http_buff .= "<td class='$torrentstate'><b>"._htmlIze(substr(${$TORRENT_UNIVERSE->{$k}->{torrent}}->{content}->{info}->{name},0,96))."</b></td>";
		$http_buff .= "<td><a href='/$k/Selected'>$k</a></td>";
		$http_buff .= "<td>".int($list{$k}->{active})."/".int($list{$k}->{peers})."</td>";
		$http_buff .= "<td>".int($donestats->{pieces})."/".int(Bitflu::Store::GetNumberOfPieces($k)+1)."</td>";
		$http_buff .= "<td>".int($donestats->{size}/1024/1024)."/".int($TORRENT_UNIVERSE->{$k}->{total_size}/1024/1024)."</td>";
		$http_buff .= "<td><img src='/Images/80_145_80' height=8 width=".int($percent_done)." alt='$percent_done% completed'>";
		$http_buff .= "<img src='/Images/145_95_80' height=8 width=".int(100-$percent_done)." alt='$percent_done% completed'></td>";
		$http_buff .= "</tr>\n";
		
		if(lc($k) eq lc($injectAt)) {
			$injected ||= BuildCommandList($k);
			$http_buff .= "<tr class=$blunt><td>$injected</td><td></td><td></td><td></td><td></td><td></td></tr>";
		}
		
	}
	$http_buff .= "</table>\n";
	
	
	
return $http_buff;
}


sub CreateGifImage {
	my($red, $green, $blue) = @_;
	my $buff =  pack("H*", "47494638376101000100800100");
	   $buff .= pack("c",$red);
	   $buff .= pack("c",$green);
	   $buff .= pack("c",$blue);
	   $buff .= pack("H*", "ffffff2c00000000010001000002024401003b");
	return $buff;
}

sub CreateCSS {
	my $buff = << "EOF";
html {
    color: #000000; 
    background: #fefefe;
    font-family: serif;
}
h1, h2, h3 {
    text-align: left; 
    font-family: sans-serif;
}
h1 {
    font-size: 150%;
    line-height: 150%;
}
h2 {
    font-size: 130%;
}
h3 {
    font-size: 100%;
}
.selected {
	background: #b5c0cd;
}
.completed {
	color: #969145;
}
.stalled {
	color: #913737;
}
.downloading {
	color: #497c47;
}
.idle {
	color: #2a2a35;
}
.table1 {
	background: #e8e8e8;
}
.table2 {
	background: #f0f0f0;
}
.msgbox {
	background: #dedede;
	padding: 4px;
	border-style: solid;
	border-width: 2px;
	border-color: #c6c3bd;
	text-align: center;
}

.linkbox {
	background: #dfdcdf;
	border-style: solid;
	border-width: 1px;
	margin-right: 4px;
	padding: 4px;
	border-color: #000000;
}
.linkbox:hover {
	background: #e3dfc3;
}
a:link, a:visited {
    color: #1b56ce;
    font-weight: bold;
}
a:link:hover, a:visited:hover {
    color: #d81e1e;
}
EOF
return $buff
}

sub _htmlIze {
	my($string) = @_;
	$string =~ s/&/&nbsp;/gm;
	$string =~ s/>/&gt;/gm;
	$string =~ s/</&lt;/gm;
	return $string;
}

sub _prfy {
	my($string) = @_;
	$string =~ tr/[a-z][A-Z][0-9]_\.& -//cd;
	return $string;
}

1;





##################################################################################################################################
package Bitflu::Publisher;

	my $UGLY_FILE_FIND_REF = ();
	use constant FAKESHA => 'X' x 20;
	use constant FAKESUM => FAKESHA.FAKESHA;
	
	
	#################################################################
	# Ehr.. main()
	# returns (torrent_hash, human_message) ; torrent_has is undef on error
	sub main {
		my(@cmdlist) = @_;
		
		my $fakesha = FAKESHA;
		my $fakesum = FAKESUM;
		my $ref = {};           # Dull
		my $torrent = {};       # Torrent Reference
		my $fatalerr = undef;
		
		Carp::confess("BUGBUG: fakesum defined while entering main?! someone forget to cleanup stuff ; panic!") if $TORRENT_UNIVERSE->{$fakesum};
		
		if(-d $fakesum) {
			return(undef, "Directory '$fakesum' exists , please remove it before running publish");
		}
		
		# 'Parse' arguments
		foreach(@cmdlist) {
			my($k,$v) = split(/=/,$_);
			push(@{$ref->{$k}},$v);
		}
		
		
		$TORRENT_UNIVERSE->{$fakesum}->{torrent} = \$torrent; # Glue $torrent <-> $TORRENT_UNIVERSE
		# ..and pollute our $torrent reference
		$torrent->{content}->{'creation date'} = int(time());                        # Time of creation
		$torrent->{content}->{'created by'}    = "BitFlu ".main::BITFLU_VERSION;     # Watchout for broken torrents! :-)
		$torrent->{content}->{info}->{pieces}  = $fakesha;                           # A fake Checksum ; needed to get INIT() happy
		$torrent->{content}->{info}->{name}    = ${$ref->{name}}[0];                 # Name
		$torrent->{content}->{info}->{'piece length'} = 512*1024;                    # Currently hardcoded to 512kib
		
		# -> Create multiTracker entry
		# See http://wiki.depthstrike.com/index.php/P2P:Protocol:Specifications:Multitracker
		my $trackers_count = 0;
		foreach(@{$ref->{announce}}) {
			my @multitrackers  = ();
			foreach(split(/#/,$_)) {
				$torrent->{content}->{'announce'} ||= $_;
				push(@multitrackers,$_);
				$trackers_count++;
			}
			push(@{$torrent->{content}->{'announce-list'}}, \@multitrackers);
		}
		# Hups: No backups? Well: Delete useless announce-list
		delete($torrent->{content}->{'announce-list'}) if $trackers_count < 2;
		delete($torrent->{content}->{announce})        unless $ref->{announce}; # -> trackerless torrent
		
		# Add DHT nodes: NOTE: Bitflu supports IPv6 DHT-Nodes ; Silly Mainline doesn't and may
		# crash if it eats a .torrent created by bitflu with IPv6 nodes!
		foreach(@{$ref->{nodes}}) {
			if($_ =~ /^([0-9a-fA-F\.:]+):(\d+)$/) {
				push(@{$torrent->{content}->{nodes}}, [$1, $2]);
			}
			else {
				$fatalerr .= "Invalid DHT-Node '$_' ";
			}
		}
		
		
		# Check if the Torrent is 'bootable' . Well: Bitflu could use
		# a torrent without announce and nodes without any problems.
		# But this check avoids the distribution of non-mainline-compatible torrents.
		if(!$torrent->{content}->{announce} && !$torrent->{content}->{nodes}) {
			$fatalerr .= "No 'announce' and/or 'nodes' keyword ; ";
		}
		
		#Walk directory tree and find all files
		$UGLY_FILE_FIND_REF = (); #Clean it
		File::Find::find( {wanted => \&HandlePublishItem, no_chdir=>1}, ( Bitflu::Store::__get_publishbase()) );

		if(ref($UGLY_FILE_FIND_REF->{files}) ne "ARRAY") {
			$fatalerr .= "No valid files inside 'publish' found ; ";
		}
		
		
		# Last stop before we start to publish the file: Check if someone
		# has set fatalerr:
		if(defined($fatalerr)) {
			delete($TORRENT_UNIVERSE->{$fakesum});
			return (undef, "$fatalerr  Nothing published. See 'help publish' for more information");
		}
		
		
		# Init TorrentStorage for our fake-torrent..
		Bitflu::Store::InitTorrent($fakesum) or Carp::confess("InitTorrent aborted");

		# ..and slap it together
		my $slapt_result = slapit($UGLY_FILE_FIND_REF,$torrent);
		# What we have done so far:
		# We got a working copy of the content inside workdir/$fakesha
		# We got a messy mess inside $TORRENT_UNIVERSE->{$fakesum}
		# We got a bencoded torrent inside $encoded_torrent
		
		# No. 0 : Get rid of the $TORRENT_UNIVERSE pollution
		delete($TORRENT_UNIVERSE->{$fakesum});
		
		# No. 0.5 : Check if we could publish it..
		if(defined($slapt_result)) {
			return (undef,"Failed to publish content: '$slapt_result'");
		}
		
		# No. 1: Fixup our torrent hashref and decode the torrent
		$torrent->{content}->{info}->{files} = $UGLY_FILE_FIND_REF->{files};
		
		# .. and add a NULLBYTE to each path element
		# DO NOT USE THIS LIST ANYMORE WITH SYSCALLS!
		foreach(@{$torrent->{content}->{info}->{files}}) {
			my $uref = $_;
			foreach(@{$uref->{path}}) {
				$_ .= "_" if $_ =~ /^\d+$/; #Ugly hack: Add a something to num-only entries to force a string directory.
			}
		}
		
		# Check for single-file torrent:
		if(int(@{$torrent->{content}->{info}->{files}}) == 1) {
			# Wowies! Single file torrent.. follow the stupidity and
			# shift some stuff around
			$torrent->{content}->{info}->{name}   ||= ${${$torrent->{content}->{info}->{files}}[0]->{path}}[0];
			$torrent->{content}->{info}->{length} = ${$torrent->{content}->{info}->{files}}[0]->{length};
			delete($torrent->{content}->{info}->{files});
		}
		
		$torrent->{content}->{info}->{name} ||= "Unnamed torrent\@".int(time()); #Fallback name
		
		# No. 2 : Mumble $torrent to get correct values from bencode() :-(
		$torrent->{content}->{info}->{name} .= "_" if $torrent->{content}->{info}->{name} =~ /^\d+$/; #Ugly hack: Add something to get a string
		
		# No. 3 : Save the torrent somewhere
		my $encoded_torrent = Bitflu::Bencoding::encode($torrent->{content});
		my $savefile = Bitflu::Store::__get_torrentbase()."/publishTorrent_$$"; # We are placing it there to avoid crashes at startup if thisone is corrupt
		open(PUBLISHED, ">$savefile") or Carp::confess("Unable to write $savefile : $!\n");
			print PUBLISHED $encoded_torrent;
		close(PUBLISHED);
		
		
		# No. 4 : TestLoad it to get a torrent_hash
		my $torrent_hash = Bitflu::Bencoding::torrent2hash($savefile)->{torrent_hash};
		   $torrent_hash or Carp::confess("PANIC: TestLoading torrent failed?!");
		
		if($TORRENT_UNIVERSE->{$torrent_hash}) {
			unlink($savefile) or Carp::confess("Unable to unlink $savefile : $!");
			return (undef, "$torrent_hash does exist ; unable to re-publish same SHA1 hash");
		}
		
		# No. 5 : Rename the old dir
		$TORRENT_UNIVERSE->{$fakesum} = $TORRENT_UNIVERSE->{$torrent_hash} = "FAKE";
		rename(Bitflu::Store::_getbase($fakesum), 
		       Bitflu::Store::_getbase($torrent_hash)) or Carp::confess("Fatal: rename($fakesum,$torrent_hash) failed: $!");
		
		delete($TORRENT_UNIVERSE->{$fakesum});
		delete($TORRENT_UNIVERSE->{$torrent_hash});
		
		my $load_result = main::__load_torrent($savefile);
		unlink($savefile) or Carp::confess("Unable to unlink $savefile : $!\n");
		
		unless($load_result) {
			Carp::confess("Unable to load torrent created a few seconds ago?!");
		}
		else {
			#Fixup config
			$TORRENT_UNIVERSE->{$torrent_hash}->{config}->{downloaded} = $UGLY_FILE_FIND_REF->{size};
			$TORRENT_UNIVERSE->{$torrent_hash}->{config}->{uploaded}   = $UGLY_FILE_FIND_REF->{size};
			return (defined, "Content of ".Bitflu::Store::__get_publishbase()." has been moved into torrent $torrent_hash and is shared");
		}
	}
	

	
	
	
	
	
	sub HandlePublishItem {
		my $tocrop = Bitflu::Store::__get_publishbase();
		my $item = $_;
		#print ">CALL $File::Find::dir | $_ | $File::Find::name\n";
		if(-f $item) {
			my @patharray = split(/\//,__torrentpath($item,$tocrop));
			my $size      = int(-s $item);
			if($size == 0) {
				main::l(main::L_WARN, "[Publish] Empty file '@patharray' skipped"); #Fixme: We can't handle empty files.. slapit chokes on it
				return undef;
			}
			$UGLY_FILE_FIND_REF->{size} += $size;
			push(@{$UGLY_FILE_FIND_REF->{files}}, { length=>$size, path=>\@patharray });
		}
	}
	
	#################################################################
	# Splitup and SHA-Hash content inside 'publish'-dir
	# Note: This sub changes the $torrent->{content}->{info}->{pieces} ref.
	sub slapit {
		my($findref,$torrent) = @_;
		
		my $slapbytes_left = $torrent->{content}->{info}->{'piece length'} or Carp::confess("Somehow we got no piece-length?!");
		my $slappiece      = 0;      #First piece
		my $slapbuff       = undef;  #Buffer for read data
		my $slapbuffconcat = undef;  #Buffer for a full piece
		my $slapsha        = undef;  #Buffer for SHA-Sums
		my $done_size      = 0;      #How much we had read
		my $total_size     = $findref->{size}; #..how much we need to read
		
		foreach(@{$findref->{files}}) {
			my $cref = $_;
			my $unixpath = Bitflu::Store::__get_publishbase().'/'.join('/',@{$cref->{path}});
			open(SLAP, $unixpath) or return "Unable to open $unixpath : $!";
				while($slapbytes_left) {
					my $bytes_read = read(SLAP,$slapbuff,$slapbytes_left);
					Carp::confess("FATAL: Reading from $unixpath failed: '$!'") unless defined $bytes_read;
					$slapbuffconcat .= $slapbuff;   #Concat read stuff to piecebuff
					$slapbytes_left -= $bytes_read; #..and dec. read bytes
					$done_size      += $bytes_read; #..and inc. total read count
					main::l(main::L_DEBUG_PBLSH, "READ: $bytes_read from $unixpath :: left $slapbytes_left ($done_size -> $total_size)");
					
					if(($slapbytes_left == 0 or $done_size == $total_size) and length($slapbuffconcat) != 0) {
						my $filepath      = Bitflu::Store::_getfreedir(FAKESUM)."/".int($slappiece);
						my $filepath_done = Bitflu::Store::_getdonedir(FAKESUM)."/".int($slappiece);
						open(PIECE, ">$filepath") or return "Unable to write to $filepath : $!";
							print PIECE $slapbuffconcat;
						close(PIECE);
						
						$slapsha .= Digest::SHA1::sha1($slapbuffconcat);
						rename($filepath,$filepath_done) or return "Unable to rename $filepath into $filepath_done : $!";
						main::l(main::L_WARN, "[Publish]: Piece $slappiece ($unixpath) written & hashed...");
						#Clean/Fix slapt's
						$slapbuffconcat = undef;
						$slapbytes_left = $torrent->{content}->{info}->{'piece length'};
						$slappiece++;
					}
					last if $bytes_read < 1; #EOF ; jump to next piece
				}
			close(SLAP);
		}
		$torrent->{content}->{info}->{pieces} = $slapsha; #Fixup fake-checksum
		return undef;
	}
	
	
	#################################################################
	# Crop's away the local path
	sub __torrentpath {
		my($value,$tocrop) = @_;
		my($out) = $value =~ /^$tocrop\/?(.*)$/;
		return $out;
	}
	
	
1;






package Bitflu::Commands;

	########################################################################
	# Disconnect client
	sub Quit {
		my($h,$s) = @_;
		main::killSock($h,$s);
	}
	
	########################################################################
	# Returns an array-ref with all logins
	sub ListUsers {
		my @userlist = ();
		foreach my $chunk (split(/;/,$config->{client}->{userauth})) {
			my($cuser,$cpass) = $chunk =~ /^([^:]+):(\S+)$/;
			push(@userlist,$cuser);
		}
		return(defined,\@userlist);
	}
	
	########################################################################
	# Changes password of given user
	sub ChangeUserPassword {
		my($user,$old_password, $new_password) = @_;
		
		$user =~ tr/A-Za-z0-9_\.//cd;
		
		unless(Bitflu::Utils::UserPasswordIsOkay($user,$old_password)) {
			return(undef, "Password for '$user' not changed: old password is incorrect");
		}
		
		
		my $newstring = undef;
		foreach my $chunk (split(/;/,$config->{client}->{userauth})) {
			my($cuser,$cpass) = $chunk =~ /^([^:]+):(\S+)$/;
			if($cuser eq $user) {
				$newstring .= "$cuser:".Digest::SHA1::sha1_hex("$user:$new_password").";";
			}
			else {
				$newstring .= "$chunk;";
			}
		}
		
		chop($newstring);
		$config->{client}->{userauth} = $newstring;
		return(defined, "Password for '$user' has been changed, use 'config save' to make changes permanent");
	}
	
	########################################################################
	# Creates a new user
	sub AddNewUser {
		my($user, $password) = @_;
		
		$user =~ tr/A-Za-z0-9_\.//cd;
		
		my $user_did_exist = undef;
		
		foreach my $chunk (split(/;/,$config->{client}->{userauth})) {
			my($cuser,$cpass) = $chunk =~ /^([^:]+):(\S+)$/;
			if($cuser eq $user) { $user_did_exist = 1; last;}
		}
		
		return(undef, "User '$user' does already exist") if $user_did_exist;
		
		$config->{client}->{userauth} .= ";" if $config->{client}->{userauth};
		$config->{client}->{userauth} .= "$user:".Digest::SHA1::sha1_hex("$user:$password");
		return(defined, "User '$user' created , use 'config save' to make changes permanent");
	}
	
	########################################################################
	# Removes an user
	sub DeleteUser {
		my($user,$password) = @_;
		
		$user =~ tr/A-Za-z0-9_\.//cd;
		
		unless(Bitflu::Utils::UserPasswordIsOkay($user,$password)) {
			return(undef, "User '$user' not removed: incorrect password");
		}
		
		my $newstring = undef;
		foreach my $chunk (split(/;/,$config->{client}->{userauth})) {
			my($cuser,$cpass) = $chunk =~ /^([^:]+):(\S+)$/;
			next if $cuser eq $user;
			$newstring .= "$chunk;";
		}
		chop($newstring);
		$config->{client}->{userauth} = $newstring;
		return(defined, "User removed, use 'config save' to make changes permanent");
	}
	
	########################################################################
	# Set PAUSE flag
	sub Pause {
		my($hash) = @_;
		return(undef, "unable to pause $hash : hash does not exist in queue") unless $TORRENT_UNIVERSE->{$hash};
		$TORRENT_UNIVERSE->{$hash}->{config}->{paused} = 1;
		return(defined, "$hash has been paused");
	}
	
	sub AutoCommit {
		my(@args) = @_;
		
		my $hash   = $args[0];
		my $action = $args[1];
		my $ration = sprintf("%.2f",$args[2]);
		
		return(undef, "unable to set autocommit flag for $hash : hash does not exist in queue") unless $TORRENT_UNIVERSE->{$hash};
		
		
		if($action eq "off") {
			$TORRENT_UNIVERSE->{$hash}->{config}->{autocommit} = 0;
			return(defined, "$hash: Autocommit disabled");
		}
		elsif($action eq "on") {
			if($ration >= main::MIN_ACRATION) {
				$TORRENT_UNIVERSE->{$hash}->{config}->{autocommit} = $ration;
				return(defined, "$hash: Autocommit starts at ration $ration");
			}
			else {
				return(undef, "$hash: Ration '$ration' too low ; please specify a ration > 1.0");
			}
		}
		elsif($action eq "status") {
			return(defined, "Autocommit for $hash is not enabled") unless $TORRENT_UNIVERSE->{$hash}->{config}->{autocommit};
			return(defined, "Autocommit starts after reaching a ration of ".$TORRENT_UNIVERSE->{$hash}->{config}->{autocommit});
		}
		return(undef, "autocommit: unknown subcommand '$action' ; see 'help autocommit' for more information");
	}
	
	########################################################################
	# Remove PAUSE flag
	sub Resume {
		my($hash) = @_;
		return(undef, "unable to resume $hash : hash does not exist in queue") unless $TORRENT_UNIVERSE->{$hash};
		$TORRENT_UNIVERSE->{$hash}->{config}->{paused} = 0;
		return(defined, "$hash has been resumed");
	}
	
	########################################################################
	# Dump all bitmasks/bitfields
	sub GlobalBitmask {
		my $buff = "--- Bitfields ---\n";
		foreach(keys(%{$TORRENT_UNIVERSE})) {
			$buff .= "Bitfield of $_\n----------------------------\n".unpack("B*", $TORRENT_UNIVERSE->{$_}->{bitfield})."\n\n";
		}
		return(defined,$buff);
	}
	
	########################################################################
	# Return version and uptime
	sub Version {
		return (defined, "This is BitFlu ".main::BITFLU_VERSION." (".main::BITFLU_BUILD."); Running since ".int( (int(time())-($config->{SYS}->{boottime})) / 60 )." min. as '$0' with pid $$");
	}
	
	########################################################################
	# Shutdown bitflu
	sub ShutdownProcess {
		kill(2,$$);
	}
	
	########################################################################
	# Printout torrent-hash
	sub AnalyzeTorrent {
		my($file) = @_;
		my $ref = Bitflu::Bencoding::torrent2hash($file);
		return (undef, "Unable to analyze $file") unless defined $ref->{content};
		delete $ref->{content}->{info}->{pieces}; #Do not return binary trash
		return (defined,$ref);
	}
	
	########################################################################
	# Verify Storage Directory
	sub VerifyTorrent {
		my($hash) = @_;
		return(undef, "unable to verify $hash : hash does not exist in queue") unless $TORRENT_UNIVERSE->{$hash};
		Bitflu::Store::VerifyStorageDir($hash) or main::l(main::L_PANIC, "Please restart $0 to rebuild the Bitfield of $hash");
		return(defined, "$hash passed verification");
	}
	
	########################################################################
	# Dump client stats for Telnet view
	sub TelnetClientStatsList {
		my($match) = @_;
		my $c = 0;
		my $buff = " Client ID            # Torrent Info Hash                        |STATE| CCII | DOWNLOADING        |PIECES| UPLOADED |FDX|   PEER-IP            |";
		$buff .= " INTR  | HUNT\n";
		foreach my $cc (keys(%TCSTAT)) {
			next unless $TCSTAT{$cc}->{INFO_HASH} =~ /$match/i;
			$c++;
			my $client = $TCSTAT{$cc}->{CLIENT_ID};
			$client =~ tr/a-zA-Z0-9:\.-//cd;
			my $px = "";
			foreach(keys(%{$TCSTAT{$cc}->{PIECE_INWORK}})) { $px .= "$_;"; }
			
			my $xcpieces = unpack("B*", $TCSTAT{$cc}->{BITFIELD});
			   $xcpieces =~ tr/1//cd; $xcpieces = length($xcpieces);
			$buff .= (sprintf(">%-20s # %-40s | %03d | %d%d%d%d | %-18s | %04d | %08d |%d%d%d| %-20s | %-5s | %-5s\n", substr($client,0,20),
			$TCSTAT{$cc}->{INFO_HASH}, $TCSTAT{$cc}->{BTS}, $TCSTAT{$cc}->{ME_CHOKE}, $TCSTAT{$cc}->{PEER_CHOKE},
			$TCSTAT{$cc}->{ME_INTERESTED},$TCSTAT{$cc}->{PEER_INTERESTED},
			$px, $xcpieces, $TCSTAT{$cc}->{STATS_UPLOADED_TOTAL}, $TCSTAT{$cc}->{EXTENSIONS}->{FAST},$TCSTAT{$cc}->{EXTENSIONS}->{DHT},$TCSTAT{$cc}->{EXTENSIONS}->{EPROTO},
			"$TCSTAT{$cc}->{IP}:$TCSTAT{$cc}->{PORT}",
			_dv($TCSTAT{$cc}->{TIMEOUT_INTERESTED}), _dv($TCSTAT{$cc}->{TIMEOUT_HUNTER})));
		}
		return(defined, $buff."\n -> $c clients out of ".main::GetMaxConnections());
	}
	
	########################################################################
	# Dump download list for telnet view
	sub TelnetDownloadList {
		my %list = ();
		
		foreach my $s (keys(%TCSTAT)) {
			my $ih = $TCSTAT{$s}->{INFO_HASH};
			$list{$ih}->{peers}++;
			$list{$ih}->{active}++ if $TCSTAT{$s}->{ME_CHOKE} == 0;
			foreach(keys(%{$TCSTAT{$s}->{PIECE_INWORK}})) {
				$list{$ih}->{sucking}++;
			}
		}
		my $buff = (sprintf(" *** UPLOAD ($config->{SYS}->{runtime_unchoked}) : %.3f KiB/s | DOWNLOAD %.3f KiB/s\n", $BWSTATS{UPX}/1024 , $BWSTATS{DOWNX}/1024));
		$buff   .= (sprintf("> %-32s |%-40s| || %-9s | %-9s| SIZE (DONE/TOTAL) | DONE % | RATION | BW: D/U   | \n",
		                     'Name', '================= Hash =================', 'A/T>PEERS', 'DONE/TOTAL'));
		foreach my $k (sort keys(%{$TORRENT_UNIVERSE})) {
			my $donestats = Bitflu::Store::GetDoneStats($k);
			my $flags     = "   ";
			   substr($flags,0,1,"C") if $TORRENT_UNIVERSE->{$k}->{is_complete};
			   substr($flags,1,1,"P") if $TORRENT_UNIVERSE->{$k}->{config}->{paused};
			   substr($flags,2,1,"A") if $TORRENT_UNIVERSE->{$k}->{config}->{autocommit};
			
			my $t_buff .= sprintf("* %-32s (%-40s) || %4d/%4d | %4d/%4d | %5d/%5d (MB)  | %6.2f | %6.2f |%5.1f %4.1f | %s\n", 
						substr(${$TORRENT_UNIVERSE->{$k}->{torrent}}->{content}->{info}->{name},0,32),
						$k, $list{$k}->{active},$list{$k}->{peers}, $donestats->{pieces}, Bitflu::Store::GetNumberOfPieces($k)+1,
						int($donestats->{size}/1024/1024), int($TORRENT_UNIVERSE->{$k}->{total_size}/1024/1024),
						(($donestats->{pieces}/(Bitflu::Store::GetNumberOfPieces($k)+1))*100),
						( Bitflu::Utils::GetRation($k) ),
						$TORRENT_UNIVERSE->{$k}->{stats_downloaded} , $TORRENT_UNIVERSE->{$k}->{stats_uploaded},
						$flags
						);
			
			if($donestats->{pieces} == Bitflu::Store::GetNumberOfPieces($k)+1) {
				$buff .= Bitflu::Ansi::Bold($t_buff); #Completed -> YellowBold
			}
			elsif($list{$k}->{sucking} != 0) {
				$buff .= Bitflu::Ansi::Green($t_buff); #Downloading -> Green
			}
			elsif($list{$k}->{peers} == 0) {
				$buff .= Bitflu::Ansi::Red($t_buff); #No peers -> Red
			}
			else {
				$buff .= $t_buff;
			}
		}
		chop($buff); # Removes last \n
		return(defined,$buff);
	}
	
	########################################################################
	# Load a torrent file
	sub LoadTorrent {
		my($path) = @_;
		
		my $torrent_you_shall_load = $path;
		my $torrent_you_shall_delete = undef;
		
		if($path =~ /^http:\/\//i) {
			my $uriget = Bitflu::Tracker::doHTTP({full_uri=>$path});
			if($uriget->{CODE} != 200) {
				return(undef, "unable to load $path ; Server responded with code $uriget->{CODE}");
			}
			else {
				my $savefile = Bitflu::Store::__get_torrentbase()."/httpLoad";
				open(HTTPLOAD, ">".$savefile) or Carp::confess("Unable to write $savefile : $!");
				print HTTPLOAD $uriget->{BODY};
				close(HTTPLOAD);
				$torrent_you_shall_load = $torrent_you_shall_delete = $savefile;
			}
		}
		
		# Parse file and return sha1-hash
		my $xhash  = Bitflu::Bencoding::torrent2hash($torrent_you_shall_load)->{torrent_hash};
		my $errmsg = undef;
		
		if(defined($xhash)) {
			my $xhist = Bitflu::Store::HistoryCheck($xhash);
			if($xhist->{exists}) {
				$errmsg = "$path has been added for download at ".gmtime($xhist->{addtime})." ; remove $xhist->{path} to allow a re-download";
			}
			else {
				unless(defined(main::__load_torrent($torrent_you_shall_load))) {
					$errmsg = "unable to load $path";
				}
			}
		}
		else {
			$errmsg = "$path doesn't seem to be a valid .torrent file";
		}
		
		
		if(defined($torrent_you_shall_delete)) {
			unlink($torrent_you_shall_delete) or Carp::confess("Unable to delete $torrent_you_shall_delete : $!");
		}
		
		if(defined($errmsg)) {
			return(undef,$errmsg);
		}
		else {
			return(defined,"$path loaded with hash $xhash");
		}
	}
	
	########################################################################
	# Publish
	sub Publish {
		my (@args) = @_;
		return Bitflu::Publisher::main(@args);
	}
	
	########################################################################
	# Force autoload
	sub AutoloadNow {
		if($config->{client}->{kick_autoload} == 0) {
			return(undef, "autoloading disabled; reenable it using 'config set kick_autoload \$something_above_zero'");
		}
		else {
			$config->{SYS}->{kick_autoload} = 0;
			return(defined, "autoload started");
		}
	}
	
	########################################################################
	# Remove a download
	sub CancelTorrent {
		my($hash) = @_;
		return(undef, "unable to cancel $hash : hash does not exist in queue") unless $TORRENT_UNIVERSE->{$hash};
		my $cdir = Bitflu::Utils::RemoveDownloadFromProcess($hash, Bitflu::Store::__get_canceledbase());
		return(defined, "$hash has been canceled and renamed into $cdir");
	}
	
	########################################################################
	# Remove a commited download
	sub CancelAndLinkTorrent {
		my($hash) = @_;
		return(undef, "unable to cancel $hash : hash does not exist in queue") unless $TORRENT_UNIVERSE->{$hash};
		my $cdir = Bitflu::Utils::RemoveDownloadFromProcess($hash, Bitflu::Store::__get_commitedbase());
		return(defined, "$hash has been moved into $cdir");
	}
	
	########################################################################
	# spam all connected telnet clients
	sub Wall {
		my($rawline_part, $exclude) = @_;
		my ($wmsg) = $rawline_part =~ /^\S+\s(.+)$/;
		my @xtelnet =$telnet_handle->can_write(main::DYNASLEEP_MIN);
		foreach (@xtelnet) {
			next if $_ eq $exclude;
			main::_xwA($_, "[Bitflu message] : $wmsg\n# ");
		}
	}
	
	sub ForkCommit {
		my($args, $do_remove, $partlist) = @_;

		my $pid = fork();
		if(defined($pid) && $pid == 0) {
			#-> CHILD
			$SIG{INT} = $SIG{HUP} = undef; #Do NOT save config on HUP
			# CloseDown all connections
			my @xtelnet =$telnet_handle->can_write(main::DYNASLEEP_MIN);
			my @xtorrent=$torrent_handle->can_write(main::DYNASLEEP_MIN);
			my @xhttp   =$http_handle->can_write(main::DYNASLEEP_MIN);
			#attemp to clean all connections held on this fork
			foreach(@xtelnet)  { close($_) if $_; }
			foreach(@xtorrent) { close($_) if $_; }
			foreach(@xhttp)    { close($_) if $_; }
			
			close($telnet_socket) if $telnet_socket;
			close($torrent_socket) if $torrent_socket;
			close($http_socket) if $http_socket;
			
			my $process_name = $0;
			foreach my $xhash (@$args) {
				$0 = "$process_name [commit:$xhash]";
				Commit($xhash,$do_remove, $partlist);
			}
			exit;
		}
		else {
			return(defined, "Started commiting process with pid $pid");
		}
	}
	
	########################################################################
	# Commit
	sub Commit {
		my($hash, $do_remove, $partlist) = @_;
		
		if(!$TORRENT_UNIVERSE->{$hash}) {
			Bitflu::RPC::wall("$hash: unable to commit ; hash does not exist");
			return;
		}
		else {
			Bitflu::RPC::wall("$hash: Commit started with pid $$");
		}
		my $assemble_start = time();
		my $assemble_result = Bitflu::Store::AssembleData($hash, $partlist);
		my $assemble_end   = time();
		Bitflu::RPC::wall("$hash: Commit finished after ".int($assemble_end-$assemble_start)." seconds");
		
		if($assemble_result) {
			Bitflu::RPC::wall("$hash: Commit errors: $assemble_result");
		}
		elsif($do_remove) {
			Bitflu::RPC::cancel_and_link("$hash");
			Bitflu::RPC::wall("$hash: Removed from queue");
		}
		else {
			Bitflu::RPC::wall("$hash: Commit completed ; download not removed from queue");
		}
		
	}
	
	
	sub Help {
		my(@subpages) = @_;
		
		my @buff = ();
		
		push @buff, "load /tmp/x.torrent             - Start download of /tmp/x.torrent";
		push @buff, "load http://example.com/..      - Grab .torrent and start download";
		push @buff, "commit \$hash                    - Assemble download and remove file from queue";
		push @buff, "xcommit \$hash                   - Same as commit but does not stop seeding";
		push @buff, "pcommit \$hash 3 6 8             - Commit file 3, 6 and 8 of \$hash (-> see 'multifile \$hash list)";
		push @buff, "vd                              - Printout download queue / stats";
		push @buff, "cancel \$hash                    - Cancel this download";
		push @buff, "pause \$hash                     - Stop downloading/seeding given torrent";
		push @buff, "resume \$hash                    - Restart downloading/seeding given torrent";
		push @buff, "autocommit \$hash [command]      - Enable or disable autocommit; Type 'help autocommit' for more information";
		push @buff, "multifile \$hash [command]       - Multifile-Torrent commands; Type 'help multifile' for more information";
		push @buff, "details \$hash                   - Display information aboug \$hash";
		push @buff, "autoload                        - Forces re-scan of autoloaddir";
		push @buff, "publish announce=http://...     - Create a torrent and seed it; Type 'help publish' for more information";
		push @buff, "version                         - Printout version and uptime";
		push @buff, "config                          - Change configuration ; Type 'help config' for more information";
		push @buff, "user                            - Change/Create users ; Type 'help user' for more information";
		push @buff, "help                            - This output";
		push @buff, "analyze file.torrent (ADVANCED) - Display decoded bencoding";
		push @buff, "verify \$hash         (ADVANCED) - Check Storage directory.";
		push @buff, "bitmask              (ADVANCED) - Display bitmask/bitfield of all downloads";
		push @buff, "cs                   (ADVANCED) - Display client-statistics";
		push @buff, "wall \$message        (ADVANCED) - Broadcast messate to all connected telnet clients";
		push @buff, "!                               - ReExecute last command";
		push @buff, "quit                            - Disconnect telnet client";
		push @buff, "die                             - Shutdown bitflu process";
		
		foreach(@subpages) {
			if($_ eq 'publish') {
				push @buff, "-----------------------------------------";
				push @buff, "publish: How to create a .torrent file";
				push @buff, "-----------------------------------------";
				push @buff, "Step #1: Place your content inside ./publish";
				push @buff, "Step #2: Create the torrent file:";
				push @buff, " # publish announce=http://tracker.example.com:6969/announce [name=\"description\"]";
				push @buff, " -> Use 'nodes' to create a DHT torrent:";
				push @buff, " # publish nodes=127.0.0.1:6681 [nodes=127.0.0.2:6682]";
				push @buff, "-NOTE-: 'name' is optional ; you can specify multiple trackers: 'announe=http://tracker1 announce=http://tracker2 ...'";
				push @buff, "         LoadBalancing is also possible, use '#' to splitup balanced trackers: announce=http://tracker1#http://tracker2";
				push @buff, "Step #3: Distribute the created torrent file: torrents/\$hash";
				push @buff, "-NOTE-: You can delete the contents of ./publish after running the 'publish' command";
			}
			elsif($_ eq "config") {
				push @buff, "-----------------------------------------";
				push @buff, "config: Modify or save configuration";
				push @buff, "-----------------------------------------";
				push @buff, "config show                     - Show current configuration";
				push @buff, "config set key value            - Change setting (Example: config set maxup 50)";
				push @buff, "config save                     - Store current configuration";
			}
			elsif($_ eq "autocommit") {
				push @buff, "-----------------------------------------";
				push @buff, "autocommit: Commit completed files";
				push @buff, "-----------------------------------------";
				push @buff, "autocommit \$hash off            - Disable autocommit feature";
				push @buff, "autocommit \$hash on 1.3         - Commit \$hash if completed and ration >= 1.3";
				push @buff, "autocommit \$hash status         - Show autocommit ration-value";
			}
			elsif($_ eq "multifile" or $_ eq "mf") {
				push @buff, "-----------------------------------------";
				push @buff, "multifile: Pick single files from torrent";
				push @buff, "-----------------------------------------";
				push @buff, "multifile \$hash list            - Show status of all included files";
				push @buff, "multifile \$hash stop all        - Stop download of all included files";
				push @buff, "multifile \$hash start all       - Start download of all included files";
				push @buff, "multifile \$hash stop 3 6 8      - Stop download of file 3, 6 and 8";
				push @buff, "multifile \$hash start 2 9 22    - Start download of file 2, 9 and 22";
			}
			elsif($_ eq "user") {
				push @buff, "-----------------------------------------";
				push @buff, "user: Create/Change bitflu users";
				push @buff, "-----------------------------------------";
				push @buff, "user list                       - List existing users";
				push @buff, "user add \$USER \$PASSWORD        - Creates a new user";
				push @buff, "user password \$USER \$OLD \$NEW   - Change password";
				push @buff, "user delete \$USER \$PASSWORD     - Remove account";
				push @buff, "WARNING: #1 Run 'config save' to store any changes you made!";
				push @buff ,"         #2 Your passwords are travelling unencrypted! (telnet & http)";
			}
		}
		
		return(defined,\@buff);
	}
	
	
	
	sub Details {
		my($hash) = @_;
		
		my $buff = "Details for $hash\n";
		   $buff .="====================================================\n";
		unless($TORRENT_UNIVERSE->{$hash}) {
			$buff .= "Hash $hash does not exist";
			return(undef,$buff);
		}
		
		my %stats = ();
		my %missmap = ();
		my $donestats = Bitflu::Store::GetDoneStats($hash);
		
		$stats{total_pieces} = Bitflu::Store::GetNumberOfPieces($hash)+1;
		$stats{total_pleft}  = $stats{total_pieces} - $donestats->{pieces};
		
		
		foreach my $xpiece (0..Bitflu::Store::GetNumberOfPieces($hash)) {
			unless(Bitflu::Utils::GetBit(\$TORRENT_UNIVERSE->{$hash}->{bitfield},$xpiece)) {
				$missmap{$xpiece} = defined;
			}
		}
		
		
		foreach my $cc (keys(%TCSTAT)) {
			my $tref = $TCSTAT{$cc};
			next if $tref->{INFO_HASH} ne $hash;
			$stats{total_clients}++;
			$stats{total_me_unchoked}++   if !$tref->{ME_CHOKE};
			$stats{total_peer_unchoked}++ if !$tref->{PEER_CHOKE};
			$stats{total_me_inter}++      if $tref->{ME_INTERESTED};
			$stats{total_peer_inter}++    if $tref->{PEER_INTERESTED};
			foreach my $xpiece (keys(%missmap)) {
				if(Bitflu::Utils::GetBit(\$tref->{BITFIELD},$xpiece)) {
					delete($missmap{$xpiece});
				}
			}
			my $xcpieces = unpack("B*", $TCSTAT{$cc}->{BITFIELD});
			   $xcpieces =~ tr/1//cd; $xcpieces = length($xcpieces);
			if($xcpieces == $stats{total_pieces}) { $stats{total_seeding}++;  }
			else                                  { $stats{total_leeching}++; }
		}
		
		
		
		my $missmap_str = join(',',keys(%missmap));
		if(length($missmap_str) > 65) {
			$missmap_str = int(keys(%missmap))." pieces are not seeded";
		}
		elsif(length($missmap_str) == 0) {
			$missmap_str = "None";
		}
		
		
		
		$buff .= sprintf("Connected peers                   : %d\n",$stats{total_clients});
		$buff .= sprintf("Peers seeding this torrent        : %d\n",$stats{total_seeding});
		$buff .= sprintf("Peers leeching this torrent       : %d\n",$stats{total_leeching});
		$buff .= sprintf("Peers with new pieces             : %d\n",$stats{total_me_inter});
		$buff .= sprintf("Peers willing to download from us : %d\n",$stats{total_peer_inter});
		$buff .= sprintf("Peers willing to upload to us     : %d\n",$stats{total_me_unchoked});
		$buff .= sprintf("Peers we are uploading to         : %d\n",$stats{total_peer_unchoked});
		$buff .= sprintf("Total number of pieces to download: %d\n",$stats{total_pieces});
		$buff .= sprintf("Number of pieces not downloaded   : %d\n",$stats{total_pleft});
		$buff .= sprintf("Missing pieces with no seeders    : %s\n",$missmap_str);
		
		
		
		
		return(defined,$buff);
	}
	
	
	sub SelectiveIexclude {
		my($hash,$do_include,@list) = @_;
		unless($TORRENT_UNIVERSE->{$hash}) {
			return(undef, "Hash $hash does not exist");
		}
		
		my $offsetmap = Bitflu::Store::CalculateFileOffsets($hash);
		my $excldir   = Bitflu::Store::_getexcludedir($hash);
		my $freedir   = Bitflu::Store::_getfreedir($hash);
		my $workdir   = Bitflu::Store::_getworkdir($hash);
		my $numpieces = Bitflu::Store::GetNumberOfPieces($hash);
		my %exclmap   = ();
		my $filenum   = 0;
		
		if($list[0] eq "all") {
			@list = ();
			foreach (0..int(@$offsetmap)) {
				push(@list, ($_+1));
			}
		}
		
		
		
		if(!$do_include) { foreach(@list) { $exclmap{int($_)} = 1; } }
		foreach my $com (@$offsetmap) {
			$filenum++;
			next if $exclmap{$filenum};
			foreach my $cpart ($com->{start_piece}..$com->{end_piece}) {
				if(-e $excldir."/".$cpart) {
					$exclmap{$filenum} = 1;
					last;
				}
			}
		}
		if($do_include) { foreach(@list) { delete($exclmap{int($_)}); } }
		
		# Everything inside %exclmap should be excluded if possible ; everything else has to
		# be included.
		my $piecemap = "0" x $numpieces;
		
		
		$filenum = 0;
		foreach my $com (@$offsetmap) {
			$filenum++;
			next if $exclmap{$filenum}; # Is excluded
			foreach my $cpiece ($com->{start_piece}..$com->{end_piece}) {
				substr($piecemap,$cpiece,1,1);
			}
		}
		
		
		my %xdlpiggyback = ();
		
		foreach my $cpiece (0..$numpieces) {
			my $exclname = $excldir."/".$cpiece;
			my $freename = $freedir."/".$cpiece;
			my $workname = $workdir."/".$cpiece;
			if(substr($piecemap,$cpiece,1) == 1) {
				# -> INCLUDE
				if(-e $exclname) {
					rename($exclname,$freename) or Carp::confess("Unable to rename $exclname -> $freename : $!");
				}
			}
			else {
				# -> EXCLUDE
				if(-e $freename) {
					rename($freename,$exclname) or Carp::confess("Unable to rename $freename -> $exclname : $!");
				}
				elsif(-e $workname) {
					$xdlpiggyback{$cpiece} = 1;
				}
			}
		}
		
		main::CancelAllPiecesOfHash(\%xdlpiggyback,$hash);
		foreach my $cp (keys(%xdlpiggyback)) {
			rename($freedir."/".$cp, $excldir."/".$cp) or Carp::confess("Unable to exclude $cp\@$hash : $!");
		}

		return SelectiveDoList($hash);
	}
	
	
	sub SelectiveDoList {
		my($hash) = @_;
		
		unless($TORRENT_UNIVERSE->{$hash}) {
			return(undef, "$hash does not exist");
		}
		
		my @buff = undef;
		my $offsetmap = Bitflu::Store::CalculateFileOffsets($hash);
		my $torrent   = ${$TORRENT_UNIVERSE->{$hash}->{torrent}};
		my $filenum   = 0;
		my $donedir   = Bitflu::Store::_getdonedir($hash);
		my $excldir   = Bitflu::Store::_getexcludedir($hash);
		
		push @buff, "Status   |FileId| Done | Path";
		push @buff, "------------------------------------------------------------------";
		foreach my $file_hash (@{$torrent->{content}->{info}->{files}}) {
			$filenum++;
			my $path = join('/', @{$file_hash->{path}});
			my $com  = ${$offsetmap}[$filenum-1];
			my $excluded = 0;
			my $missing  = 0;
			my $total    = 0;
			foreach my $cpart ($com->{start_piece}..$com->{end_piece}) {
				if(-e $excldir."/".$cpart) {
					$excluded++;
				}
				elsif(! (-e $donedir."/".$cpart)) {
					$missing++;
				}
				$total++;
			}
			my $lineout = sprintf("%4d | %3d%% | %s",$filenum, (100-(($missing+$excluded)/$total*100)),substr($path,0,0xFF));
			if($excluded)   { $lineout = "Excluded | ".$lineout; }
			elsif($missing) { $lineout = Bitflu::Ansi::Green("Loading  | ".$lineout); }
			else            { $lineout = Bitflu::Ansi::Yellow("Completed| ".$lineout); }
			
			push @buff, $lineout;
		}
		
		push @buff, "------------------------------------------------------------------";
		push @buff, "$hash has $filenum file(s)";
		
		if($filenum == 0) {
			return(defined, "$hash is not a multifile-torrent");
		}
		else {
			return(defined, \@buff);
		}
	}
	

sub _dv {
	my($n) = @_;
	
	my $i = $n-int(time());
	return "-" if $i < 0;
	return $i;
}


sub ExecuteIpcCommands {
	my($rawline) = @_;
	foreach my $rawline_part (split(/\r?\n/,$rawline)) {
		my ($command, @args) = _deToken($rawline_part);
		
		if($command eq "wall") {
			Wall($rawline_part, undef);
		}
		elsif($command eq "cancelandlink") {
			foreach(@args) { CancelAndLinkTorrent($_) }
		}
		else {
			Carp::confess("Unknown RPC command '$command'; ABORTING");
		}
		
	}
}

sub _MubleAllKeywordToAllTorrentHashes {
	my @newa = ();
	foreach(@_) {
		if($_ eq "all") {
			foreach(keys(%$TORRENT_UNIVERSE)) {
				push(@newa,$_);
			}
		}
		else {
			push(@newa,$_);
		}
	}
	return @newa;
}

sub ExecuteTelnetCommands {
	my($handle,$socket,$rawline) = @_;
	
	my @out = ();
	
	foreach my $rawline_part (split(/\r?\n/,$rawline)) {
		my ($command, @args) = _deToken($rawline_part);
		
		if($command eq "!") {
			($command, @args) = _deToken($ACSTAT{$socket}->{LASTCOMMAND});
		}
		else {
			$ACSTAT{$socket}->{LASTCOMMAND} = $rawline_part;
		}
		
		if($command eq "quit" or $command eq "q" or $command eq "exit") {
			Quit($handle,$socket);
		}
		elsif($command eq "die" or $command eq "shutdown") {
			ShutdownProcess();
		}
		elsif($command eq "autocommit") {
			push @out, AutoCommit(@args);
		}
		elsif($command eq "pause") {
			foreach(_MubleAllKeywordToAllTorrentHashes(@args)) { push @out, Pause($_); }
		}
		elsif($command eq "resume") {
			foreach(_MubleAllKeywordToAllTorrentHashes(@args)) { push @out, Resume($_); }
		}
		elsif($command eq "load") {
			foreach(@args) { push @out, LoadTorrent($_); }
		}
		elsif($command eq "analyze") {
			foreach(@args) { push @out, AnalyzeTorrent($_); }
		}
		elsif($command eq "cancel") {
			foreach(@args) { push @out, CancelTorrent($_); }
		}
		elsif($command eq "verify") {
			foreach(_MubleAllKeywordToAllTorrentHashes(@args)) { push @out, VerifyTorrent($_); }
		}
		elsif($command eq "bitmask") {
			push @out, GlobalBitmask();
		}
		elsif($command eq "details") {
			foreach(_MubleAllKeywordToAllTorrentHashes(@args)) { push @out, Details($_); }
		}
		elsif($command eq "autoload") {
			push @out, AutoloadNow();
		}
		elsif($command eq "cs") {
			push @out, TelnetClientStatsList(@args);
		}
		elsif($command eq "vd" or $command eq "ls") {
			push @out, TelnetDownloadList();
		}
		elsif($command eq "version") {
			push @out, Version();
		}
		elsif($command eq "mf" or $command eq "multifile") {
			my $sel_hash = $args[0];
			my $sel_action=$args[1];
			if($sel_action eq "list") {
				push @out, SelectiveDoList($sel_hash);
			}
			elsif($sel_action eq "stop" or $sel_action eq "start") {
				splice(@args,0,2);
				my $do_include = 1;
				   $do_include = 0 if $sel_action eq "stop";
				push @out, SelectiveIexclude($sel_hash,$do_include,@args);
			}
			else {
				push @out, (undef, "Suboption '$sel_action' does not exist, see 'help $command' for more information");
			}
		}
		elsif($command eq "publish") {
			push @out, Publish(@args);
		}
		elsif($command eq "help" or $command eq "?") {
			push @out, Help(@args);
		}
		elsif($command eq "user") {
			if($args[0] eq "list") {
				push @out, ListUsers();
			}
			elsif($args[0] eq "password" && $args[3]) {
				push @out, ChangeUserPassword($args[1], $args[2],$args[3]);
			}
			elsif($args[0] eq "add" && $args[2]) {
				push @out, AddNewUser($args[1], $args[2]);
			}
			elsif($args[0] eq "delete" && $args[2]) {
				push @out, DeleteUser($args[1],$args[2]);
			}
			else {
				push @out, (undef, "Syntax error ; See 'help user' for more information'");
			}
		}
		elsif($command eq "config") {
			if($args[0] eq "show") {
				push @out, (defined, Data::Dumper::Dumper($config->{client}));
			}
			elsif($args[0] eq "set") {
				if(defined($config->{client}->{$args[1]})) {
					if($config->{client_locked}->{$args[1]}) {
						push @out, (undef, "Sorry: option '$args[1]' can't be changed at runtime!");
					}
					else {
						$config->{client}->{$args[1]} = $args[2];
						push @out, (defined, "Option '$args[1]' set to '$args[2]'");
					}
				}
				else {
					push @out, (undef, "Option '$args[1]' does not exist");
				}
			}
			elsif($args[0] eq "save") {
				if(my $cfh = $config->{SYS}->{configfh}) {
					seek($cfh,0,0) or Carp::confess("Seeking failed");
					truncate($cfh,0); # FatalError on error
					print $cfh "# Configuration saved at ".gmtime()."\n";
					foreach my $ckey (sort(keys(%{$config->{client}}))) {
						print $cfh $ckey." = ".$config->{client}->{$ckey}."\n";
					}
					print $cfh "# END\n";
					$cfh->autoflush(1); # force-flush
					# perl should auto-flush on newlines 
					push @out, (defined, "Configuration saved");
				}
				else {
					push @out, (undef, "Configuration not saved: no configfile specified/existing");
				}
			}
			else {
				push @out, (undef, "Unknown sub-command '$args[0]' ; Hint: type 'help $command'");
			}
		}
		elsif($command eq "commit" or $command eq "xcommit") {
			if(int(@args)) {
				my $do_remove = 0;
				   $do_remove = 1 if $command eq "commit";
				push @out, ForkCommit(\@args,$do_remove);
			}
			else {
				push @out, (undef, "$command needs an argument, type 'help' for more information");
			}
		}
		elsif($command eq "pcommit") {
			if($args[0] && defined($args[1])) {
				my @xtorrent = ($args[0]);
				splice(@args,0,1);
				my %xh = ();
				foreach(@args) { $xh{$_} = 1; };
				push @out, ForkCommit(\@xtorrent,0,\%xh);
			}
			else {
				push @out, undef, "$command needs an argument and file-id's, type 'help' for more information";
			}
		}
		elsif($command eq "wall") {
			Wall($rawline_part,$socket);
		}
		else {
			push @out, (undef, "Command '$command' not understood");
		}
	}
	
	for(my $i=0; $i < int(@out); $i+=2) {
		if(ref($out[$i+1]) eq "HASH") {
			$out[$i+1] = Data::Dumper::Dumper($out[$i+1]);
		}
		elsif(ref($out[$i+1]) eq "ARRAY") {
			my @ar = @{$out[$i+1]};
			$out[$i+1] = undef;
			foreach(@ar) {
				$out[$i+1] .= $_."\n";
			}
			chop($out[$i+1]);
		}
		unless(defined($out[$i])) {
			$out[$i+1] = Bitflu::Ansi::Red($out[$i+1]);
		}
		main::_xwA($socket, $out[$i+1]."\n");
	}
	
	
}


sub _deToken {
	my($line) = @_;
	
	my @p = ();
	
	my $aesc = undef;
	my $buff = undef;
	foreach my $token (split(/(\s|\\\"|")/,$line)) {
		next if $token eq '';
		if(defined($aesc)) {
			if($token eq '"') {
				push(@p,_toDE($aesc));
				$aesc=undef;
			}
			else {
				$aesc .= $token;
			}
			next;
		}
		elsif($token eq '"' && !defined($aesc)) {
			# Start escaping
			$aesc = '';
			$aesc .= $buff;
			$buff = undef;
			next;
		}
		elsif($token eq " ") {
			push(@p,_toDE($buff)) if defined $buff;
			$buff = undef;
			next; #Skip command seperation
		}
		$buff .= $token;
	}
	push(@p,_toDE($buff)) if defined $buff;
	push(@p,_toDE($aesc)) if defined $aesc;
	return @p;
}

sub _toDE {
	my($string) = @_;
	$string =~ s/\\\"/"/g;
	return $string;
}









1;

##################################################################################################################################
##################################################################################################################################
package Bitflu::Utils;


sub XXCopy {
	my($from_fh, $to_fh, $bytes_left) = @_;
	
	my $bytes_read    = 0;     # return of read
	my $bytes_written = 0;     # how much we wrote
	my $cbuff         = 0;     # current max-buffer
	my $total_written = 0;     # total amount of written data
	my $now_write     = 0;     # return of syswrite
	my $buffer        = undef; # payload buffer
	
	if($bytes_left < 0) {
		return undef; # no thanks!
	}
	
	for(;;) {
		if($bytes_left >= POSIX::BUFSIZ) { $cbuff = POSIX::BUFSIZ; }
		else                             { $cbuff = $bytes_left;   }

		$bytes_read = read($from_fh,$buffer,$cbuff); # Fill buffer with new data
		Carp::confess("Error while reading from fh $from_fh : $!") unless defined($bytes_read);
		
		$bytes_left -= $bytes_read;
		
		$bytes_written = 0;
		while($bytes_written != $bytes_read) {
			$now_write = syswrite($to_fh,$buffer,POSIX::BUFSIZ,$bytes_written); # Write read data into $to_fh
			Carp::confess("Error while writing to fh $to_fh : $!") unless defined($now_write);
			$bytes_written += $now_write;
		}
		
		$total_written += $bytes_written;
		if($bytes_left == 0 or $bytes_read == 0) { # Hit NumWant or EOF
			return $total_written;
		}
		elsif($bytes_left < 0) {
			Carp::confess("BUG");
		}
		
	}
}


sub GetRation {
	my($hash) = @_;
	Carp::confess("$hash: not downloading!") unless $TORRENT_UNIVERSE->{$hash};
	return sprintf("%6.2f",($TORRENT_UNIVERSE->{$hash}->{config}->{uploaded}/($TORRENT_UNIVERSE->{$hash}->{config}->{downloaded}+1)));
}


sub SetBit {
	my($bfref, $num) = @_;
	my $string = "0" x $num;
	   $string .= 1;
	$$bfref = $$bfref | pack("B*", $string);
}

sub GetBit {
	my($bfref, $num) = @_;
	my $bitfield = unpack("B*", "$$bfref");
	return(substr(unpack("B*", "$$bfref"),$num,1));
}

########################################################################
# A very flanky base64 decoder
sub DecodeBase64 {
	my($string) = @_;
	$string =~ tr/A-Za-z0-9+=//cd;     # Remove badchars
	$string =~ s/=+$//;                # Throw away padding
	my $b64len = int(length($string)*3/4);
	$string =~ tr/A-Za-z0-9+\// -_/;   # PCRE magic: Convert to UUencode :-) (chr(32) -> chr(95))
	return undef unless $string;       # Wowies? only padding?!
	return undef if $b64len > (ord("_")-ord(" "));
	$string = chr(ord(" ")+$b64len)."$string";
	$string = unpack("u",$string);
	return $string;
}

########################################################################
# Checks if password for given user is correct
sub UserPasswordIsOkay {
	my($user,$password) = @_;

	foreach my $chunk (split(/;/,$config->{client}->{userauth})) {
		my($cuser,$cpass) = $chunk =~ /^([^:]+):(\S+)$/;
		if($cuser eq $user) {
			my $shallbe = Digest::SHA1::sha1_hex("$user:$password");
			if($shallbe eq $cpass) {
				return 1;
			}
			else {
				return undef;
			}
		}
	}
	return undef; #Didn't even exist!
}


########################################################################
# Try to get the BootID on a Linux host
# This may/will silently fail on non-linux os'es
sub GetLinuxBootId {
	open(LBOOTID, "/proc/sys/kernel/random/boot_id") or return "NO-BOOT-ID";
	my $buff = <LBOOTID>;
	chomp($buff);
	close(LBOOTID);
	Carp::confess("Empty boot-id ?!") unless $buff;
	return $buff;
}

########################################################################
# Creates a FAST-EXTENSIONS FastSet
sub CreateFastSet {
	my($hash, $ip, $pieces, $numfast) = @_;
	
	$numfast ||= 10; # default
	
	if($pieces < $numfast) {
		return {};
	}
	
	my %a = ();
	my $order = 0;
	# Convert IPv4 into value
	$ip = hex(sprintf("%02X%02X%02X%02X", split(/\./,$ip)));
	my $x = sprintf("%X",(0xFFFFFF00 & $ip));
	   $x .= $hash;
	while(int(keys(%a)) < $numfast) {
		$x = Digest::SHA1::sha1_hex(pack("H*",$x));
		for(my $i=0; $i<5 && int(keys(%a)) < $numfast; $i++) {
			my $j = $i*8;
			my $y = hex(substr($x,$j,8));
			my $index = $y % $pieces;
			defined($a{$index}) or $a{$index} = ++$order;
		}
	}
	return \%a;
}

########################################################################
# Decodes Compact IP-Chunks
sub DecodeCompactIp {
	my($compact_list) = @_;
	my @peers = ();
		for(my $i=0;$i<length($compact_list);$i+=6) {
			my $chunk = substr($compact_list, $i, 6);
			my $a    = unpack("C", substr($chunk,0,1));
			my $b    = unpack("C", substr($chunk,1,1));
			my $c    = unpack("C", substr($chunk,2,1));
			my $d    = unpack("C", substr($chunk,3,1));
			my $port = unpack("n", substr($chunk,4,2));
			my $ip = "$a.$b.$c.$d";
			push(@peers, {ip=>$ip, port=>$port, peer_id=>"COMPACT"});
		}
	return @peers;
}

########################################################################
# Creates single Compact-IP chunk
sub EncodeCompactIp {
	my($r) = @_;
	my $buff = undef;
	
	my $ip     = $r->{ip};
	my $port   = $r->{port};
	my $funny_assert = 0;
	foreach my $cx (split(/\./,$ip)) { $buff .= pack("c",$cx); $funny_assert++; }
	Carp::confess("BUGBUG => $ip") if $funny_assert != 4;
	$buff .= pack("n",$port);
	return $buff;
}

sub RemoveDownloadFromProcess {
	my($hash, $destination) = @_;
	Carp::confess("Unable to remove non-existing download $hash") unless $TORRENT_UNIVERSE->{$hash};

	#Ok, this is'nt such an easy job..
	#1 : Mumble all clients
	foreach my $rs (keys(%TCSTAT)) {
		next unless $TCSTAT{$rs}->{INFO_HASH} eq $hash;
		$TCSTAT{$rs}->{BTS} = main::BTS_UNDEF;
		$TCSTAT{$rs}->{TIMEOUT} = 0; #Let the client timeout on next can_write run
		delete($TCSTAT{$rs}->{PIECE_INWORK});
		delete($TCSTAT{$rs}->{INFO_HASH});
	}
	## Kill the torrent
	Bitflu::Store::UnstoreTorrent($hash);
	Bitflu::Store::WipeAllPieces($hash);
	## Rename the Torrent dir
	my $cdir = Bitflu::Store::CancelDirectory($hash, $destination);
	## ..and kill the torrent_universe
	delete($TORRENT_UNIVERSE->{$hash});
	return $cdir;
}


1;



##################################################################################################################################
##################################################################################################################################
package Bitflu::Tracker;

use constant HTTP_REQUEST_TIMEOUT => 10;
use constant MAX_TRACKER_RETRY    => 3; #Try to contact the tracker 3 times (including redirects)
use constant TRACKER_MIN_INTERVAL => 360; # Minimal interval value for Tracker replys

use List::Util 'shuffle';


#################################################################
# Ask tracker and populate .nodes directory
sub StartBoot {
	my($hash) = @_;
	
	my $nowtime = time();
	my $skew    = int(rand(0xF));
	
	return 0 if $TORRENT_UNIVERSE->{$hash}->{config}->{tracker_nextquery} > $nowtime; # Not yet!
	
	# SafeSet a new interval
	$TORRENT_UNIVERSE->{$hash}->{config}->{tracker_nextquery} = $nowtime + TRACKER_MIN_INTERVAL + $skew;
	
	# Save configuration before contacting tracker
	Bitflu::Store::SaveConfig($hash, $TORRENT_UNIVERSE->{$hash}->{config});
	
	my $tracker_ref = Query($hash);
	return 0 unless $tracker_ref; # All trackers are dead!
	
	# Update tracker_nextquery
	if($tracker_ref->{interval} > TRACKER_MIN_INTERVAL) {
		$TORRENT_UNIVERSE->{$hash}->{config}->{tracker_nextquery} = $nowtime + $tracker_ref->{interval} + $skew;
	}
	
	# and populate .nodes
	Bitflu::Store::AddTorrentNodes($tracker_ref->{peers}, $hash, "TRACKER");
	return 1;
}

#################################################################
# Check what the next event for the tracker shall be
sub NextTrackerEvent {
	my($hash) = @_;
	
	if($TORRENT_UNIVERSE->{$hash}->{config}->{last_tracker_event} eq "") {
		# -> No event sent ever
		$TORRENT_UNIVERSE->{$hash}->{config}->{last_tracker_event} = 'started';
		return 'started';
	}
	elsif($TORRENT_UNIVERSE->{$hash}->{config}->{last_tracker_event} ne 'completed' &&
	      $TORRENT_UNIVERSE->{$hash}->{is_complete}) {
		# -> Completed download!
		$TORRENT_UNIVERSE->{$hash}->{config}->{last_tracker_event} = 'completed';
		return 'completed';
	}
	else {
		return '';
	}
}



#################################################################
# Fillup .nodes directory using a Tracker
sub Query {
	my($info_hash) = @_;
	Carp::confess("No valid info hash") unless $info_hash;
	my $ref = ${$TORRENT_UNIVERSE->{$info_hash}->{torrent}}  or Carp::confess("No ref for this info hash :-(");
	Carp::confess("BUG: $ref->{torrent_hash} != $info_hash") if $ref->{torrent_hash} ne $info_hash;

	my @tracker_list = ($ref->{content}->{announce});
	my $decoded_response = undef;
	my @peers            = ();
	
	my $tracker_event = NextTrackerEvent($info_hash); #Note: IF we fail to contact all trackers we will
	                                                  #go to the next event anyway.. i do not care about this..
	

	
	# Torrent has an announce-list : Build tracker_list array
	if(ref($ref->{content}->{'announce-list'}) eq "ARRAY") {
		@tracker_list = ();
		foreach my $tracker_array ((@{$ref->{content}->{'announce-list'}})) {
			foreach my $xtracker (shuffle(@{$tracker_array})) {
				push(@tracker_list,$xtracker);
			}
		}
	}
	
	# .. and kick blacklisted trackers
	for(my $i=0;$i<int(@tracker_list);$i++) {
		if($config->{client}->{trackerblacklist} && $tracker_list[$i] =~ /$config->{client}->{trackerblacklist}/) {
			main::l(main::L_WARN, "[TrackerList] Removing blacklisted tracker '$tracker_list[$i]' from announce-list");
			$tracker_list[$i] = undef;
		}
	}
	
	
	foreach my $current_tracker (@tracker_list) {
		next if $current_tracker !~ /^http:\/\//i; # non-http entry?!
		
		my $httpH = {peer_id=>$config->{SYS}->{peer_id}, key=>$config->{SYS}->{peer_key},
		             stats_downloaded=>$TORRENT_UNIVERSE->{$info_hash}->{config}->{downloaded},
		             stats_uploaded=>$TORRENT_UNIVERSE->{$info_hash}->{config}->{uploaded},
		             stats_left=>($TORRENT_UNIVERSE->{$info_hash}->{total_size}-$TORRENT_UNIVERSE->{$info_hash}->{config}->{downloaded}),
		             port=>$config->{client}->{port}, info_hash=>$info_hash, uri=>$current_tracker, event=>$tracker_event};
		
		
		my $result = undef;
		
		for(0..MAX_TRACKER_RETRY) {
			$result = doHTTP($httpH);
			if($result->{CODE} == 301 or $result->{CODE} == 302) {
				# -> Redirect
				delete($httpH->{uri});
				$httpH->{full_uri} = $result->{HEADER}->{location}
			}
			elsif($result->{CODE} == 200) {
				last;
			}
		}
		
		
		if($result->{CODE} != 200) {
			main::l(main::L_WARN, "Unable to contact tracker $current_tracker for $info_hash ; last response: $result->{CODE}");
			next;
		}
		elsif(!$result->{BODY}) {
			main::l(main::L_WARN, "[TrackerList] Tracker $current_tracker for $info_hash returned an empty body");
			next;
		}
		elsif(! ($decoded_response = Bitflu::Bencoding::decode($result->{BODY}) ) ) {
			main::l(main::L_WARN, "[TrackerList] Tracker $current_tracker for $info_hash returned a non-bencoded response");
			next;
		}
		
		if( ref($decoded_response->{peers}) ne "ARRAY" ) {
			# -> Compact response
			@peers = Bitflu::Utils::DecodeCompactIp($decoded_response->{peers});
		}
		else {
			# -> Normal response
			@peers = @{$decoded_response->{peers}};
		}
		main::l(main::L_WARN, "[TrackerList] Tracker $current_tracker has been used for $info_hash");
		return ({interval=>$decoded_response->{interval}, peers=>\@peers});
	}
	
	main::l(main::L_WARN, "[TrackerList] $info_hash : no trackers contacted");
	return undef;
	
}



sub _http_encode_info_hash {
	my($in) = @_;
	my $buff = undef;
	foreach my $k ((1..(length($in)/2))) {
		$buff .= "%".substr($in,(($k-1)*2),2);
	}
	return $buff;
}




sub doHTTP {
	my($h) = @_;
	
	foreach (qw(stats_left stats_uploaded stats_downloaded)) {
		if($h->{$_} < 1) {
			#stats_left has a good chance to be below zero if we
			#received even 1 bad chunk..
			main::l(main::L_INFO, "Fixing invalid size of $h->{$_} -> $_");
			$h->{$_} = 0;
		}
	}
	
	
	my $peer_id       = $h->{peer_id}; #Fixme: needs HTTP uri encoding
	my $key           = $h->{key};
	my $port          = int($h->{port});
	my $downloaded    = int($h->{stats_downloaded});
	my $uploaded      = int($h->{stats_uploaded});
	my $left          = int($h->{stats_left});
	my $event         = $h->{event};
	my $URI           = $h->{uri};
	my $FULL_URI      = $h->{full_uri};
	my $torrent_hash  = _http_encode_info_hash($h->{info_hash});
	
	my ($tracker_host, $tracker_port, $tracker_base) = undef;
	my %ret = ();
	
	my $query = "GET ";
	
	if($FULL_URI) {
		($tracker_host,$tracker_port,$tracker_base) = $FULL_URI =~ /^http:\/\/([^\/:]+):?(\d*)\/(.+)$/i;
		$tracker_port ||= 80;
		$query .= "/$tracker_base HTTP/1.0\r\nHost: $tracker_host:$tracker_port\r\n\r\n";
	}
	else {
		#1: Decode tracker infos
		($tracker_host,$tracker_port,$tracker_base) = $URI =~ /^http:\/\/([^\/:]+):?(\d*)\/(.+)$/i;
		$tracker_port ||= 80;
		main::l(main::L_INFO, "Tracker($URI) decoded: HOST:$tracker_host / PORT:$tracker_port / BASE:$tracker_base");
		
		#2: Create the GET request
		my $nextarg = "?";
		   $nextarg = "&" if($tracker_base =~ /\?/); # announce url allready has an ? -> start with a &
		$query .= "/$tracker_base".$nextarg."info_hash=$torrent_hash";
		$query .= "&peer_id=$peer_id&port=$port&uploaded=$uploaded&downloaded=$downloaded&left=$left&key=$key&event=$event&compact=1";
		$query .= " HTTP/1.0\r\nHost: $tracker_host:$tracker_port\r\n\r\n";
	}
	
	
	return({ERROR=>"Unable to extract parts"}) unless $tracker_host;
	
	
	my $socket = IO::Socket::INET->new(PeerPort=>$tracker_port, PeerHost=>$tracker_host, Proto=>'tcp', Timeout=>2) or return({ERROR=>"$!"});
	
	alarm(HTTP_REQUEST_TIMEOUT);
	$SIG{ALRM} = sub { die "TIMEOUT\n"; };
	eval {
		print $socket $query;
		($ret{CODE}) = <$socket> =~ /^HTTP\/\d\.\d (\d{3})/;
		while(<$socket>) {
			chomp($_); chop($_);
			if($_ =~ /^([^:]+): (.+)$/) {
				$ret{HEADER}->{lc($1)} = $2;
			}
			last if length($_) == 0;
		}
		
		while(<$socket>) {
			$ret{BODY} .= $_;
		}
	};
	alarm(0);
	$SIG{ALRM} = undef;
	
	if($@) {
		return({ERROR=>"Timed out"});
	}
	else {
		return \%ret;
	}
}

1;

##################################################################################################################################
##################################################################################################################################
package Bitflu::Bencoding;


	sub decode {
		my($string) = @_;
		my @chars = split(//,$string);
		return _decode(\@chars);
	}
	
	sub encode {
		my($ref) = @_;
		return _encode($ref);
	}
	
	
	
	sub _encode {
		my($ref) = @_;
		
		my $encoded = undef;
		
		if(ref($ref) eq "HASH") {
			$encoded .= "d";
			foreach(sort keys(%$ref)) {
				$encoded .= length($_).":".$_;
				$encoded .= _encode($ref->{$_});
			}
			$encoded .= "e";
		}
		elsif(ref($ref) eq "ARRAY") {
			$encoded .= "l";
			foreach(@$ref) {
				$encoded .= _encode($_);
			}
			$encoded .= "e";
		}
		elsif($ref =~ /^(\d+)$/) {
			$encoded .= "i$1e";
		}
		else {
			# -> String
			$encoded .= length($ref).":".$ref;
		}
		return $encoded;
	}
	
	
	sub _decode {
		my($ref) = @_;
		
		my $cx = shift(@$ref);
		if($cx eq 'd') {
			# -> A Dict : Cycle-Decode chunks until we get to an 'e'
			my $stolen_item = shift(@{$ref});
			my %ret_hash = ();
			while(defined($stolen_item) && $stolen_item ne 'e') {
				unshift(@{$ref},$stolen_item);
				my $key = _decode($ref);
				my $val = _decode($ref);
				$ret_hash{$key} = $val;
				$stolen_item = shift(@{$ref}); #NextOne
			}
			return \%ret_hash;
		}
		elsif($cx =~ /^\d$/) {
			# -> LengthPrefixed string
			my $string = undef;
			my $length = $cx;
			
			my $stolen_item = shift(@{$ref});
			while(defined($stolen_item) && $stolen_item ne ':') {
				$length .= $stolen_item;
				$stolen_item = shift(@{$ref});
			}
			for(1..$length) {
				$string .= shift(@{$ref});
			}
			return $string;
		}
		elsif($cx =~ 'i') {
			# -> An int, cycle until we find a non-int (e)
			my $fullnum = undef;
			while(defined (my $append = shift(@{$ref}))) {
				last if $append !~ /^\d$/;
				
				$fullnum .= $append;
			}
			return $fullnum;
		}
		elsif($cx eq 'l') {
			# -> A List (aka Array) ; Cycle-Decode until we got an 'e'
			my $stolen_item = shift(@{$ref});
			my @ret_array = ();
			while(defined($stolen_item) && $stolen_item ne 'e') {
				unshift(@{$ref}, $stolen_item);
				push(@ret_array, _decode($ref));
				$stolen_item = shift(@{$ref});
			}
			return \@ret_array;
		}
		elsif(defined($cx)) {
			warn "$0::Bitflu::Bencoding : Unexpected entry '$cx' found; Aborting parsing due to syntax error.\n";
			return undef;
		}
	}

#################################################################
# Load a torrent file
sub torrent2hash {
	my($file) = @_;
	my $buff = undef;
	open(BENC, $file) or return {};
	while(<BENC>) {
		$buff .= $_;
	}
	close(BENC);
	return {} unless $buff;
	my $href = decode($buff);
	return {} unless ref($href) eq "HASH";
	my $torrent_hash =  Digest::SHA1::sha1_hex(encode(${$href}{info}));
	return {torrent_hash=>$torrent_hash, content=>$href};	
}



1;


##################################################################################################################################
##################################################################################################################################
package Bitflu::Protocol;

use constant MSG_CHOKE          => 0;
use constant MSG_UNCHOKE        => 1;
use constant MSG_INTERESTED     => 2;
use constant MSG_NOT_INTERESTED => 3;
use constant MSG_HAVE           => 4;
use constant MSG_BITFIELD       => 5;
use constant MSG_REQUEST        => 6;
use constant MSG_PIECE          => 7;
use constant MSG_CANCEL         => 8;  # Not implemented
use constant MSG_PORT           => 9;  # DHT Extension
use constant MSG_WANT_METAINFO  => 10; # Unused
use constant MSG_METAINFO       => 11; # Unused
use constant MSG_SUSPECT_PIECE  => 12; # Unused
use constant MSG_SUGGEST_PIECE  => 13; # FastPeers
use constant MSG_HAVE_ALL       => 14; # FastPeers   | IMPLEMENTED  (Recv)      (Silly extension: no need to send this)
use constant MSG_HAVE_NONE      => 15; # FastPeers   | IMPLEMENTED  (Recv)       (Silly extension: no need to send this)
use constant MSG_REJECT_REQUEST => 16; # FastPeers   | IMPLEMENTED (Recv|Send)
use constant MSG_ALLOWED_FAST   => 17; # FastPeers   | IMPLEMENTED (Recv|Send)
use constant MSG_HOLE_PUNCH     => 18; # NAT-Unused
use constant MSG_UTORRENT_MSG   => 19; # Unused (??)
use constant MSG_UT_EXT         => 20; # Extended Handshake protocol


#################################################################
# Assemble a FastPieceAllowed message
sub CreateAllowedFast {
	my $x = pack("N",5);
	$x   .= pack("c", MSG_ALLOWED_FAST);
	$x   .= pack("N",$_[0]);
	return $x;
}

#################################################################
# Assembles a piece request
# Needs {index, offset, size}
sub CreateRequest {
	my($hr,$tref) = @_;
	Carp::confess("BUGBUG: Wrong CreateRequest call! :".Data::Dumper::Dumper($hr)) if ($hr->{offset} <0 or $hr->{index} < 0 or $hr->{size} < 1);
	main::l(main::L_DEBUG_CRX, "CreateRequest { PIECE : $hr->{index} | | OFFSET : $hr->{offset} | SIZE : $hr->{size} }");
	
	if($tref->{ME_CHOKE} == 1 && !defined($tref->{FASTALLOWED}->{$hr->{index}})) {
		Carp::confess("CreateRequest wisely did not allow such a stupid request");
	}
	
	my $x = pack("N",13);
	$x .= pack("c", MSG_REQUEST);
	$x .= pack("N", $hr->{index});
	$x .= pack("N", $hr->{offset});
	$x .= pack("N", $hr->{size});
	return $x;
}

#################################################################
# Creates a CANCEL message
sub CreateCancel {
	my($hr) = @_;
	Carp::confess("BUGBUG: Wrong CreateCancel call! :".Data::Dumper::Dumper($hr)) if ($hr->{offset} <0 or !defined($hr->{index}) or $hr->{size} < 1);
	my $x = pack("N",13);
	$x .= pack("c", MSG_CANCEL);
	$x .= pack("N", $hr->{index});
	$x .= pack("N", $hr->{offset});
	$x .= pack("N", $hr->{size});
	return $x;
}

#################################################################
# Creates A REJECT message
sub CreateReject {
	my($hr) = @_;
	my $x = pack("N",13);
	$x .= pack("c", MSG_REJECT_REQUEST);
	$x .= pack("N", $hr->{index});
	$x .= pack("N", $hr->{offset});
	$x .= pack("N", $hr->{size});
	return $x;
}

#################################################################
# Send a bitfield
sub CreateBitfield {
	my($bitfield) = @_;
	my $x = pack("N", (1+length($bitfield)));
	$x   .= pack("c", MSG_BITFIELD);
	$x   .= $bitfield;
	return $x;
}

#################################################################
# Assembles a handshake request
# Needs {info_hash, peer_id}
sub CreateHandshake {
	my($hr) = @_;
	Carp::confess("BUGBUG") unless $hr->{info_hash} or $hr->{peer_id};

	my $extensions = AssembleExtensions($hr);

	my $x .= pack("c", 19);
	$x .= "BitTorrent protocol";
	$x .= $extensions;
	$x .= pack("H40", $hr->{info_hash});
	$x .= $hr->{peer_id};
	return $x;
}

#################################################################
# Creates the Extensions-part of an Handshake
sub AssembleExtensions {
	my($h) = @_;
	my $ext = "0" x 64;
	
	if(defined($h->{EXT_eproto})) {
		#Enables Enhanced Messages
		substr($ext,43,1,1);
	}
	if(defined($h->{EXT_fastpeers})) {
		substr($ext,61,1,1);
	}
	if(defined($h->{EXT_dht})) {
		substr($ext,63,1,1);
	}
	return pack("B64",$ext);
}


sub CreatePiece {
	my($hr) = @_;
	my $x .= pack("N", 9+$hr->{size});
	$x    .= pack("c", MSG_PIECE);
	$x    .= pack("N", $hr->{index});
	$x    .= pack("N", $hr->{offset});
	$x    .= $hr->{data};
	return $x;
}

sub CreateExtended {
	my($type,$bencoded) = @_;
	my $x .= pack("N", 2+length($bencoded));
	   $x .= pack("c", MSG_UT_EXT);
	   $x .= pack("c", int($type));
	   $x .= $bencoded;
	return $x;
}

#################################################################
# Returns a Keepalive request
sub CreateKeepalive {
	return(pack("cccc",0000));
}

sub CreateHave {
	my $x = pack("N",5);
	$x   .= pack("c", MSG_HAVE);
	$x   .= pack("N",$_[0]);
	return $x;
}

#################################################################
# Returns an Unchoke request
sub CreateUnchoke {
	my($tref) = @_;
	Carp::confess() unless $tref;
	Carp::confess("BUG") unless $tref->{PEER_CHOKE};
	$tref->{PEER_CHOKE} = 0;
	$tref->{'UNCHOKED@'} = int(time());
	my $x = pack("N",1);
	   $x .= pack("c",MSG_UNCHOKE);
	return $x;
}

#################################################################
# Returns a Choke request
sub CreateChoke {
	my($tref) = @_;
	Carp::confess() unless $tref;
	Carp::confess("BUG") if $tref->{PEER_CHOKE};
	$tref->{PEER_CHOKE} = 1;
	$tref->{'CHOKED@'} = int(time());
	my $x = pack("N",1);
	   $x .= pack("c",MSG_CHOKE);
	return $x;
}

#################################################################
# returns an INTERESTED request
sub CreateInterested {
	my($tref) = @_;
	Carp::confess() unless $tref;
	return if $tref->{ME_INTERESTED};
	$tref->{ME_INTERESTED} = 1;
	
	my $x = pack("N",1);
	   $x .= pack("c",MSG_INTERESTED);
	return $x;
}

#################################################################
# returns an NOT-INTERESTED request
sub CreateNotInterested {
	my ($tref) = @_;
	Carp::confess() unless $tref;
	return unless $tref->{ME_INTERESTED};
	$tref->{ME_INTERESTED} = 0;
	my $x = pack("N",1);
	   $x .= pack("c",MSG_NOT_INTERESTED);
	return $x;
}

#################################################################
# Parse handshake response, returns unparsed bytes in _->{ol}
sub ParseHandshake {
	my($buff) = @_;

	if(length($buff) >= 68) {
		if(unpack("c",substr($buff,0,1)) == 19 &&
		   substr($buff,1,19) eq "BitTorrent protocol") {
			my $extensions_raw = unpack("B*", substr($buff,20,8));
			my $info_hash = substr($buff,28,20);
			my $client_id = substr($buff,48,20);
			
			my $ext = ();
			   $ext->{DHT}    = 1 if substr($extensions_raw, 63,1) == 1;
			   $ext->{FAST}   = 1 if substr($extensions_raw, 61,1) == 1;
			   $ext->{EPROTO} = 1 if substr($extensions_raw, 43,1) == 1;
			
			return({ol=>substr($buff,68), info_hash=>unpack("H40",$info_hash), client_id=>$client_id, extensions=>$ext});
		}
		else {
			return({ol=>$buff});
		}
	}
	else {
		return({ol=>$buff});
	}
}


1;

package Bitflu::RPC;

##################################################################################################################################
# Execute a 'telnet' command
sub __execute {
	my($cmd) = @_;
	$cmd =~ tr/\n\r//d;
	$cmd .= "\n"; #Seperator
	# Could deadlock for sure..
	POSIX::write($ipc_write, $cmd, length($cmd));
	return undef;
}

##################################################################################################################################
# Execute the 'wall' command
sub wall {
	my($string) = @_;
	return __execute("wall $string");
}

sub cancel_and_link {
	my($hash) = @_;
	return __execute("cancelandlink $hash");
}

1;


##################################################################################################################################
##################################################################################################################################
package Bitflu::Store;
	
	use constant WORKDIR => ".working";
	use constant FREEDIR => ".free";
	use constant DONEDIR => ".done";
	use constant CONFDIR => ".settings";
	use constant NODESDIR => ".nodes";
	use constant DHTBOOT => ".dhtboot";
	use constant EXCLUDEDIR => ".exclude";
	use constant TORRENTS => "torrents";
	use constant STORAGELOCK => ".BitfluLock";
	
	use constant EXCLUDE_PREFIX => "_exclude_";
	
	#################################################################
	# Init Bitflu::Store -> Creates the root directores
	sub _INIT {
		my $rootbase        = __get_rootbase() or Carp::confess("BUGBUG: NO ROOTBASE");
		my $torrentbase     = __get_torrentbase() or Carp::confess("BUGBUG : NO TORRENTBASE");
		my $publishbase     = __get_publishbase() or Carp::confess("BUGBUG : NO PUBLISHBASE");
		my $autoloadbase    = __get_autoloadbase() or Carp::confess("BUGBUG : NO AUTOLOADBASE");
		my $commitedbase    = __get_commitedbase() or Carp::confess("BUGBUG : NO COMMITEDBASE");
		my $canceledbase    = __get_canceledbase() or Carp::confess("BUGBUG : NO CANCELEDBASE");
		my $downloadingbase = __get_downloadingbase() or Carp::confess("BUGBUG : NO DOWNLOADINGBASE");
		my $historybase     = __get_historybase() or Carp::confess("BUGBUG : NO HISTORYBASE");
		unless(-d $rootbase) {
			mkdir($rootbase) or Carp::confess("Unable to create rootbase '$rootbase' : $!");
		}
		
		chdir($rootbase) or Carp::confess("Unable to chdir($rootbase) : $!");
		
		foreach my $cdir ( ($torrentbase, $publishbase, $autoloadbase, $commitedbase, $canceledbase,
		                    $downloadingbase, $historybase) ) {
			unless(-d $cdir) {
				mkdir($cdir) or Carp::confess("Unable to create $cdir : $!");
			}
		}

	}
	
	#################################################################
	# Lock the storage directory
	sub LockStorage {
		my($lock_content) = @_;
		my $file = STORAGELOCK;
		
		Carp::confess("NO CONTENT TO LOCK") unless $lock_content;
		
		if(-e $file) {
			my $lbuff = undef;
			open(OLDLOCK, $file) or Carp::confess("Unable to open existing lockfile $file : $!");
			while(<OLDLOCK>) { $lbuff .= $_; }
			close(OLDLOCK);
			return $lbuff;
		}
		else {
			open(SLOCK, ">$file") or Carp::confess("Unable to write lockfile $file : $!");
			print SLOCK "$lock_content";
			close(SLOCK);
			return $lock_content;
		}
	}
	
	#################################################################
	# Remove Lockfile from storage directory
	sub UnlockStorage {
		my $file = STORAGELOCK;
		return undef unless -e $file;
		unlink($file) or Carp::confess("Unable to remove lockfile $file : $!");
		return 1;
	}
	
	#################################################################
	# Stors given torrent
	sub StoreTorrent {
		my($from,$to) = @_;
		$to ||= Carp::confess("BUGBUG");
		$to = __get_torrentbase()."/".$to;
		my @statA = stat($from);
		my @statB = stat($to);
		if($statA[1] == $statB[1]) {
			return $to; #<- same file!
		}
		my $buff = undef;
		open(IN, "$from") or Carp::confess("Unable to open $from : $!");
		open(OUT, ">$to") or Carp::confess("Unable to open $to : $!");
		while(<IN>) { print OUT $_; }
		close(IN);
		close(OUT);
		return $to;
	}
	
	#################################################################
	# Removes a torrent from store
	sub UnstoreTorrent {
		my($hash) = @_;
		_getbase($hash); #Assert check: Does the file exist?
		my $file = TORRENTS."/".$hash;
		unlink($file) or Carp::confess("Failed to unlink $file : $!");
	}
	
	#################################################################
	# Add a hash to history
	sub HistoryAdd {
		my($hash) = @_;
		my $hdir = _gethistorydir($hash);
		
		unless(-d $hdir) {
			mkdir($hdir) or Carp::confess("Unable to create $hdir : $!");
		}
	}
	
	#################################################################
	# Check if file exists within history
	sub HistoryCheck {
		my($hash) = @_;
		my $href = ();
		my $hdir = _gethistorydir($hash);
		my @STAT = stat($hdir);
		$href->{exists}  = (-d $hdir);
		$href->{addtime} = $STAT[9];
		$href->{path}    = $hdir;
		return $href;
	}
	
	
	#################################################################
	# Removes all pieces from a download
	sub WipeAllPieces {
		my($hash) = @_;
		my $donedir = _getdonedir($hash);
		my $freedir = _getfreedir($hash);
		my $workdir = _getworkdir($hash);
		
		for(0..GetNumberOfPieces($hash)) {
			my $piecenum = $_;
			foreach( ($donedir,$freedir,$workdir) ) {
				my $cf = $_."/".int($piecenum);
				if(-f $cf) {
					unlink($cf) or Carp::confess("Unable to unlink $cf : $!");
					last;
				}
			}
		}
	}
	
	#################################################################
	# Moves a basedir
	sub CancelDirectory {
		my($hash,$destdir) = @_;
		
		my $source = _getbase($hash);
		Carp::confess("Invalid Destdir $destdir") unless -d $destdir;
		Carp::confess("Invalid torrent $hash")  unless $TORRENT_UNIVERSE->{$hash};
		Carp::confess("Invalid Sourcedir $source") unless -d $source;
		my $dstname = _cure_name(${$TORRENT_UNIVERSE->{$hash}->{torrent}}->{content}->{info}->{name});
		
		for(0..0xFFFF) {
			my $dest = $destdir."/".$dstname;
			   $dest .= ".$_" if $_;
			next if -e $dest;
			rename($source, $dest) or Carp::confess("Unable to rename $source -> $dest : $!");
			return($dest);
		}
		Carp::confess("CancelDirectory($hash,$destdir) failed to find a new home for $hash :-(");
	}
	
	
	#################################################################
	# Adds a list of nodes to .nodes
	sub AddTorrentNodes {
		my($aref, $hash, $source) = @_;
		Carp::confess("No source?!") unless $source;
		my $nodesdir = _getnodesdir($hash);
		foreach(@$aref) {
			my $ip =   $_->{ip};
			my $port = int($_->{port});
			_addNode($_->{ip}, $_->{port}, $nodesdir, $source);
		}
	}
	
	#################################################################
	# Private sub to add nodes
	sub _addNode {
		my($ip,$port,$basedir,$source) = @_;
		$port = int($port);
		return 0 unless $ip =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/; # Fixme: needs sub _isValidIP(v6) ?
		Carp::confess("No basedir?") unless $basedir;
		Carp::confess("No source?")  unless $source;
		my $file = "$basedir/$ip:$port";
		open(XF, ">$file") or Carp::confess("Unable to write $file : $!");
		print XF $source;
		close(XF);
		return 1;
	}
	
	#################################################################
	# Stores DHT-Node inside .dhtboot directory
	sub AddDhtNode {
		my($hash, $ip, $port) = @_;
		my $bootdir =_getdhtbootabledir($hash);
		return _addNode($ip,$port,$bootdir, 'DHT');
	}
	
	#################################################################
	# Return torrent basedir
	sub __get_torrentbase {
		return TORRENTS;
	}
	
	sub __get_publishbase {
		return $config->{client}->{publishdir};
	}
	
	sub __get_downloadingbase {
		return $config->{client}->{downloaddir};
	}
	
	#################################################################
	# Return main-root basedir
	sub __get_rootbase {
		return $config->{client}->{rootdir};
	}
	
	#################################################################
	# Returns Torrent-Autoload directory
	sub __get_autoloadbase {
		return $config->{client}->{autoloaddir};
	}
	
	sub __get_commitedbase {
		return $config->{client}->{commitdir};
	}
	
	sub __get_canceledbase {
		return $config->{client}->{canceldir};
	}
	
	sub __get_historybase {
		return $config->{client}->{historydir};
	}
	
	
	#################################################################
	# Get base of given hash
	sub _getbase {
		my($hash) = @_;
		Carp::confess("NO HASH") unless $TORRENT_UNIVERSE->{$hash};
		return __get_downloadingbase()."/".$hash;
	}
	
	#################################################################
	# Returns history-Directory
	sub _gethistorydir {
		my($hash) = @_;
		Carp::confess("INVALID HASH : $hash") if ($hash !~ /^([a-z0-9]{40})$/);
		return __get_historybase()."/".$hash;
	}
	
	#################################################################
	# Get workdir of given hash
	sub _getworkdir {
		return _getbase($_[0])."/".WORKDIR;
	}

	#################################################################
	# Get freedir of given hash
	sub _getfreedir {
		return _getbase($_[0])."/".FREEDIR;
	}
	
	#################################################################
	# Get DistributedHashTablePeers dir
	sub _getnodesdir {
		return _getbase($_[0])."/".NODESDIR;
	}
	
	sub _getdhtbootabledir {
		return _getbase($_[0])."/".DHTBOOT;
	}

	#################################################################
	# Get donedir of given hash
	sub _getdonedir {
		return _getbase($_[0])."/".DONEDIR;
	}
	
	#################################################################
	# Parts we are NOT going to download..
	sub _getexcludedir {
		return _getbase($_[0])."/".EXCLUDEDIR;
	}
	
	#################################################################
	# Get configdir of given hash
	sub _getconfdir {
		return _getbase($_[0])."/".CONFDIR;
	}

	#################################################################
	# Returns number of pieces of given info_hash (0 = 1 piece!)
	sub GetNumberOfPieces {
		my($hash) = @_;
		my $pieces = $TORRENT_UNIVERSE->{$hash}->{pieces};
		Carp::confess("No pieces?") unless defined $pieces;
		return $pieces;
	}
	
	#################################################################
	# Saves config-ref for given hash
	sub SaveConfig {
		my($hash, $ref) = @_;
		my $base = _getconfdir($hash);
		foreach my $setting (keys(%$ref)) {
			open(CONF, ">$base/$setting") or Carp::confess("Unable to write '$base/$setting' : $!");
			print CONF $ref->{$setting};
			close(CONF);
		}
	}

	#################################################################
	# Returns the config-ref for given hash
	sub GetConfig {
		my($hash) = @_;
		my $base = _getconfdir($hash);
		my $ref = ();
		opendir(CONFDIR, $base) or return $ref;
		while(my $key = readdir(CONFDIR)) {
			my $fp = $base."/".$key;
			next unless -f $fp;
			open(FILE, $fp) or Carp::confess("Unable to read '$fp' : $!");
			while(<FILE>) {
				$ref->{$key} .= $_;
			}
			close(FILE);
		}
		closedir(CONFDIR);
		return $ref;
	}
	
	#################################################################
	# Returns downloaded pieces + total size : Note: this is just a 'guess'
	sub GetDoneStats {
		my($hash) = @_;
		my $numpieces = 0;
		my $size      = 0;
		
		my $bitfield = unpack("B*", $TORRENT_UNIVERSE->{$hash}->{bitfield});
		   $bitfield =~ tr/1//cd;
		   $numpieces = length($bitfield);
		   $size      = (($numpieces-1)*($TORRENT_UNIVERSE->{$hash}->{piece_len}))+$TORRENT_UNIVERSE->{$hash}->{lastpiece_len} if $numpieces > 0;
		return({pieces=>$numpieces, size=>$size});
	}

	#################################################################
	# Returns basename of torrent-file
	sub _get_torrent_basename {
		my($hash) = @_;
		Carp::confess("No valid hash") unless $TORRENT_UNIVERSE->{$hash};
		my $torrent = ${$TORRENT_UNIVERSE->{$hash}->{torrent}};
		my $conout = _cure_name($torrent->{content}->{info}->{name} or "OUT");
		return($conout);
	}

	##################################################################
	# Returns full-path basename of torrent-file
	sub _get_concat_name {
		my($hash) = @_;
		my $conout = _cure_name(_get_torrent_basename($hash));
		my $base   = _getbase($hash);
		return($base."/".$conout);
	}
	
	
	#################################################################
	# Removes evil chars from a filename.. this isn't perfect :-/
	sub _cure_name {
		my($string) = @_;
		$string =~ s/\.\.\//___/g;
		$string =~ tr/\/|//d;
		return $string;
	}

	
	sub AssembleData {
		my($hash,$pcomlist) = @_;
		
		my $outbase    = _get_concat_name($hash);                        # Multifile torrent -> output dir | Single file -> filename
		my $donedir    = _getdonedir($hash);                             # .done
		my $excldir    = _getexcludedir($hash);                          # .excluded
		my $torrent    = ${$TORRENT_UNIVERSE->{$hash}->{torrent}};       # torrent hashref
		my $offsetmap  = CalculateFileOffsets($hash);                    # Reference with offset-map
		my $buffer     = undef;                                          # WRITE buffer
		my $filenum    = 0;
		my $missing    = 0;
		my $do_pcommit = int(keys(%$pcomlist));
		my $num_okcom  = 0;
		
		return("$outbase exists; commit aborted") if -d $outbase;
		mkdir($outbase) or Carp::confess("FATAL: Unable to create '$outbase' : $!");
		
		foreach my $file_hash (@{$torrent->{content}->{info}->{files}}) {
			my @file_path   = @{$file_hash->{path}};
			my $file_name   = $outbase."/";
			for(my $i=0; $i < (int(@file_path)-1); $i++) {
				my $safepart = _cure_name($file_path[$i]);
				$file_name .= $safepart."/";
				unless(-d $file_name) {
					main::l(main::L_DEBUG, "mkdir($file_name)");
					mkdir($file_name) or Carp::confess("Unable to create '$file_name' : $!");
				}
			}
			$file_name .= _cure_name($file_path[-1]);
			
			my $file_offsetmap = ${$offsetmap}[$filenum];
			my $file_unlinkable= 0;
			my $this_missing   = 0;
			if($do_pcommit && !$pcomlist->{$filenum+1}) {
				main::l(main::L_WARN, "Skipping $file_name");
				$filenum++;
			}
			else {
				main::l(main::L_WARN, "Writing $file_name");
				open(OUTFILE, ">$file_name") or Carp::confess("Unable to create $file_name : $!");
				for my $partnum ($file_offsetmap->{start_piece}..$file_offsetmap->{end_piece}) {
					my $xseek = 0;
					   $xseek = $file_offsetmap->{start_offset} if $partnum == $file_offsetmap->{start_piece};
					my $xsize = $file_offsetmap->{normal_length}-$xseek;
					   $xsize = $file_offsetmap->{end_offset}-$xseek if $partnum == $file_offsetmap->{end_piece};
					
					my $partfile      = $donedir."/".$partnum;
					my $partfile_skip = $excldir."/".$partnum;
					
					if(-f $partfile) {
						open(PART, $partfile) or Carp::confess("Unable to open $partfile : $!");
					}
					elsif(-f $partfile_skip) {
						$file_unlinkable = 1;
						$missing++;
						$this_missing++;
						last;
					}
					else {
						main::l(main::L_WARN, "$partfile is missing, simulating part using /dev/zero");
						open(PART, "/dev/zero") or Carp::confess("Unable to open /dev/zero : $!");
						$missing++;
						$this_missing++;
					}
					
					if($xseek != 0) { seek(PART,$xseek,0) or Carp::confess("SEEKFAIL: $!"); }
					my $xwritten = Bitflu::Utils::XXCopy(*PART, *OUTFILE, $xsize);
					Carp::confess("XXCopy failed: $xwritten != $xsize") if $xsize != $xwritten;
					close(PART);
					
				}
				close(OUTFILE);
				if($file_unlinkable) {
					main::l(main::L_WARN, "Removing excluded and incomplete file $file_name");
					unlink($file_name) or Carp::confess("Unable to unlink $file_name : $!");
				}
				$num_okcom++ if $this_missing == 0;
				$filenum++;
			}
		}
		
		
		if($filenum == 0 && $do_pcommit == 0) {
			# -> SingleFile torrent
			$outbase .= "/"._cure_name(_get_torrent_basename($hash));
			main::l(main::L_WARN, "Writing single-file $outbase");
			open(CONCAT, ">$outbase") or Carp::confess("Unable to open '$outbase' for writing: $!");
			foreach my $partnum (0..GetNumberOfPieces($hash)) {
				my $partfile = $donedir."/".$partnum;
				if(-f $partfile) {
					open(PART, $partfile) or Carp::confess("Unable to open $partfile : $!");
				}
				else {
					main::l(main::L_WARN, "$partfile is missing, simulating part using /dev/zero");
					open(PART, "/dev/zero") or Carp::confess("Unable to open /dev/zero : $!");
					$missing++;
				}
				my $xsize = GetLengthOfPiece($hash,$partnum);
				my $xwritten = Bitflu::Utils::XXCopy(*PART, *CONCAT, $xsize);
				Carp::confess("XXCopy failed: $xwritten != $xsize") if $xsize != $xwritten;
				close(PART);
			}
			$num_okcom++ if $missing == 0;
			$filenum++;
		}
		
		
		if($missing > 0) {
			return("$num_okcom files (out of $filenum) commited");
		} else {
			return undef;
		}
	}
	
	
	sub CalculateFileOffsets {
		my($hash) = @_;
		my $torrent = ${$TORRENT_UNIVERSE->{$hash}->{torrent}};        # torrent hashref
		my $plength = $torrent->{content}->{info}->{'piece length'};   # length of a normal piece
		my $piecenum= GetNumberOfPieces($hash);                        # Pieces inside torrent
		
		my $cnum = 0;
		my $tnum = 0;
		my @A    = ();
		foreach my $fhash (@{$torrent->{content}->{info}->{files}}) {
			my $this_start = $cnum;
			my $this_len   = $fhash->{length};
			my $this_end = $cnum  = $cnum+$this_len;
			my $xpiece_start = int($this_start/$plength);
			my $xpiece_end   = int($this_end  /$plength);
			my $xoffset_start = $this_start-($xpiece_start*$plength);
			my $xoffset_end   = $this_end-($xpiece_end*$plength);
			$A[$tnum] = { start_piece => $xpiece_start , end_piece => $xpiece_end, start_offset => $xoffset_start,
			              end_offset=>$xoffset_end, normal_length=>$plength, real_length=>$this_len };
			$tnum++;
		}
		return \@A;
	}
	

	#################################################################
	# Create base layout of a workdir
	sub InitTorrent {
		my($hash) = @_;
		my $torrent = ${$TORRENT_UNIVERSE->{$hash}->{torrent}};
		Carp::confess("InitTorrent($hash) : hash is unknown") unless $torrent;
		Carp::confess("BugBug! InitTorrent called a 2nd time?!") if $TORRENT_UNIVERSE->{$hash}->{total_size};
		## Calculate total_size
		my $total_size = 0;
		if($torrent->{content}->{info}->{files}) {
			foreach(@{$torrent->{content}->{info}->{files}}) {
				my $h = $_;
				$total_size += $h->{length};
			}
		}
		else {
			$total_size = $torrent->{content}->{info}->{length};
		}
		$TORRENT_UNIVERSE->{$hash}->{total_size} = $total_size;
		$TORRENT_UNIVERSE->{$hash}->{config}->{downloaded} = 0;
		$TORRENT_UNIVERSE->{$hash}->{config}->{uploaded}   = 0;
		$TORRENT_UNIVERSE->{$hash}->{config}->{tracker_nextquery} = undef;
		
		## Now we are going to calculate the number of pieces,
		## -> THIS makes GetNumberOfPieces functional!
		my $unpacked_sha = length(unpack("H*", $torrent->{content}->{info}->{pieces}));
		Carp::confess("$hash is a br0ken torrent : $unpacked_sha < 40") if $unpacked_sha < 40;
		$TORRENT_UNIVERSE->{$hash}->{pieces}     = int($unpacked_sha/40)-1;
		
		# Calculate size of LAST piece
		my $allbutone = $total_size - (GetNumberOfPieces($hash)
		                               * $torrent->{content}->{info}->{'piece length'});
		
		$TORRENT_UNIVERSE->{$hash}->{lastpiece_len} = $allbutone;
		$TORRENT_UNIVERSE->{$hash}->{piece_len}     = $torrent->{content}->{info}->{'piece length'};
		
		
		if($allbutone > $TORRENT_UNIVERSE->{$hash}->{piece_len}) {
			main::l(main::L_WARN, "Torrent with $hash seriously broken! (Last piece larger than normal piece)");
			delete($TORRENT_UNIVERSE->{$hash}); #Byebye!
			return undef;
		}
		
		
		if(-d $hash && ! (-d _getbase($hash))) {
			# FIXME: REMOVEME after 2-3 releases
			main::l(main::L_WARN, "Migrating < 0.21 bitflu data for torrent $hash");
			rename($hash,_getbase($hash)) or Carp::confess("Unable to rename $hash into "._getbase($hash)." : $!");
		}
		
		## -> Init working directories
		if(-d _getbase($hash)) {
			main::l(main::L_INFO, "Starting Rollback for existing workdir: $hash ...");
			opendir(RBACK, _getworkdir($hash)) or Carp::confess("Unable to open workdir for $hash : $!");
			
			# The rollback places 'lost' pieces (= Pieces inside .working) somewhere
				while(defined (my $x = readdir(RBACK))) {
					next if substr($x, 0,1) eq ".";
					main::l(main::L_INFO, "  -> Rollback $x");
					
					my (undef,$this_plen) = GetWorkPieceOffset($hash,$x); #Check if this piece is complete
					
					if($this_plen > 0) {
						# Nope, still data to go: .free
						main::l(main::L_INFO, " -> FREE");
						SetPieceAsFree($hash,$x);
					}
					else {
						# Piece IS complete, verify SHA1
						
						open(DATA, _getworkdir($hash)."/$x") or Carp::confess("Unable to open $x : $!");
						my $verify_ok = VerifyPiece($torrent->{content}->{info}->{pieces},$x, *DATA);
						close(DATA);
						if($verify_ok) {
							# Ok, someone forgot to move this file into .done ;-)
							main::l(main::L_INFO, " -> DONE");
							SetPieceAsDone($hash,$x);
						}
						else {
							# Great: A corrupted-but-completed .workpiece : ZeroOut and requeue it as .free
							main::l(main::L_WARN, "Removing corrupted piece $x\@$hash");
							ZeroWorkPiece($hash,$x);
							SetPieceAsFree($hash,$x);
						}
					}
				}
			closedir(RBACK);
		}
		else {
			mkdir(_getbase($hash));             # Creates base directory
			mkdir(_getfreedir($hash));          # Creates .free
			for(0..GetNumberOfPieces($hash)) {
				# Init FakeBucks
				my $dst = _getfreedir($hash)."/".int($_);
				open(DUMMY, ">$dst") or Carp::confess("Unable to write file $dst : $!");
				close(DUMMY);
			}
			
		}
		
		mkdir(_getworkdir($hash));          # Creates .workdir
		mkdir(_getdonedir($hash));          # Creates .done
		mkdir(_getconfdir($hash));          # Creates .settings
		mkdir(_getdhtbootabledir($hash));   # Creates .dhtboot dir
		mkdir(_getnodesdir($hash));         # Creates .nodes dir
		mkdir(_getexcludedir($hash));       # Creates .exclude dir
		
		# (Re?)-Add dht nodes for torrent
		if(ref($torrent->{content}->{nodes}) eq "ARRAY") {
			main::l(main::L_WARN, "Adding nodes of $hash to DHT-Bootable list");
			foreach my $ar (@{$torrent->{content}->{nodes}}) {
				next unless ref($ar) eq "ARRAY";
				my $ip = $$ar[0];
				my $port = int($$ar[1]);
				Bitflu::Store::AddDhtNode($hash, $$ar[0], $$ar[1]) or main::l(main::L_WARN, "AddDhtNode($hash, $$ar[0], $$ar[1]) skipped.");
			}
		}
		
		main::l(main::L_WARN, "Updating bitfield for $hash ; this may take some seconds...");
		my $bitfield = GetBitfieldFromDisk($hash);
		$TORRENT_UNIVERSE->{$hash}->{bitfield} = $bitfield;
		
		# Try to get this pieces first to become a seeder
		$TORRENT_UNIVERSE->{$hash}->{hotpieces} = Bitflu::Utils::CreateFastSet($hash,sprintf("%d.%d.%d.0", int(rand(0xFF)),int(rand(0xFF)),int(rand(0xFF))),(GetNumberOfPieces($hash)+1),14);
		
		main::l(main::L_WARN, "Init of '$hash' finished!");
		return 1;
	}
	
	#################################################################
	# Calculate bitfield from data stored on disk
	sub GetBitfieldFromDisk {
		my($hash) = @_;
		my $bitfield = "";
		foreach my $piece (0..GetNumberOfPieces($hash)) {
			if(-e _getdonedir($hash)."/".int($piece)) {
				$bitfield .= "1";
			}
			else {
				$bitfield .= "0";
			}
		}
		return pack("B*",$bitfield);
	}
	
	#################################################################
	# Check for 'lost' chunks due to OS-Crashes (fsck)
	sub VerifyStorageDir {
		my($hash) = @_;
		
		my $donedir = _getdonedir($hash);
		my $freedir = _getfreedir($hash);
		my $workdir = _getworkdir($hash);
		my $excldir = _getexcludedir($hash);
		my $torrent = ${$TORRENT_UNIVERSE->{$hash}->{torrent}};
		
		my $HAD_TO_FIX = undef;
		
		foreach my $piece (0..GetNumberOfPieces($hash)) {
			my $xdone = $donedir."/".int($piece);
			my $xwork = $workdir."/".int($piece);
			my $xfree = $freedir."/".int($piece);
			my $xexcl = $excldir."/".int($piece);
			my $xappr = 0; # How often does the piece appear?
			
			foreach my $xp (($xdone,$xwork,$xfree,$xexcl)) {
				if($xappr && -e $xp) {
					main::l(main::L_WARN, "VerifyStorageDir: Removing $xp : duplicate piece!");
					unlink($xp) or Carp::confess("Unable to unlink $xp : $!\n");
					$HAD_TO_FIX++;
				}
				elsif(-e $xp) {
					$xappr++;
				}
			}
			
			if(!$xappr) {
				# Piece does not exist
				main::l(main::L_WARN, "$piece\@$hash got lost somehow, re-creating a newone ...");
				open(NP, ">".$xfree) or Carp::confess("Unable to write $xfree : $!\n");
					print NP undef;
				close(NP);
				$HAD_TO_FIX++;
			}
			elsif(-e $xdone) {
				open(DATA,$xdone) or Carp::confess("Unable to open $xdone : $!\n");
				if(VerifyPiece($torrent->{content}->{info}->{pieces},$piece, *DATA)) {
					main::l(main::L_WARN, "$piece\@$hash: Verify PASSED");
					close(DATA);
				}
				else {
					main::l(main::L_WARN, "$piece\@$hash: Verify FAILED");
					close(DATA);
					unlink($xdone) or Carp::confess("Unable to unlink $xdone: $!\n");
					open(NP, ">".$xfree) or Carp::confess("Unable to write $xfree : $!\n");
						print NP undef;
					close(NP);
					$HAD_TO_FIX++;
				}
			}
			elsif(-e $xwork) {
				# main::l(main::L_WARN, "$piece\@$hash: In transaction ; okay ...");
			}
			elsif(-e $xfree) {
				# main::l(main::L_WARN, "$piece\@$hash: Is free ; okay ...");
			}
			elsif(-e $xexcl) {
				# main::l(main::L_WARN, "$piece\@$hash: Is excluded ; okay ...");
			}
			else {
				Carp::confess("NOTREACHED : $xappr : $xwork\n");
			}
		}
		
		if($HAD_TO_FIX) {
			main::l(main::L_WARN, "$hash: Fixed $HAD_TO_FIX error(s) for $hash");
			return 0;
		}
		else {
			main::l(main::L_WARN, "$hash: no corruption found :-)");
			return 1;
		}
	}
	
	
	#################################################################
	# SHA1-Verify piece : Returns '1' if okay
	sub VerifyPiece {
		my($shastring,$piece, $fh) = @_;
		my $torrent_sha = unpack("H40", substr($shastring,$piece*20,20));
		my $sha = Digest::SHA1->new();
		seek($fh,0,0);
		   $sha->addfile($fh);
		my $file_sha = $sha->hexdigest;
		if($file_sha eq $torrent_sha) {
			return 1;
		}
		else {
			return 0;
		}
	}
	
	#################################################################
	# Returns true if PIECE of HASH is in freelist (= Not done and not in-transition)
	sub PieceIsFree {
		my($hash,$piece) = @_;
		return (-e _getfreedir($hash)."/".int($piece));
	}
	
	#################################################################
	# Mark piece to work
	sub SetPieceAsInwork {
		my($hash,$piece) = @_;
		main::l(main::L_DEBUG, "SetPieceAsInwork($hash,$piece)");
		Carp::confess("$piece is NOT free.. i can't set it to WORK!") unless PieceIsFree($hash,$piece);
		rename(_getfreedir($hash)."/".int($piece), _getworkdir($hash)."/".int($piece)) or Carp::confess("rename failed : $!");
		Carp::confess("Wowies") if PieceIsFree($hash,$piece);
	}

	#################################################################
	# Move piece into .done folder
	sub SetPieceAsDone {
		my($hash,$piece) = @_;
		main::l(main::L_DEBUG, "SetPieceAsDone($hash,$piece)");
		my $from = _getworkdir($hash)."/".int($piece);
		my $to   = _getdonedir($hash)."/".int($piece);
		rename($from,$to) or Carp::confess("Unable to rename $from to $to : $!");
	}
	
	#################################################################
	# Zeroes-out a piece from the workqueue
	sub ZeroWorkPiece {
		my($hash,$piece) = @_;
		my $file = _getworkdir($hash)."/".int($piece);
		Carp::confess("Unableo to zero non-existing work-piece $file") unless -e $file;
		open(KILL, ">$file") or Carp::confess("Unable to write to $file : $!");
		close(KILL);
	}
	
	#################################################################
	sub ZeroFreePiece {
		my($hash,$piece) = @_;
		my $file = _getfreedir($hash)."/".int($piece);
		Carp::confess("Unableo to zero non-existing free-piece $file") unless -e $file;
		open(KILL, ">$file") or Carp::confess("Unable to write to $file : $!");
		close(KILL);
	}
	
	#################################################################
	sub GetPieceOffset {
		my($hash,$piece, $base) = @_;
		my $file = "$base/".int($piece);
		Carp::confess("GetPieceOffset(@_) : $file does NOT exist") unless -e $file;
		
		my $size  = (-s $file);
		my $fplen = GetLengthOfPiece($hash,$piece);
		my $plen  = main::PIECESIZE;
		
		if( ($fplen - $size) < main::PIECESIZE ) {
			$plen = ($fplen - $size);
		}
		
		main::l(main::L_DEBUG,"GetPieceOffset($hash,$piece) => $size | $plen");
		
		return ($size,$plen);
	}
	
	#################################################################
	sub GetFreePieceOffset {
		my($hash,$piece) = @_;
		return GetPieceOffset($hash,$piece, _getfreedir($hash));
	}
	
	#################################################################
	sub GetWorkPieceOffset {
		my($hash,$piece) = @_;
		return GetPieceOffset($hash,$piece, _getworkdir($hash));
	}
	
	#################################################################
	# 'rollback' a piece from workdir to freedir
	sub SetPieceAsFree {
		my($hash,$piece) = @_;
		main::l(main::L_DEBUG, "SetPieceAsFree($hash,$piece)");
		unless(-e _getworkdir($hash)."/".int($piece)) {
			main::l(main::L_WARN, "BUGCHECK! Someone attemped to free an unused piece!");
			main::l(main::L_WARN, " HASH: $hash | PIECE : $piece");
			main::l(main::L_WARN, "================= TCSTAT DUMP ===============");
			main::l(main::L_WARN, Data::Dumper::Dumper(\%TCSTAT)."\n\n\n\r\n");
			sleep(2);
			main::l(main::L_WARN, "Throwing CONFESS\n");
			Carp::confess("\n\nSetPieceAsFreeBUG == BFD EXTIED!\n\n");
		}
		rename(_getworkdir($hash)."/".int($piece),_getfreedir($hash)."/".int($piece)) or die;
	}


	sub FreeWorkPieces {
		my($tref) = @_;
		foreach my $workpiece (keys(%{$tref->{PIECE_INWORK}})) {
			SetPieceAsFree($tref->{INFO_HASH}, $workpiece);
		}
		delete($tref->{PIECE_INWORK});
	}



	#################################################################
	# Check if we are allowed to request given piece from client
	sub PieceRequestOkForClient {
		my($piece,$client_ref) = @_;
		
		
		if($TORRENT_UNIVERSE->{$client_ref->{INFO_HASH}}->{config}->{paused}) {
			return 0;
		}
		elsif($client_ref->{ME_CHOKE} == 0) {
			return 1;
		}
		elsif($client_ref->{EXTENSIONS}->{FAST} && defined($client_ref->{FASTALLOWED}->{$piece})) {
			return 1;
		}
		else {
			return 0;
		}
	}
	
	#################################################################
	# Suggest a new piece depending on clients current state + Extensions
	sub SuggestNewPiece {
		my($client_ref,$exclude) = @_;
		if($client_ref->{ME_CHOKE} == 1 && $client_ref->{EXTENSIONS}->{FAST}) {
			main::l(main::L_DEBUG_SUGST, "-> FastSet piece");
			return SuggestNewFastSetPiece($client_ref,$exclude);
		}
		else {
			main::l(main::L_DEBUG_SUGST, "-> Normal Piece");
			return SuggestNewUnchokedPiece($client_ref,$exclude);
		}
	}
	
	#################################################################
	# Suggest a new Piece from a FastSet
	sub SuggestNewFastSetPiece {
		my($client_ref, $exclude) = @_;
		Carp::confess("No fast extension!") unless $client_ref->{EXTENSIONS}->{FAST};
		
		if(defined($client_ref->{SNP_CACHE}) && defined($client_ref->{FASTALLOWED}->{$client_ref->{SNP_CACHE}})
		   && $client_ref->{SNP_CACHE} != $exclude && PieceIsFree($client_ref->{INFO_HASH},$client_ref->{SNP_CACHE}) && (!defined($exclude) or $exclude != $client_ref->{SNP_CACHE}) ) {
			return $client_ref->{SNP_CACHE};
		}
		else {
			foreach my $index (List::Util::shuffle(keys(%{$client_ref->{FASTALLOWED}}))) {
				if(Bitflu::Utils::GetBit(\$client_ref->{BITFIELD}, $index) &&
				   (!defined($exclude) or $exclude != $index) &&
				   Bitflu::Store::PieceIsFree($client_ref->{INFO_HASH}, $index)) {
					$client_ref->{SNP_CACHE} = $index;
					return $index;
				}
			}
		}
		return undef;
	}
	
	#################################################################
	# Suggest a new piece to request for this client reference
	sub SuggestNewUnchokedPiece {
		my($client_ref, $exclude) = @_;
		main::l(main::L_DEBUG_SUGST, "SuggestNewPiece($client_ref->{INFO_HASH},$exclude)");
		
		if($TORRENT_UNIVERSE->{$client_ref->{INFO_HASH}}->{is_complete}) {
			return undef;
		}
		elsif(!$client_ref->{BITFIELD}) { #No bitfield = no hits for sure
			main::l(main::L_DEBUG_SUGST, "Aborting search as client has no bitfield");
			return undef;
		}
		elsif(defined($client_ref->{SNP_CACHE}) && (!defined($exclude) or $exclude != $client_ref->{SNP_CACHE})
		        && PieceIsFree($client_ref->{INFO_HASH},$client_ref->{SNP_CACHE})) {
			return $client_ref->{SNP_CACHE};
		}
		else {
			main::l(main::L_DEBUG_SUGST, "## CACHE MISS");
		}
		
		
		
		# Tactic 1 : Client EXCLUDED a piece, we are going to see if we could featch the NEXT piece
		if(defined($exclude)) {
			my $plusone = $exclude+1;
			if($plusone <= GetNumberOfPieces($client_ref->{INFO_HASH}) &&
			   Bitflu::Utils::GetBit(\$client_ref->{BITFIELD},$plusone) &&
			   PieceIsFree($client_ref->{INFO_HASH},$plusone)) {
				main::l(main::L_DEBUG_SUGST, "NextPiece for $client_ref->{INFO_HASH} $plusone");
				$client_ref->{SNP_CACHE} = $plusone;
				return $plusone;
			}
		}

		foreach my $hotpiece (keys(%{$TORRENT_UNIVERSE->{$client_ref->{INFO_HASH}}->{hotpieces}})) {
			if( Bitflu::Utils::GetBit(\$client_ref->{BITFIELD},$hotpiece) && 
			    Bitflu::Utils::GetBit(\$TORRENT_UNIVERSE->{$client_ref->{INFO_HASH}}->{bitfield},$hotpiece) == 0 &&
			    (!defined($exclude) or $exclude != $hotpiece) &&
			    PieceIsFree($client_ref->{INFO_HASH},$hotpiece)) {
				$client_ref->{SNP_CACHE} = $hotpiece;
				return $hotpiece;
			}
		}

		# Tactic 2 : Try to grab a random piece
		for(0..30) {
			my $sug = int(rand(GetNumberOfPieces($client_ref->{INFO_HASH})));
			if(   Bitflu::Utils::GetBit(\$client_ref->{BITFIELD},$sug) 
			   && Bitflu::Utils::GetBit(\$TORRENT_UNIVERSE->{$client_ref->{INFO_HASH}}->{bitfield},$sug) == 0
			   && (!defined($exclude) or $exclude != $sug)
			   && PieceIsFree($client_ref->{INFO_HASH},$sug)) {
				main::l(main::L_DEBUG_SUGST, "RandomPiece for $client_ref->{INFO_HASH} $sug");
				$client_ref->{SNP_CACHE} = $sug;
				return $sug;
			}
		}

		
		# Tactic 3 : Walk down the full .free list
		# This isn't such a big performance hit: There ain't much pieces left if we are doing Tactic #3
		my $freelist = 0;
		opendir(FREELIST, _getfreedir($client_ref->{INFO_HASH})) or Carp::confess("Unable to open freedir");
		while(defined (my $x = readdir(FREELIST))) {
			next if substr($x, 0,1) eq ".";
			$freelist++;
			next if (defined($exclude) && int($x) == $exclude);
			next unless Bitflu::Utils::GetBit(\$client_ref->{BITFIELD},$x);
			closedir(FREELIST);
			main::l(main::L_DEBUG_SUGST, "WalkedDownPiece for $client_ref->{INFO_HASH} : $x");
			$client_ref->{SNP_CACHE} = $x;
			return $x;
		}
		closedir(FREELIST);

		#Idee Fixme: Nun haben wir ja 2 walkdowns.. beide sind je nach situation ungsntig:
		# Wir knnten einfach die Stats vom torrent angucken und z.B. bei ber 80% done randomPiece berspringen und den besseren
		# walkdown nehmen..

		main::l(main::L_DEBUG_SUGST, "found NOTHING for $client_ref->{INFO_HASH} ; Freelist has $freelist piece(s) left");
		
		## We Should PENALTY this guy!
		
		return undef;
	}
	
	
	
	
	
	#################################################################
	# Retruns length of Piece # in hash X
	sub GetLengthOfPiece {
		my($hash, $piece) = @_;
		my $num = 0;
		if($piece == GetNumberOfPieces($hash)) {
			$num = $TORRENT_UNIVERSE->{$hash}->{lastpiece_len};
		}
		else {
			$num = $TORRENT_UNIVERSE->{$hash}->{piece_len};
		}
		return $num;
	}
	
	
	
	
	
	#################################################################
	# Read data from disk
	sub ReadPiece {
		my($ref) = @_;
		Carp::confess("ReadPIece was called without a HASH!") unless $ref->{hash};
		my $torrent = ${$TORRENT_UNIVERSE->{$ref->{hash}}->{torrent}};
		Carp::confess("BUGBUG: Torrent $ref->{hash} not found in TORRENT_UNIVERSE") unless ref($torrent) eq "HASH";
		
		my $file = _getdonedir($ref->{hash})."/".int($ref->{index});
		my $buff = undef;
		open(DAT, $file) or return undef;
		seek(DAT, $ref->{offset},0);
		my $bytes_read = read(DAT,$buff,$ref->{size});
		close(DAT);
		main::l(main::L_DEBUG_RPICE, "ReadPiece($ref) : OFFSET:$ref->{offset}, REQUESTED: $ref->{size} , READ: $bytes_read");
		return ($buff,$bytes_read); #We don't care if read failed..
	}
	
	
	
	
	
	#################################################################
	# Store and Verify a piece
	#
	# The piece will *NOT* be anymore in .work IF $rollover is != 0 
	#
	sub StorePiece {
		my($ref) = @_;
		Carp::confess("StorePiece was called without a HASH!") unless $ref->{hash};
		my $torrent = ${$TORRENT_UNIVERSE->{$ref->{hash}}->{torrent}};
		Carp::confess("BUGBUG: Torrent $ref->{hash} not found in TORRENT_UNIVERSE") unless ref($torrent) eq "HASH";
		
		
		my $client_points = 0;                  #Will be set to the points the client earned for this piece
		my $next_offset   = 0;                  #Will be set to the next offset if not a rollover did NOT happen
		my $piece         = $ref->{index};      #Piece we are storing
		my $next_size     = main::PIECESIZE;    #Size of next request if a rollover did NOT happen
		my $rollover      = undef;              #Did a rollover happen?
		my $verified      = undef;              #Was the piece VERIFIED (= SHA OK)
		my $work_file     = _getworkdir($ref->{hash})."/".int($piece); # File we are writing to

		my $length_of_full_piece    = GetLengthOfPiece($ref->{hash},$piece);
		
		my $length_of_saved_data    = (-s $work_file);
		my $offset_of_current_piece = $ref->{offset};
		
		main::l(main::L_DEBUG_STORE, "STRORE PIECE:$piece | DATLEN: $length_of_saved_data | APPEND: ".length($ref->{data}));
		
		if( !(-e $work_file) ) {
			Carp::confess("Piece $work_file was *NOT* requested ; is NOT in workdir !");
		}
		elsif($offset_of_current_piece != $length_of_saved_data) {
			# -> Nope
			main::l(main::L_WARN, "Received unuseable data for $piece\@$ref->{hash}: Got $offset_of_current_piece data but expected to get $length_of_saved_data stuff");
			$rollover = 1;       # Go to next piece
			$next_offset = -1;
			$client_points -= 10;
			ZeroWorkPiece($ref->{hash},$piece);  #Piece is BR0ken -> Wipe data..
			SetPieceAsFree($ref->{hash},$piece); #..and requeue it as .free
		}
		else {
			# -> Ok, we can now store this chunk
			open(DATA, ">>$work_file") or Carp::confess("Unable to write to $work_file : $!");
			print DATA $ref->{data} or Carp::confess("WRITE failed: $!");
			close(DATA); #Needed to flush all buffers
			$length_of_saved_data = $next_offset = (-s $work_file); # -> update ;-)
			
			if($length_of_saved_data == $length_of_full_piece) {
				main::l(main::L_DEBUG_STORE, "STORE -> Piece is COMPLETE");

				# -> Complete piece!
				$rollover = 1; # Do a rollover, even if the chunk was garbage
				$next_offset = -1;
				
				open(DATA, $work_file) or Carp::confess("Unable to open $work_file : $!");
				my $verify_ok = VerifyPiece($torrent->{content}->{info}->{pieces},$piece, *DATA);
				close(DATA);
				
				if($verify_ok) {
					# -> Chunk was okay
					$client_points += 5;
					$verified = 1;
					SetPieceAsDone($ref->{hash},$piece);
				}
				else {
					#Failed to verify hash
					main::l(main::L_WARN, "STORE : Verification of $piece\@$ref->{hash} FAILED : Starting ROLLBACK");
					$client_points -= 30;
					ZeroWorkPiece($ref->{hash},$piece);  #Verification failed ; Remove the br0ken data from disc..
					SetPieceAsFree($ref->{hash},$piece); #..and requeue the piece
				}
			}
			elsif($length_of_saved_data > $length_of_full_piece) {
				#Should never happen
				Carp::confess("BUGBUG! Saved too much data in $work_file");
			}
			elsif( ($length_of_full_piece - $length_of_saved_data) < main::PIECESIZE ) {
				#May be the last piece..  # FIXME: Duplicate code: GetFreePieceOffset does almost the same
				$next_size = $length_of_full_piece - $length_of_saved_data;
			}
		}
		return({index=>$piece, size=>$next_size, offset=>$next_offset, rollover=>$rollover, verified=>$verified, points=>$client_points});
	}

1;

package Bitflu::Ansi;
use constant ANSI_ESC    => "\x1b[";
use constant ANSI_BOLD   => '1;';
use constant ANSI_BLACK  => '30m';
use constant ANSI_RED    => '31m';
use constant ANSI_GREEN  => '32m';
use constant ANSI_YELLOW => '33m';
use constant ANSI_BLUE   => '34m';
use constant ANSI_RSET   => '0m';


sub Green {
	my($s) = @_;
	my ($string,$end) = AnsiCure($s);
	$s = ANSI_ESC.ANSI_BOLD.ANSI_GREEN.$string.ANSI_ESC.ANSI_RSET;
	$s .= $end;
	return $s;
}

sub Blue {
	my($s) = @_;
	my ($string,$end) = AnsiCure($s);
	$s = ANSI_ESC.ANSI_BOLD.ANSI_BLUE.$string.ANSI_ESC.ANSI_RSET;
	$s .= $end;
	return $s;
}

sub Yellow {
	my($s) = @_;
	my ($string,$end) = AnsiCure($s);
	$s = ANSI_ESC.ANSI_BOLD.ANSI_YELLOW.$string.ANSI_ESC.ANSI_RSET;
	$s .= $end;
	return $s;
}

sub Red {
	my($s) = @_;
	my ($string,$end) = AnsiCure($s);
	$s = ANSI_ESC.ANSI_BOLD.ANSI_RED.$string.ANSI_ESC.ANSI_RSET;
	$s .= $end;
	return $s;
}

sub Bold {
	my($s) = @_;
	my ($string,$end) = AnsiCure($s);
	$s = ANSI_ESC."40m".ANSI_ESC.ANSI_BOLD.ANSI_YELLOW.$string.ANSI_ESC.ANSI_RSET;
	$s .= $end;
	return $s;
}

sub AnsiCure {
	my($s) = @_;
	my $badend = chop($s);
	if($badend ne "\n") { $s .= $badend ; $badend = undef }
	return($s,$badend);
}
1;


